#include <numeric> //accumulate

#include "computeId.h"
#include "migraine.h" // for DEBUG

CcomputeId calculQ;

using namespace std;
//***********************************************************
//pour l'elimination gaussienne dans PiCalcPimDG2002,
//directement import['e]e de Griffiths 2002
int Gaussian(picalctype **a,pirestype *b,const int m) { //FR new prototype 04/2010
int i,j,k,p;
picalctype tmp;
//vector<picalctype> ptmp;//modif avant c'etait un pointeur
picalctype* ptmp; //FR 04/2010

#ifdef DEBUG
if(gaussdebug){
   	printf("\n\n\tSystem of equations\n");
		for(int r=0;r<m;r++) {
			printf("\n\t");
			for(int s=0;s<m;s++) printf("a[%d][%d]=%3.4e ",r,s,(double) a[r][s]);
			printf("b[%d]=%.4e",r,(double) b[r]);
		}
	}
#endif

for(i=0;i< m-1; i++) {
	for(p=i;p<m;p++) if(a[p][i] != 0.0) break;
	if(p == m) {
	        printf("\n\tGaussian elimination failed 1\n");
            if (cinGetOnError) cin.get();
	        exit(1);
    }
	if(p != i) {
	//(qui?) verifier que ca va pas trop lentement -> comparer avec ou sans pointeurs
	//FR 04/2010 je suis revenu a des array...
		ptmp=a[p];a[p]=a[i];a[i]=ptmp;
		tmp=b[p];b[p]=b[i];b[i]=tmp;
	}
	for(j=i+1;j<m;j++) {
		tmp=(a[j][i])/(a[i][i]);
		for(k=0;k<m;k++) a[j][k] -= tmp*a[i][k];
		b[j] -= tmp*b[i];
	}
#ifdef DEBUG
	if(gaussdebug){
	   	printf("\n\n\tElimination step %d\n",i+1);
		for(int r=0;r<m;r++) {
			printf("\n\t");
			for(int s=0;s<m;s++) printf("a[%d][%d]=%3.4e ",r,s,(double) a[r][s]);
			printf("b[%d]=%.4e",r,(double) b[r]);
		}
	}
#endif
}
if(a[m-1][m-1] == 0.0) {
	        printf("\n\tGaussian elimination failed 2\n");
            if (cinGetOnError) cin.get();
	        exit(1);
    }
b[m-1] /= a[m-1][m-1];
for(i=m-2;i>=0;i--) {
	for(j=i+1;j<m;j++) b[i] -= a[i][j]*b[j];
	b[i] /= a[i][i];
	}
#ifdef DEBUG
if(gaussdebug){
   	printf("\n\n\tfinal step\n");
		for(int r=0;r<m;r++) {
			printf("\n\t");
			for(int s=0;s<m;s++) printf("a[%d][%d]=%3.4e ",r,s,(double) a[r][s]);
			printf("b[%d]=%.4e",r,(double) b[r]);
	}
}
#endif
return 0;
}
/*------------------end gaussian()------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/


int CcomputeId::setNh(CMigMod* MigMod,CMutMod* MutMod) {
      pson=MigMod->MigPSON;
      //Nhdeme=(*max_element(MigMod->Ma.begin(),MigMod->Ma.end()))*10000./pson;
      Nhdeme=max(Nhdeme,(picalctype) MutMod->ThetaDIG*10000./pson);
    //cout << "Nhdeme=" << Nhdeme << endl;
  return(0);
}

int CcomputeId::initMig(CMigMod* MigMod) {
          picalctype fac=(2.*Nhdeme*pson);
          eigvals.resize(pson);
          eigR.resize(pson);
          pirestype tmp;
         for (vector<vector<picalctype> >::iterator it=eigR.begin();it!=eigR.end();it++) it->resize(pson,0);
          Migmat.resize(pson);
         for (int ii=0;ii<pson;ii++) {
             Migmat[ii].resize(pson,0);
             for (int jj=0;jj<pson;jj++) Migmat[ii][jj]=MigMod->Mab[ii][jj]/fac; //Mab is always globally (N_T) scaled; this recovers the migration matrix
             Migmat[ii][ii]=0.;
             tmp=accumulate(Migmat[ii].begin(),Migmat[ii].end(),0.);
             Migmat[ii][ii]=1.-tmp;
//ostream_vector(MigMod->Mab[ii],cout);getchar();
//cout<<MigMod->DispPars[0]<<" "<<MigMod->DispPars[1];getchar();
//ostream_vector(MigMod->Mab[ii],cout);getchar();
//ostream_vector(Migmat[ii],cout);getchar();
         }
 return(0);
}

int CcomputeId::initMut(CMutMod* MutMod,picalctype& mufactor,string mutationModel) { // mufactor is the IBD/IIS correction (for the eigenvalues of the mutation matrix...)
    int stateNum=MutMod->Mut_AlleleNbr;
    picalctype twoNhmu;
    if(mutationModel=="PIM") twoNhmu=(MutMod->ThetaDIG/pson) *(stateNum-1)/stateNum; // this brings back to real mut from DIG seudo mut for PIM (not for SMM or GSM or ISM)
    else twoNhmu=MutMod->ThetaDIG/pson;
    mu=mufactor*twoNhmu/(2.*Nhdeme);
	ibdTable.resize(MutMod->Mut_AlleleNbr);
    //cout<<"TwoNhmu="<<twoNhmu<<"; mufactor="<<mufactor<<"; mu="<<mu << endl;
//getchar();
return(0);
}

int CcomputeId::eigenanalysis() {
    //jama_eig uses STL arrays so Migmat should be one such array
     CEigensystem<picalctype,picalctype> es(Migmat,true); // 'true' means symmetric matrix
     eigvals=es.getdvec(); // eigenvalues
     eigR=es.getV(); // (transpos'es //t mathematica) = dans le bon sens... = vectors in columns
     /*il me manque eigL. For symmetric mig matrices this is transpose(eigR)*/
//ostream_vector(eigR[0],cout);getchar();
//ostream_vector(eigR[1],cout);getchar();
return(0);
}

picalctype CcomputeId::IdbyD(int d1,int d2) { //identity by descent between genes in demes d1 and d2
      picalctype tmp,Q;
      gamma=pow( (1.-mu),(picalctype) 2.);
      picalctype fac=gamma/Nhdeme; //gB ...
    //cout<<"Nhdeme=" << Nhdeme<<"; mu= "<<mu<<"; fac= "<<fac << endl;//getchar();      // to solve eqs for within-deme identities: we need to construct a pson*pson matrix (Mmat array -- not ideal name here)
      for (int ii=0;ii<pson;ii++)  {
          bvec[ii]=0;
          for (int jj=0;jj<pson;jj++) {
          //first compute terms [ii,ii][jj,jj] of Inv(ImgA).gA=:W. This takes the form \sum_{kl} R_{ii,kl} D_{kl,kl} L_{kl,jj}
            tmp=0;
            for (int kk=0;kk<pson;kk++)  for (int ll=0;ll<pson;ll++) {
               tmp+=eigR[ii][kk]*eigR[ii][ll]*(eigvals[kk]*eigvals[ll]/(1.-gamma*eigvals[kk]*eigvals[ll]))*eigR[jj][kk]*eigR[jj][ll];
            }
            tmp*=fac;
            Mmat[ii][jj]=tmp;  //(WB)_{ii ii jj jj} or (WB)_w [ii,jj]
            bvec[ii]+=tmp; //sums to (WB)_w . 1
          } // end loop over jj
          Mmat[ii][ii]+=1.; //I+WB_w
      }
     // note that eigensystem(lvec) is not a simple as that of the original migration matrix
     // perhaps we could recompute it. But this should be enough:
     Gaussian(Mmat,bvec,pson); // solves (I+WB_w)Q_w= (WB)_w . 1
//cout<<bvec[0]<<" "<<bvec[1];getchar();
     if (d1==d2) return(bvec[d1]); //Identity by descent within deme d1
     // ELSE
     Q=0;
     for (int jj=0;jj<pson;jj++) {
         tmp=0;
     // we need (W.c)[d1,d2]=\sum_jj W[d1,d2][j,j]c_{j,j}.     W[d1,d2][j,j] takes the form \sum_{kl} R_{d1d2,kl} D_{kl,kl} L_{kl,jj}
            for (int kk=0;kk<pson;kk++)  for (int ll=0;ll<pson;ll++) {
               tmp+=eigR[d1][kk]*eigR[d2][ll]*(eigvals[kk]*eigvals[ll]/(1.-gamma*eigvals[kk]*eigvals[ll]))*eigR[jj][kk]*eigR[jj][ll];
            }
        tmp*=fac;
        Q+=tmp*(1-bvec[jj]);
     }
     return(Q);
}

picalctype CcomputeId::NotIdbyD() { //non - identity by descent between genes in a deme
    gamma=pow( (1.-mu),( picalctype ) 2. );
    //cout<<"Nhdeme=" << Nhdeme<<"; mu= "<<mu<< endl;
    return(Nhdeme*(1.-gamma)/(gamma+Nhdeme*(1.-gamma)));
}


