/***************************************************************************
© F. Rousset, R. Leblois 2007

francois.rousset@univ-montp2.fr
raphael.leblois@mnhn.fr

This file is part of Migraine. This software is a computer program
whose purpose is to perform statistical analyses.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

 ***************************************************************************/
#ifndef H_DGCMIGCMUT
#define H_DGCMIGCMUT
#include <iomanip>
#include <string>
#include <cstdlib>
#include <fstream>
#include <cstdio>
#include <iostream>
#include <fstream>

#include "dg.h"


//-----------------------------------------------------
//------declarations pour DG2002.cpp-------------------
//-----------------------------------------------------

//classe model de mutation
class CMutMod{
	public:

	parstype ThetaDIG;
    parstype FounderThetaDIG;//for OnePopFounderFlush, this is the theta value for the founder event
	parstype AncThetaDIG;//for OnePopVarSize & OnePopFounderFlush, this is the ancestral theta value
	parstype ThetaFactor;//pour corriger PIM/KAM
	parstype pGSM;
	int Mut_AlleleNbr;//nombre d'etat alleliques au locus considéré
	std::vector< std::vector<picalctype> > Pij;//matrice de transition
	std::vector<parstype> Station;//distrib stationnaire
	std::vector< std::vector<picalctype> > REigenVectors;//right eigenvector matrix
	std::vector< std::vector<picalctype> > LEigenVectors;//left eigenvector matrix
	std::vector<picalctype> EigenValues;//right eigenvector matrix	
	CMutMod();//constructeur
	int MakeMatrixPIM(int StateNum,parstype _Thetatrue);//remplit thetat et une matrice de mut PIM RL 04062008 only one procedure for PIMtrue and PIMDIG
	int MakeMatrixSMM(int StateNum,parstype _Theta);//remplit theta et une matrice demutation SMM
	int MakeMatrixGSM(int locAlleleNbr,parstype _Theta, parstype pGSM);
	int PrintMut();
	parstype CalculStation();//calcul la distribution stationnaire
	int ComputeEigensystems();//Compute right and left eigenvectors, and eigen values of the matrix Pij
	int ComputeEigensystems2();//Compute right and left eigenvectors, and eigen values of the matrix Pij
	int PrintStation();//affiche la distribution stationnaire des etats alleliques
};

//classe model de migration
class CMigMod{
	public:
	std::vector<parstype> DispPars; // parameter [estimand] std::vector describing dispersal;
	std::vector< std::vector<pascumultype> > Mab;//matrice de migration
	std::vector<pascumultype> Ma;//vecteur d'emmigration
	std::vector<pascumultype> PopSize;//vecteur des tailles relatives des pops
	pascumultype FounderPopSize;//Taille relative de la pop foundatrice dans OnePopFounderFlush
	std::vector<pascumultype> AncPopSize;//vecteur de tailles relatives de pop ancestrales
	pascumultype LogisticGrowthRate;//pour variation logistic de la taille de pop
	std::vector<pascumultype> ChangeTime;//vecteur des temps de changement demo (divergence ou changement de taille de pop)
	std::vector<int> PopMergedTab;//vecteur des pop ayant "disparues" fusionn['e]es avec la pop0
	std::vector<pascumultype> x;//si mpim
	std::vector<pascumultype> y;//si mpim
	int MigPSON;//nombre de pop totales, equiv de la variable globale PSON
//	int MakeMatrixSS(const std::vector<parstype> &Disp_Pars);//rempli une matrice de migration
	double init_file_et_critere; //-1 [`a] la construction=quand fichier lu; variable >=0 later
	cumultype MinChangeTime;//prochain temps de changement (divergence ou taille de, pop)
	cumultype PrevChangeTime;//precedent temps de changement (divergence ou taille de, pop)
	bool InAncStablePopbool;//boolean to know if we are in the ancestral stable population (for PopVarSize,FounderFlush & IM)
	int PopMergedNbr;//nombre de populations ayant disparues = ayant fusionn['e]es avec la pop0
//#ifdef ASYMPTO // que pour asympto:
	double limit_FSurUnMoinsF;
	std::valarray<cumultype>costable;
	cumultype DeuxPiSurNx;
//#endif
	float Nsigma2(); //Ntot sigma2...
	CMigMod();//constructeur
	int MakeMatrixSS(int& PopSampledOrNot,const std::vector<parstype> &Disp_Pars);//rempli une matrice de migration et Mig->MigPSON....
	int MakeMatrixNpop(int& PopSampledOrNot,const std::vector<parstype> &Disp_Pars);//rempli une matrice de migration pour Npop et Mig->MigPSON
	int MakeOnePop();//rempli tout ce qui est necessaire pour OnePop model et Mig->MigPSON
	//	CMigMod(int PopNum);//constructeur
//	int MakeMatrixSS(pascumultype MigRate);//rempli une matrice de migration stepping stone
	int PrintMig();
	int MakePopSize(std::vector<parstype> PopSizeTab);//rempli un vecteur de popsize toutes egales pour IBD ou valeures specifiques pour Npop
	int MakeAncPopSizeChangeTime(std::vector<parstype> AncPopSizeTab,std::vector<scaledTimeType> ChangeTimeTab);//rempli un vecteur de temps de divergence et de tailles de pop ancestrales
	int ResetChangeTime(std::vector<scaledTimeType> ChangeTimeTab); //met a jour le vecteur des temps de divergence a pres une fusion
	int PrintPopSize();
	int PrintAncPopSize();
	int PrintChangeTime();
	int CheckM();//check if Mab=Mi*Mj pour mpim
	int CheckMSmm();//check if migration is stepping stone
	int MergePop(int Sink,int Source,int PopSampledOrNot,std::vector<parstype> Disp_Pars,std::vector<parstype> AncPopSize);//set a new migration model after a population merge in the Split model
	int ApplyIntensityFunction(int pop, double intensityValue);
};

#ifdef NGIVEN //Cond non suff mais necess...
///// HMMMM virtual destructor pour CMigMOd ?
class CN_known : public CMigMod { //additional stuff that requires N to be known
	valarray<cumultype>lncharfn2; //log(values (charfunc squared))
//	int converge_coal();
	cumultype circularFourierinv(const int dx,const int gener);
//	cumultype charfuncpower2t(const int dx,const int gener);
public:
	scaledTimeType scaled_time_limit;
	double InvAsymptoNe; //The inverse effective size as *coal rate per generation*
	double FSurUnMoinsF;
	int Nhaploid;
	CN_known();
	CN_known(int Nhaploid);
	CN_known(CMigMod &Mig,int Nhaploid,int& GeoSpan);
	int calcule_Fourier_converge(int& GeoSpan);
};

#endif //NGIVEN

#endif
