/*
 * genetIS.cpp
 *
 *  Created on: 25 juil. 2013
 *      Author: cmerle
 */

#include "genetIS.h"
#include <cmath>

using namespace std;

/**************************************************************************
 *  Pairwise composite likelihood of a given sample                       *
 **************************************************************************/

inline int nb_pairs(int k){
  return k * (k-1) / 2;
}


double CalculPCL::logPCL(const std::vector<int> & n) {
  double result = 0.;
  for (size_t i = 0; i < n.size(); ++i){
    if (n[i] >= 2){
      result +=  nb_pairs(n[i]) * logPCL2((int)i, (int)i);
    }
    if (n[i] >= 1){
      for (size_t j = i+1; j < n.size() ; ++j){
        if (n[j] >= 1){
          result += n[i] * n[j] * logPCL2((int)i, (int)j) ;// manque un /2 CM 07.10.14 Ba non pas du tout puisque j>i 20.07.15
        }
      }
    }
  }
  return result;
}


/*inline double logPCL2(int i1, int i2, double rho, double theta) {
	int stepnumber = abs(i1-i2);
	return  stepnumber * log(rho) - .5 * log( 1.0 + 2.0 * theta ) ;
}*/

//double logPCL(const vector<int> & n, double theta) {
//	double result = 0.;
//	double rho = theta /(1.0 + theta + sqrt(1.0 + 2.0 * theta));
//	for (size_t i = 0; i < n.size(); ++i){
//		if (n[i] >= 2){
//			result +=  nb_pairs(n[i]) * logPCL2((int)i, (int)i, rho, theta);
//		}
//		if (n[i] >= 1){
//			for (size_t j = i+1; j < n.size() ; ++j){
//				if (n[j] >= 1){
//          result += .5 * n[i] * n[j] * logPCL2((int)i, (int)j, rho, theta) ;// manque un /2 CM 07.10.14
//          }
//          
//				}
//			}
//		}
//	}
//  //cout << "_________result = " <<result << endl; // CM 07.10.14
//	return result;
//}
//



/**************************************************************************
 *  Random stuff                                                          *
 **************************************************************************/

double tirageexpo(double ell){
  double u =  randunif();
  double y = - log(1.0 - u) / ell;
  return y;
}


/**************************************************************************
 *  IS algorithm for OnePopVarSize                                        *
 **************************************************************************/


// the c_\ell(rho) function of Rousset et al. (XXXX)
inline double cellrho(int ell, double rho){
	const double tmp = sqrt(1.0 - rho * rho);
	return (1.0 / tmp) * pow(rho / (1.0 + tmp), ell);
}



// lambda as a function of time
double SIS::fct_lambda () {
  if(0.0<=t and t<=T){
    return pow(thetaact/thetaanc,t/T);
  } else {   // ici t>T
    return thetaact/thetaanc;  }
}

double SIS::theta(){
  if (t<T) {
    return thetaact * pow(thetaanc/thetaact, t/T);
  } else {
    return thetaanc;
  }
}
// jump rate
double SIS::tauxsaut(){
  return sum_n * ((sum_n - 1.0)*fct_lambda() + thetaact);

}
/*
 // Projet de code d'autres modèles démographiques


// lambda as a function of time
double SIS::fct_lambda () {
  return thetaact/theta();
}
 
double SIS::theta(){
  if (VarSizeFunction=="Exponential") {
    if (t<T) {
      return thetaact * pow(thetaanc/thetaact, t/T);
    } else {
      return thetaanc;
    }
  }else if (VarSizeFunction=="Discrete"){
    if (t<T) {
      return thetaact;
    } else {
      return thetaanc;
    }
  }else if (VarSizeFunction=="Linear"){
    if (t<T) {
      return (thetaanc-thetaact)*t/T+ thetaact ;
    } else {
      return thetaanc;
    }
  }else return -1; 
}


// jump rate
double SIS::tauxsaut(){
  return sum_n * ((sum_n - 1.0)*fct_lambda() + thetaact);
  
}
*/

// draw the next date of jump
void SIS::tiragetpsacceptrejet(){
  sum_n = sum(n);
  double mu = tauxsaut(); 	  // majorant
  double u = 0.;
  do {
    mu = tauxsaut();
    u = randunif();
    t += tirageexpo(mu);
  } while (u >= tauxsaut() / mu);
}


bool SIS::update_current_timenew(){ // CM 24.03.14
  bool IsInPopAncForTheFirstTime=false;
  if (t < T){
    tiragetpsacceptrejet();
    if (t>T) { // CM 24.03.14
      IsInPopAncForTheFirstTime=true; // CM 24.03.14
      t=T;
    } // CM 24.03.14
  }
  else {    EventNbInPopAnc++;//CM 13.02.14
     t += tirageexpo(tauxsaut());
  }
  return IsInPopAncForTheFirstTime;  // CM 24.03.14
}

void SIS::update_current_timeold(){
  if (t <= T){
    tiragetpsacceptrejet();}
  else{
    EventNbInPopAnc++;//CM 13.02.14
    t += tirageexpo(tauxsaut());}
}

int SIS::minallele(){
    int i=0;
    while (n[i]==0) {
        i++;
    }
    return i;
}

int SIS::maxallele(){
    int i=nballele-1;
    while (n[i]==0) {
        i--;
    }
    return i;
}

double SIS::pihat(const int k) {
  double rho = theta() / (theta() + sum_n);
  double res = 0.0;
    for (int i = minall; i <= maxall ; i++){
      if (n[i]!=0) {
          int stepnumber = min(abs(k-i), nballele-abs(k-i));
          res += (double) n[i] * cellrho(stepnumber, rho);
      }
  }
  return res /(sum_n + theta());
}


// si return = 0 on a une mutation vers le type alphaa - 1
// si return = 1 on a une mutation vers le type alphaa + 1
// si return = 2 on a une coalescence
int SIS::randbeta(int alphaa) {
  vector<double> proba(3, 0.0);
  proba[0] = theta() * pihat(alphaa - 1) * 0.5;
  proba[1] = theta() * pihat(alphaa + 1) * 0.5;
  proba[2] = n[alphaa] ;
  return discreterand(proba);
}

// si return = 0 on a une mutation vers le type alphaaa - 1
// si return = 1 on a une mutation vers le type alphaaa + 1
// si return = 2 on a une coalescence

int SIS::randbetapartiel(int alphaaa) {
    vector<long double> proba(3, 0.0);
    int indice = 2;
    
    // Renormalisation : 

    long double Dn=((double)sum_n+1.0) * (((double)sum_n) + theta());
    
    long double denom =Dn*pihat(alphaaa)/((double)sum_n+1.0);

 
    proba[0] =n[alphaaa]/denom;
    long double somme= proba[0];
    
    //Vérification : somme proposal = 1
    
    /*proba[1] =theta() * pihat(alphaaa - 1) * 0.5/denom;
    proba[2] = theta() * pihat(alphaaa + 1) * 0.5 /denom;
    long double sommeproposal = proba[0]+proba[1]+ proba[2];
    cout << "somme proposal = " << sommeproposal << endl;
     */
    
    double u = randunif();

    
    /*if (proba[0]<u) {
        indice++;
        proba[1] =theta() * pihat(alphaaa - 1) * 0.5;
            if (proba[0]+proba[1]<u) {
                //proba[2] = theta() * pihat(alphaaa + 1) * 0.5;
                indice++;

            }
    }*/
    if(proba[0]<u){
        indice=0;
        proba[1] =theta() * pihat(alphaaa - 1) * 0.5/denom;
        somme+=proba[1];
        if (proba[0]+proba[1]<u) {
            indice=1;
            proba[2] = theta() * pihat(alphaaa + 1) * 0.5 /denom;
            somme+=proba[2];
        }
    }
  
    /*long double sommepihat= 0.0;
    for (int k=0; k<=nballele-1; k++) {
        sommepihat+=pihat(k);
        
    }*/
    /*double cumul= n[alphaaa];
    while (cumul<u) {
        indice++;
        cumul+= proba[indice];
    }*/
    
    return indice;

}



void SIS::algo_one_stepnew() {
  // Tirage de la nouvelle date
  bool IsInPopAncForTheFirstTimeInalgo_one_step=false;// CM 24.03.14
  IsInPopAncForTheFirstTimeInalgo_one_step=update_current_timenew();// CM 24.03.14
  if (!IsInPopAncForTheFirstTimeInalgo_one_step) {// CM 24.03.14
    int alpha = discreterand_sum(n, sum_n,minall);
    n[alpha]--; sum_n = sum(n);
    
    // Tirage du type d'événement
    int beta = alpha;
    int tmp = randbetapartiel(alpha);
    if (tmp == 0) beta = (alpha - 1) % n.size();
    if (tmp == 1) beta = (alpha + 1) % n.size();
    // Mise à jour des poids et de n
    if (tmp ==2) { // coalescence
      w *= (sum_n + 1.0) / (n[alpha] + 1.0) * pihat(alpha);
    } else { // mutation bw alpha -> beta
      w *= (n[beta] + 1.0) / (n[alpha] + 1.0) * (pihat(alpha) / pihat(beta));
      n[beta]++;
      sum_n = sum(n);
      minall = minallele();
      maxall = maxallele();
    }
    
  }
  // ajoute au maximum nb_iter itérations de SIS
  // si nb_iter < 0, va jusqu'au MRCA
  }
  
  
void SIS::algo_one_stepold() {
  // Tirage de la nouvelle date
  update_current_timeold();

  // Tirage type alpha qui va subir l'événement
  int alpha = discreterand_sum(n, sum_n,minall);
  n[alpha]--; sum_n = sum(n);

  // Tirage du type d'événement
  int beta = alpha;
  //int randbeta(alpha);
  int tmp = randbetapartiel(alpha);
  if (tmp == 0) beta = (alpha - 1) % n.size();
  if (tmp == 1) beta = (alpha + 1) % n.size();
  // Mise à jour des poids et de n
  if (tmp ==2) { // coalescence
    w *= (sum_n + 1.0) / (n[alpha] + 1.0) * pihat(alpha);
  } else { // mutation bw alpha -> beta
    w *= (n[beta] + 1.0) / (n[alpha] + 1.0) * (pihat(alpha) / pihat(beta));
    n[beta]++;
    sum_n = sum(n);
      minall = minallele();
      maxall = maxallele();
  }

}

// Construction par nombre d'itérations / nombre d'evts
// ajoute au maximum nb_iter itérations de SIS
// si nb_iter < 0, va jusqu'au MRCA
void SIS::algo(int nb_iter) {
  //cout << " " << endl; // CM 08.04.14
    minall = minallele();
    maxall = maxallele();
  sum_n = sum(n);
  int lulu = 0;
  while(sum_n >= 2.0 and (nb_iter < 0 or lulu < nb_iter)) {
    algo_one_stepold(); //CM 31.03.14
    lulu++;
  }
  sum_n = sum(n);
  if(not fini and sum_n < 1.9){
	  w *= 1.0 / (double)n.size();
	  fini = true;
  }
}

// Construction par nombre d'individus / nombre de coalescence
// ajoute au maximum nb_iter itérations de SIS
void SIS::algo_IndivNb(int IndivNb) {
  minall = minallele();
  maxall = maxallele();
  int actualsum_n=sum_n;
  int lulu = 0;
  while(sum_n >= 2.0 and (IndivNb<0 or IndivNb > lulu)) { // CM 08.04.14
    algo_one_stepold();// CM 08.04.14
    if (sum_n<actualsum_n-lulu) {// CM 08.04.14
      lulu++;// CM 08.04.14
    }
  }
  
  sum_n = sum(n);
  if(not fini and sum_n < 1.9){
	  w *= 1.0 / (double)n.size();
	  fini = true;
  }
}
