/**************************************************************************
© F. Rousset, R. Leblois 2007
francois.rousset@univ-montp2.fr
raphael.leblois@mnhn.fr

This file is part of Migraine. This software is a computer program
whose purpose is to perform statistical analyses.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author, the holder of the
economic rights, and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading, using, modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate, and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and, more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

 ***************************************************************************/
//hypercube sampling code originally from Maria de Iorio
#include <vector>
#include <set>
#include "latin.h"
#include "hypercube.h"
#include "writeR.h"


using namespace std;

/*
#ifdef UNIX_PARAL
int krigIterations=1; // modif FR 080610
#else
#ifdef KRIGITNBR  //-DKRIGITNBR=<n> preprocessor option
int krigIterations=KRIGITNBR; // 0: pas d'appel de R; 1: appel de R mais pas d'iteration; >1: iterations
#else
int krigIterations=1; // 0: pas d'appel de R; 1: appel de R mais pas d'iteration; >1: iterations
#endif
#endif
*/

int _mls;
unsigned int ParameterNumber; //no need for initialization, always gets _lb.length().
int parDigits=7; // number of digits for parameter coordinates in pointls and nextpoints file. 7th meaningful for g as g->1
bool allowNewReplicates=false;
vector<int> fittedpars;
vector<bool> fittedindic;
std::vector<double> testvalue; // to test for inclusion in CI
std::vector<Ctestpoint > testPointList;

Ctestpoint ParMaxima; // not a test point, but the same structure can be used
Ctestpoint ParMinima;
int fittedparamnbr;
//int ParameterNumber; //no need for initialization, always gets _lb.length().
//float fixedSmoothness=numeric_limits<float>::quiet_NaN(); // default value
ofstream ptls;
bool noRCallBool=false,noRfileBool=false,
	 singlePointBool=false,calculQused=false,
	 userPointMinOrMax = false;	// CBR: indicates whether the Pointmin/Pointmax setting has been specified by the user
bool use_Rpackage=true;

string krigingCov="Matern"; // default value, but overridden through KrigingCov setting in migraine.cpp
CSampling DesignPoints;
CRParameters Rparams;

float groot(float fS2) {// from sigma2|cond to g. See migraine doc and parametrisation_2D.important.nb
   	long double S2=(long double)(fS2);
   	if (D1IBDbool) return float(((1.+2.*S2-sqrt(1.+8.*S2))/(2.*S2))); // solution of S2 == (1 + g)/(1 - g)^2
   	else if (D2IBDbool) { // S2 doit etre axial !! on pourrait inclure les deux cas et /2 quand necessaire
   	   // solution of S2 ((1 - g)^2 (2 - g)) == (1 + g)
       long double S2threshold=(143. + 19.*sqrt(57.))/16.,machin;
       if (S2<0.5) {
           cerr<<"(!) Impossible S2 value in hypercube_sampling. I exit";
           exit(-1);
           // reason is that S2 mini is axial MS of 4-neighbours SS....
       }
       if (S2<S2threshold) {
           machin=(S2*S2*(63. - 2.*S2) + 3.*sqrt(3.)*sqrt(pow(S2,(long double)(3.))*(4. + (143. - 8.*S2)*S2)));
           machin=pow(machin,(long double)(1./3.));
           return float((8.*S2 - (pow(2.,4./3.)*(-3. + S2)*S2)/machin - pow((long double)(2.),(long double)(2./3.))*machin)/(6.*S2));
       } else {
           machin=acos(((63. - 2.*S2)*S2*S2)/(2.*pow((long double)((S2-3.)*S2),(long double)(3./2.))));
           return float(4./3. - (2.*pow((long double)((-3. + S2)*S2),(long double)(1./2.))*cos(machin/3.))/(3.*S2));
       }
   }
return(-1);
}

int CRParameters::fill(int it) {
   iter=it;
   nextPointNumber=-1; // (re)mise a zero prudente
   if (cmp_nocase(boundsSource[std::min(it,int(boundsSource.size())-1)],"From2DCI")==0)
      nextBounds="From2DCI";
   else if (cmp_nocase(boundsSource[std::min(it,int(boundsSource.size())-1)],"From1DCI")==0)
      nextBounds="From1DCI";
   else if (cmp_nocase(boundsSource[std::min(it,int(boundsSource.size())-1)],"FromPoints")==0)
      nextBounds="FromPoints";
   else if (cmp_nocase(boundsSource[std::min(it,int(boundsSource.size())-1)],"notFromR")==0) {
      nextBounds="notFromR";
   } else {
      /* If
            PointNumber=1000,2197
            writeSequence=ReadPoints,Append
         migraine expects 2197 points in nextpoints.txt
      */
      nextBounds="pointsFromR"; // implements the default
      nextPointNumber=sequencePointNumber[std::min(it+1,int(sequencePointNumber.size())-1)]; // CBR modified 15.02.2013, was min(it,...)
   }
return(0);
}

int provideQuotes(vector<string>& v) {
    stringstream stst;
    if (v.size()==0) {v.push_back("\"\"");return(0);} //R will get an empty vector 'c("")'
    for (vector<string>::iterator it=v.begin();it!=v.end();it++) {
        if(it->size()==0) {
            cerr<<"(!) Empty element in provideQuotes !";
            if (cinGetOnError) cin.get();
            exit(-1);
        }
        if ((*it)[0]!='\"') {
            stst<<"\""<<(*it)<<"\"";
            (*it)=stst.str();
            stst.str("");
            stst.clear();
        }
    }
    return(0);
}

int writeCfloatmapAsRlist(std::map<std::string, std::vector<float> >listmap,std::string listname,std::ostream& stream) {
    stream<<listname<<"<-list()"<<endl;
    map<string, vector<float> >::iterator it;
    for(it=listmap.begin();it!=listmap.end();it++) { // loop on map elements
        vector<float>::iterator vit;
       stream<<listname<<"[[\""<<it->first<<"\"]]<-c(";
       vit=it->second.begin();
       if (it->second.size()>0) stream<<(*vit);
       vit++;
       while(vit!=it->second.end()) {stream<<","<<(*vit);vit++;} // loop on vector 'second' elements
       stream<<")"<<endl;
    }
return(0);
};



int Ctestpoint::write(std::ostream& stream, std::string initstring) {
    stringstream Rlist;
    bool unknownvar=false;
     Rlist<<initstring;
      for (unsigned int jj=0;jj<names.size();jj++) { //this must convert to the internal variables, ie latt2Ns2
          //conversion par standardizeNames avant ? suffit pas.
        if (jj>0) Rlist<<",";
        if (cmp_nocase(names[jj],"lattNb")==0) {
           if (D2IBDbool)
               Rlist<<"latt2Ns2="<<values[jj]<<"/pi ";
           else
               Rlist<<"latt2Ns2="<<values[jj];
        } else if (cmp_nocase(names[jj],"geoNb")==0 || cmp_nocase(names[jj],"Nb")==0) {
               Rlist<<"latt2Ns2="<<values[jj]<<"/GP$Nbfactor";
        } else if (cmp_nocase(names[jj],"2Nm")==0 || cmp_nocase(names[jj],"twoNm")==0) {
            Rlist<<"twoNm="<<values[jj];
        } else if (cmp_nocase(names[jj],"pGSM")==0 || cmp_nocase(names[jj],"p")==0) {
            Rlist<<"pGSM="<<values[jj];
        } else if (cmp_nocase(names[jj],"2Nmu")==0 || cmp_nocase(names[jj],"twoNmu")==0) {
            Rlist<<"twoNmu="<<values[jj];
        } else if (cmp_nocase(names[jj],"g")==0) {
            Rlist<<"g="<<values[jj];
        } else if (cmp_nocase(names[jj],"Q1")==0 || cmp_nocase(names[jj],"Q")==0) {
            Rlist<<"Q1="<<values[jj];
        } else if (cmp_nocase(names[jj],"M1")==0 || cmp_nocase(names[jj],"twoNm1")==0 || cmp_nocase(names[jj],"2Nm1")==0) {
            Rlist<<"M1="<<values[jj];
        } else if (cmp_nocase(names[jj],"m1overmu")==0) {
            Rlist<<"m1overmu="<<values[jj];
        } else if (cmp_nocase(names[jj],"M2")==0 || cmp_nocase(names[jj],"twoNm2")==0 || cmp_nocase(names[jj],"2Nm2")==0) {
            Rlist<<"M2="<<values[jj];
        } else if (cmp_nocase(names[jj],"m2overmu")==0) {
            Rlist<<"m2overmu="<<values[jj];
        } else if (cmp_nocase(names[jj],"NMratio")==0) {
            Rlist<<"NMratio="<<values[jj];
        } else if (cmp_nocase(names[jj],"mratio")==0) {
            Rlist<<"mratio="<<values[jj];
        } else if (cmp_nocase(names[jj],"Nratio")==0 || cmp_nocase(names[jj],"PopSizeRatio")==0) {
            Rlist<<"Nratio="<<values[jj];
        } else if (cmp_nocase(names[jj],"Nancratio")==0 || cmp_nocase(names[jj],"AncPopSizeRatio")==0) {
            Rlist<<"Nratio="<<values[jj];
        } else if (cmp_nocase(names[jj],"NactNfounderratio")==0 || cmp_nocase(names[jj],"ActFounderPopSizeRatio")==0) {
            Rlist<<"NactNfounderratio="<<values[jj];
        } else if (cmp_nocase(names[jj],"NfounderNancratio")==0 || cmp_nocase(names[jj],"FounderAncPopSizeRatio")==0) {
            Rlist<<"NfounderNancratio="<<values[jj];
        } else if (cmp_nocase(names[jj],"TSplit")==0 || cmp_nocase(names[jj],"T")==0
                    || cmp_nocase(names[jj],"TChange")==0 || cmp_nocase(names[jj],"Tdiv")==0) {
            Rlist<<"T="<<values[jj];
        } else if (cmp_nocase(names[jj],"Tgmu")==0) {
            Rlist<<"Tgmu="<<values[jj];
        } else if (cmp_nocase(names[jj],"DSplit")==0 || cmp_nocase(names[jj],"D")==0
				   || cmp_nocase(names[jj],"DChange")==0 || cmp_nocase(names[jj],"Duration")==0) {
            Rlist<<"D="<<values[jj];
        } else if (cmp_nocase(names[jj],"Dgmu")==0) {
            Rlist<<"Dgmu="<<values[jj];
        } else if (cmp_nocase(names[jj],"2Nancmu")==0 || cmp_nocase(names[jj],"twoNancmu")==0) {
            Rlist<<"twoNancmu="<<values[jj];
        } else if (cmp_nocase(names[jj],"2Nfoundermu")==0 || cmp_nocase(names[jj],"twoNfoundermu")==0) {
            Rlist<<"twoNfoundermu="<<values[jj];
        } else if (cmp_nocase(names[jj],"growthRate")==0 || cmp_nocase(names[jj],"growth")==0 || cmp_nocase(names[jj],"LogisticGrothRate")==0) {
            Rlist<<"growthRate="<<values[jj];
        } else if (cmp_nocase(names[jj],"2NmuSEQ")==0 || cmp_nocase(names[jj],"twoNmuSEQ")==0) {
            Rlist<<"twoNmuSEQ="<<values[jj];
        } else if (cmp_nocase(names[jj],"2NancmuSEQ")==0 || cmp_nocase(names[jj],"twoNancmuSEQ")==0) {
            Rlist<<"twoNancmuSEQ="<<values[jj];
        }
// RAPH : ici tu peux rajouter des variables...
          else {
            std::cout<<"(!) Unknown variable name '"<<names[jj]
               <<"' as setting values for testPoint. LRT computation is skipped.\n";
             unknownvar=true;
        }
     }

     Rlist<<")"<<endl;
     if ( ! unknownvar) {
         stream<<Rlist.str();
     } else stream<<"   tmplist<-NULL   ## a testPoint in settings was ignored because a parameter value was unknown."<<endl;
return(0);
} //end Ctestpoint::write()

float Ctestpoint::scanForVar(string varname) {
      for (unsigned int jj=0;jj<names.size();jj++) {
          if (cmp_nocase(names[jj],varname)==0) return(values[jj]); //returns the value associated with 'varname' in the point
      }
      return(numeric_limits<float>::quiet_NaN()); // if no value found, returns NaN
} //end Ctestpoint::scanForVar()


int CRParameters::writeR(std::ostream& stream) {
//FR->RL le but est de regrouper tous les passages de parametres de C a R dans cette fonction
// et aussi de les minimiser...
// ce n'est que le debut...
   /** info on pointls file**/
   stream<<"jobSampleNbr<-"<<_mls<<std::endl;
   stream<<"paramnbr<-"<<ParameterNumber<<" ## fitted or not"<<std::endl;
   stream<<"NbrMultilocusLiks<-"<<1+int(asymptoBool)<<" ## number of multilocus likelihood estimators per row (generally one, but two is possible)"<<std::endl;
   if (subsetPointls.size()>0) stream<<"subsetPtls<-("<<subsetPointls<<") ## selection of rows in pointls"<<std::endl;
   else stream<<"subsetPtls<-NULL ## no selection of rows in pointls"<<std::endl;
   /** info on statistical model**/
   if(IBDbool) stream<<"DemographicModel<-\"IBD\" "<<std::endl; // actual DemographicModel string is PlanarIBD or similar...
   else stream<<"DemographicModel<-\""<<DemographicModel<<"\" "<<std::endl;
   std::vector<std::string> CopyForRcodeWriting;
   stream<<"GP<-list() ## GP: Global Parameters (set by the C code for the R session, by contrast with GV, though some GV can also be set by the C code)"<<std::endl;
   CopyForRcodeWriting=CanonicalNames; //  for stl::vectors, '=' has the nice copy constructor meaning
   provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
   stream<<"GP$ParameterNames<-c("; // to declare eg log-transformations of variables
   ostream_vector(CopyForRcodeWriting,stream,",");
   stream<<") ## Canonical order of elements in parameter vector"<<endl;
   stream<<"resultsName <- paste(\"results_\",jobSampleNbr,\".txt\",sep=\"\")"<<endl;
   stream<<"abyss <- saveOldFile(resultsName) "<<endl;
   stream<<"cleanResu <- file(resultsName, \"w\")"<<std::endl;
   stream<<"write(\"Migraine "<<version<<"\",file=cleanResu)"<<std::endl;
   stream<<"write(paste(\"R code run on \",date()),file=cleanResu)"<<std::endl;
   stream<<"write(\"\",file=cleanResu)"<<std::endl;
   stream<<"write(\"Data file: "<<glob_gp_filename<<"\",file=cleanResu)"<<std::endl;
   stream<<"write(\"Settings file: "<<settingsfilename<<"\",file=cleanResu)"<<std::endl;
   stream<<"GP$graphicsFormat<-\""<<graphicsFormat<<"\""<<std::endl;
   provideQuotes(Rarguments);
   stream<<"GP$options<-c(";
   ostream_vector(Rarguments,stream,",");
   stream<<")"<<endl;
   provideQuotes(KrigSpace);
   stream<<"GP$FONKgSpace<-c(";  // to declare Kriging with non-canonical variables
   ostream_vector(KrigSpace,stream,",");
   stream<<")"<<endl;
	//CopyForRcodeWriting.resize(KrigScale.size());
	//copy(KrigScale.begin(),KrigScale.end(),CopyForRcodeWriting.begin());
   CopyForRcodeWriting=KrigScale; //  for stl::vectors, '=' has the nice copy constructor meaning
   provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
   stream<<"GP$FONKgScale<-c("; // to declare eg log-transformations of variables
   ostream_vector(CopyForRcodeWriting,stream,",");
   stream<<")"<<endl;
   stream<<"names(GP$FONKgScale)<-GP$FONKgSpace"<<endl;
   CopyForRcodeWriting=BoundsDefs; //  for stl::vectors, '=' has the nice copy constructor meaning
    for (vector<string>::iterator it=CopyForRcodeWriting.begin();it!=CopyForRcodeWriting.end();it++) {
        if (cmp_nocase((*it),"nb")==0) {
            (*it)="latt2Ns2";
            /* cf standardization of BoundsDefs: Nb was purposely not converted..*/
        }
    }
   provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
   stream<<"GP$samplingSpace<-c(";  // to declare Kriging with non-canonical variables
   ostream_vector(CopyForRcodeWriting,stream,",");
   stream<<") ## ONLY for next points"<<endl;
   CopyForRcodeWriting=samplingScale; //  for stl::vectors, '=' has the nice copy constructor meaning
   provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
   stream<<"GP$samplingScale<-c(";  // to declare Kriging with non-canonical variables
   ostream_vector(CopyForRcodeWriting,stream,",");
   stream<<") ## ONLY for next points"<<endl;
   stream<<"names(GP$samplingScale)<-GP$samplingSpace"<<endl;
   //
   CopyForRcodeWriting=extraScale; //  for stl::vectors, '=' has the nice copy constructor meaning
   provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
   stream<<"GP$extraScale<-c(";  // to declare Kriging with non-canonical variables
   for (unsigned int it=0;it<extraScale.size();it++) {
        stream<<extraNames[it]<<"="<<CopyForRcodeWriting[it];
        if(extraScale.size()>1 && it<(extraScale.size()-1)) stream<<",";
   }
   stream<<") ## "<<endl;
   //
   stream<<"## (!) plotRange can contain Nb rather than latt2Ns2"<<endl;
   writeCfloatmapAsRlist(plotRange,"GP$plotRange",stream);
   //
   writeCfloatmapAsRlist(graphicPars,"GP$graphicPars",stream);
   // migraine's default graphics parameters (par):
   stream<<"if (is.null(GP$graphicPars$cex.main)) GP$graphicPars$cex.main<-1.5 ## larger default font for main title"<<endl;
   stream<<"if (is.null(GP$graphicPars$cex.lab)) GP$graphicPars$cex.lab<-1.5 ## larger default font for axis title"<<endl;
   stream<<"if (is.null(GP$graphicPars$cex.axis)) GP$graphicPars$cex.axis<-1.4 ## larger default font for tick labels"<<endl;
   stream<<"if (is.null(GP$graphicPars$labcex)) GP$graphicPars$labcex<-0.9 ## for contour; default 0.6. "<<endl;
   stream<<"if (is.null(GP$graphicPars$tcl)) GP$graphicPars$tcl<-NA ## smaller ticks"<<endl;
   stream<<"if (is.null(GP$graphicPars$mgp)) GP$graphicPars$mgp<-c(3.5,0.5,0) ## tick labels closer to axis than default, axis label more distant"<<endl;
   stream<<"if (is.null(GP$graphicPars$mar)) GP$graphicPars$mar<-c(5,5.1,4,2)+0.1 ## bltr axis label positions on each side; default c(5, 4, 4, 2) + 0.1, starting from 0"<<endl;
   stream<<"if (is.null(GP$graphicPars$xmaxticks)) GP$graphicPars$xmaxticks<-9 ## for maketicks -> nice, not for par"<<endl;
   stream<<"if (is.null(GP$graphicPars$ymaxticks)) GP$graphicPars$ymaxticks<-10 ## for maketicks -> nice, not for par "<<endl;
   stream<<"ParArgsNames<-intersect(names(GP$graphicPars),graphics:::.Pars) ## using names of 'par' arguments, see ?par"<<endl;
   stream<<"ParArgs<-GP$graphicPars[ParArgsNames] ## only  'par' arguments"<<endl;
   stream<<"epsArgsNames<-intersect(names(GP$graphicPars),c(\"height\",\"width\")) ## "<<endl;
   stream<<"epsArgs<-GP$graphicPars[epsArgsNames] ## only  'eps' arguments"<<endl;
   //
   // other graphic control
//   stream<<"GP$colors<-\"tim\""<<endl;
   //
   CopyForRcodeWriting=PlotOptions; //  for stl::vectors, '=' has the nice copy constructor meaning
   provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
   stream<<"GP$PlotOptions<-c(";
   ostream_vector(CopyForRcodeWriting,stream,",");
   stream<<")"<<endl;
//   stream<<"if ( ! any(c(\"AllSurf\" %innc% GP$PlotOptions,\"All1DProfiles\" %innc% GP$PlotOptions,\"All2DProfiles\" %innc% GP$PlotOptions))) {"<<endl;
//   stream<<"   GP$PlotOptions<-c(GP$PlotOptions,\"Basic\")} ## Hence Basic implements the default"<<endl;
    if (oneDimCIvars.size()>0) {
        CopyForRcodeWriting=oneDimCIvars; //  for stl::vectors, '=' has the nice copy constructor meaning
        provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
        stream<<"oneDimCIvars<-c(";
        ostream_vector(CopyForRcodeWriting,stream,",");
        stream<<") ## (an empty c() value will be meaningful for nextBounts)"<<endl;
    } else { //currently provideQuotes on zero-size vector produces the string "" and in R, length("") = 1
        stream<<"oneDimCIvars<-c()"<<endl;
    }
    if (spec1DProfiles.size()>0) {
        CopyForRcodeWriting=spec1DProfiles; //  for stl::vectors, '=' has the nice copy constructor meaning
        provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
        stream<<"spec1DProfiles<-c(";
        ostream_vector(CopyForRcodeWriting,stream,",");
        stream<<")"<<endl;
    } else { //currently provideQuotes on zero-size vector produces the string "" and in R, length("") = 1
        stream<<"spec1DProfiles<-c()"<<endl;
    }
    if (spec2DProfiles.size()>0) {
        CopyForRcodeWriting=spec2DProfiles; //  for stl::vectors, '=' has the nice copy constructor meaning
        provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
        stream<<"spec2DProfiles<-c(";
        ostream_vector(CopyForRcodeWriting,stream,",");
        stream<<")"<<endl;
    } else { //currently provideQuotes on zero-size vector produces the string "" and in R, length("") = 1
        stream<<"spec2DProfiles<-c()"<<endl;
    }

    if (IBDbool) {
//        stream<<"GV$S2factor<-"<<geoBinWidth*geoBinWidth<<" ##  the correction from condS2 in lattice unit to user's units"<<std::endl;
           stream<<"GV$S2factor<-1 ##  other values add more confusion that anything else"<<std::endl;
        if (D1IBDbool) {
            stream<<"D2IBDbool<-FALSE;D1IBDbool<-TRUE "<<std::endl;
            stream<<"GV$Nbfactor<-"<<geoBinWidth<<" ##  correction from 2NmS2 to 2Ds2=Nb"<<std::endl;
            if (geoUnit.size()>0)
               stream<<"GeoUnit<-\""<<geoUnit<<"\" ## string that gives the Nb unit in plot"<<std::endl;
            else {
               if (geoBinWidth==1) {
                  stream<<"GeoUnit<-\"lattice units\" ##  Migraine has not found a non-trivial bin width and user has not provided the unit"<<std::endl;
               } else {
                  stream<<"GeoUnit<-\"\" ## Migraine has found a non-trivial bin width but user has not provided the unit"<<std::endl;
               }
            }
            stream<<"print(paste(\"Geographic bin width= \",GV$Nbfactor,sep=\"\"))"<<std::endl;
            stream<<"write(paste(\"Geographic bin width= \",GV$Nbfactor,sep=\"\"),file=cleanResu)"<<std::endl;
            stream<<"GV$mincondS2<-1"<<std::endl;
        } else { //D2IBD
            stream<<"D2IBDbool<-TRUE;D1IBDbool<-FALSE "<<std::endl;
            stream<<"GV$Nbfactor<-pi ##  from 2[N=D]m condS2 (where N=D in user's units) to 2piDs2 in user units "<<std::endl;
            stream<<"GeoUnit<-\"\" ## string that would gives the Nb unit in 1D IBD plot"<<std::endl;
            stream<<"GV$mincondS2<-1/2"<<std::endl;
        }
        stream<<"if ( ! is.null(GP$plotRange[[\"Nb\"]])) GP$plotRange[[\"latt2Ns2\"]]<-GP$plotRange[[\"Nb\"]]/GV$Nbfactor"<<std::endl;
    } else { // not IBD
       stream<<"D2IBDbool<-FALSE;D1IBDbool<-FALSE "<<std::endl;
       stream<<"GeoUnit<-\"\" ## string that would gives the Nb unit in 1D IBD plot"<<std::endl;
   }
   if (CovFnParam.size()>0) {
      stream<<"CovFnParamInSettingsBool<-TRUE"<<std::endl;
      stream<<"GV$CovFnParam<-c(";
      ostream_vector(CovFnParam,stream,",");
      stream<<")"<<std::endl;
   } else stream<<"GV$CovFnParam<-numeric(0);CovFnParamInSettingsBool<-FALSE"<<std::endl;
//   if(perLocusBool)
//       krigR<<"nlociperrow<-"<<int(perLocusBool)*nloci<<" ## >0 if per locus likelihoods are available"<<std::endl;
   if (KrigLoci.size()>0) {
      stream<<"fittedLoci<-c(";
      ostream_vector(KrigLoci,stream,",");
      stream<<")"<<endl;
   } else stream<<"fittedLoci<-integer(0) ## hmmm character(0) ?"<<std::endl;

   stream<<"## Booleans: "<<std::endl;
    /** FR->FR some boolean(s) that could be converted with more work...**/
   if (useCbool)
      stream<<"useC<-T ## to use C code for Kriging"<<std::endl;
   else
      stream<<"useC<-F ## to use C code for Kriging"<<std::endl;
    /** selection of points**/
   //iter is the value given  as argument to Rparams.fill and thus starts from 1 for first iteration
   stream<<"GP$LikStatistic<-\""<<sequenceStat[std::min(iter,int(sequenceStat.size())-1)]<<"\" ## PAC/IS..."<<std::endl;	// CBR modified 15.02.2013, was min(iter-1,...)
   stream<<"GP$designRetain<-"<<designRetain<<" ## fraction of points in the Top retained for Kriging, removing occurrences of close points "<<std::endl;
   stream<<"GP$GCVdesignRetain<-"<<GCVdesignRetain<<" ## fraction of points in the Top retained for GCV, removing occurrences of close points "<<std::endl;
   //stream<<"maxDoubles<-"<<maxKrigedReplicatedPoints<<" ## "<<std::endl; //RL->FR 092016 : a enlever définitivement
     stream<<"GCVptnbr<-"<<GCVptnbr<<" ## nbr of optimization steps for GCV"<<std::endl;
    //krigR<<"penalizeGCVBool<-FALSE ##declaration/initialization, not a control from the C code"<<std::endl;
    //krigR<<"penalizedGCVBool<-FALSE ##declaration/initialization, not a control from the C code"<<std::endl;
     stream<<"GCVoptimSteps<-"<<GCVoptimSteps<<" ## nbr of optimization steps for GCV"<<std::endl;
    /** other Kriging parameters**/
     stream<<"minKrigPtNbr<-"<<minKrigPtNbr<<" ## nbr of optimization steps for GCV"<<std::endl;
    stream<<"covFamily<-\""<<krigingCov<<"\" ## !!!!! covariance function for kriging"<<std::endl;
    stream<<"trendOrder<-"<<trendOrder<<" ## polynomial order of trend surface (fixed effect of surface fit)"<<std::endl;
    stream<<"meanNeighbor<-"<<meanNeighbor<<" ## poor hack for [Taper which does not work]"<<std::endl;
    stream<<"krigmax<-"<<krigmax<<std::endl;
    stream<<"\tkriglength<-"<<kriglength<<"## will be overwritten if krigmax<0"<<std::endl;
    stream<<"\tkrigoverlap<-"<<krigoverlap<<"## will be overwritten if krigmax<0"<<std::endl;
     if (maxKrigPtNbr==-1)
        stream<<"maxKrigPtNbr<-FALSE ## max pts for Kriging"<<std::endl;
    else
        stream<<"maxKrigPtNbr<-"<<maxKrigPtNbr<<" ## max pts for Kriging"<<std::endl;


    /** hull parameters **/
    stream<<"GV$hullExpandFactor<-"<<hullExpandFactor<<std::endl;

    /** profile grid parameters **/
   //iter is the value given  as argument to Rparams.fill and thus starts from 1 for first iteration
    stream<<"gridstepsNbr<-"<<sequenceGridSteps[std::min(iter,int(sequenceGridSteps.size())-1)]<<std::endl;	// CBR modified 15.02.2013

    /** CI, LRT, nextpoints parameters **/
    stream<<"profileType<-tolower(\""<<profileType<<"\")"<<std::endl;
    stream<<"CILevel<-"<<CILevel<<std::endl;
    stream<<"CIdlr<- -qchisq("<<1-CILevel<<",1)/2 ## for P="<<CILevel<<std::endl;
    stream<<"NextBoundsLevel<-"<<NextBoundsLevel<<std::endl;
    stream<<"NextBoundsdlr<- -qchisq("<<1-NextBoundsLevel<<",1)/2 ## for P="<<NextBoundsLevel<<std::endl;
    stream<<"nextBounds<-\""<<nextBounds<<"\""<<std::endl;
    stream<<"anyTwoNmuMin<-"<<anyTwoNmuMin<<std::endl;
    stream<<"anyTwoNmMin<-"<<anyTwoNmMin<<std::endl;
    stream<<"anyTimeMin<-"<<anyTimeMin<<std::endl;
    if(!std::isinf(anyTwoNmuMax))
        stream<<"anyTwoNmuMax<-"<<anyTwoNmuMax<<std::endl;
      else stream<<"anyTwoNmuMax<-Inf"<<std::endl;
    if(!std::isinf(anyTwoNmMax))
        stream<<"anyTwoNmMax<-"<<anyTwoNmMax<<std::endl;
      else stream<<"anyTwoNmMax<-Inf"<<std::endl;
    if(!std::isinf(anyTimeMax))
        stream<<"anyTimeMax<-"<<anyTimeMax<<std::endl;
      else stream<<"anyTimeMax<-Inf"<<std::endl;
    stream<<"nextPointNumber<-"<<nextPointNumber<<std::endl;
    stream<<"ptSamplingSeed<-"<<_ptSamplingSeed<<" ## random seed for sampling new points"<<std::endl;
    stream<<"parDigits<-"<<parDigits<<" ## digits of parameter coordinates in files"<<std::endl;
    /** one ugly numerical setting for optim() **/
    // a unit change in par/pascale should produce a unit change in fn()/fnscale
    stream<<"scalefactor<-"<<scalefactor<<" ## for (fnscale,parscale in optim())"<<std::endl;
    stream<<"write(\"\",file=cleanResu)"<<std::endl;

	return(0);
} // end CRParameters::writeR()

/*CSampling::CSampling(const int ParameterNumber,const int PointNumber){
this->ParameterNumber=ParameterNumber;
this->PointNumber=PointNumber;
DesignPoint.resize(PointNumber);
for(std::vector<std::vector<float> >::iterator ii=DesignPoint.begin();ii<DesignPoint.end();ii++){
	(*ii).resize(ParameterNumber);
	}
};*/

/*void CSampling<pointtype>::redim(const int parnbr,const int ptnbr){
DesignPoint.resize(ptnbr);
for(std::vector<std::vector<pointtype> >::iterator ii=DesignPoint.begin();ii<DesignPoint.end();ii++){
	(*ii).resize(parnbr);
	}
};*/

int CRParameters::write_R_GP_code(std::ostream& stream) {


    std::vector<std::string> CopyForRcodeWriting;


    stream<<"GP<-list()"<<std::endl;
    stream<<"GP$usedBy <- \"Migraine\""<<std::endl;
    stream<<"GP$jobSampleNbr<-"<<_mls<<std::endl;
    stream<<"GP$coreNbr<-"<<coreNbr<<std::endl;
    stream<<"GP$dataFile <- paste(\"pointls_\",GP$jobSampleNbr,\".txt\",sep=\"\")"<<std::endl;
    stream<<"GP$estimOutf <- file(paste(\"output_\",GP$jobSampleNbr,\".txt\",sep=\"\"), \"w\")"<<std::endl;
    stream<<"GP$nextPointsf <- paste(\"nextpoints_\",GP$jobSampleNbr,\".txt\",sep=\"\")"<<std::endl;
    stream<<"GP$cleanResu <- file(paste(\"results_\",GP$jobSampleNbr,\".txt\",sep=\"\"), \"w\")"<<std::endl;

    stream<<"GP$plotFiles<-list() ## sinon le premier appel de 'plotFiles[[filename]]<-' cree un vecteur de lgr 1 et le 2 appel plante"<<std::endl;
    CopyForRcodeWriting=CanonicalNames; //  for stl::vectors, '=' has the nice copy constructor meaning
    provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
    stream<<"GP$ParameterNames<-c("; // to declare eg log-transformations of variables
    ostream_vector(CopyForRcodeWriting,stream,",");
    stream<<") ## Canonical order of elements in parameter vector"<<endl;
    stream<<"GP$paramnbr <- length(GP$ParameterNames) ## fitted or not"<<std::endl;
    if(PopSizeScaleForTimebool) stream<<"GP$timeScale <- \"popSize\" "<<std::endl;
    else stream<<"GP$timeScale <- \"mutationRate\" "<<std::endl;
    stream<<"GP$graphicsFormat<-\""<<graphicsFormat<<"\""<<std::endl;

    provideQuotes(Rarguments);
    stream<<"GP$miscOptions<-c(";
    ostream_vector(Rarguments,stream,",");
    stream<<")"<<endl;
    stream<<"GP$verbosity <- 1 ## display of info about GCV"<<endl;
    /** info on statistical model**/
    if(IBDbool) stream<<"GP$DemographicModel <- \"IBD\" "<<std::endl; // actual DemographicModel string is PlanarIBD or similar...
    else stream<<"GP$DemographicModel <- \""<<DemographicModel<<"\" "<<std::endl;
    if (subsetPointls.size()>0) stream<<"GP$subsetRows <- ("<<subsetPointls<<") ## selection of rows in pointls"<<std::endl;
    else stream<<"GP$subsetRows <- NULL ## no selection of rows in pointls"<<std::endl;
    provideQuotes(KrigSpace);
    stream<<"GP$FONKgNames <- c(";  // to declare Kriging with non-canonical variables
    ostream_vector(KrigSpace,stream,",");
    stream<<")"<<endl;
    CopyForRcodeWriting=KrigScale; //  for stl::vectors, '=' has the nice copy constructor meaning
    provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
    stream<<"GP$FONKgScale <- c("; // to declare eg log-transformations of variables
    ostream_vector(CopyForRcodeWriting,stream,",");
    stream<<")"<<endl;
    stream<<"names(GP$FONKgScale) <- GP$FONKgNames"<<endl;

    //
    CopyForRcodeWriting=BoundsDefs; //  for stl::vectors, '=' has the nice copy constructor meaning
    for (vector<string>::iterator it=CopyForRcodeWriting.begin();it!=CopyForRcodeWriting.end();it++) {
        if (cmp_nocase((*it),"nb")==0) {
            (*it)="latt2Ns2";
            /* cf standardization of BoundsDefs: Nb was purposely not converted..*/
        }
    }
    provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
    stream<<"GP$samplingSpace<-c(";  // to declare Kriging with non-canonical variables
    ostream_vector(CopyForRcodeWriting,stream,",");
    stream<<") ## ONLY for next points"<<endl;

    //
    CopyForRcodeWriting=samplingScale; //  for stl::vectors, '=' has the nice copy constructor meaning
    provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
    stream<<"GP$samplingScale<-c(";  // to declare Kriging with non-canonical variables
    ostream_vector(CopyForRcodeWriting,stream,",");
    stream<<") ## ONLY for next points"<<endl;
    stream<<"names(GP$samplingScale)<-GP$samplingSpace"<<endl;
    //
    CopyForRcodeWriting=extraScale; //  for stl::vectors, '=' has the nice copy constructor meaning
    provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
    stream<<"GP$extraScale<-c(";  // to declare Kriging with non-canonical variables
    for (unsigned int it=0;it<extraScale.size();it++) {
        stream<<extraNames[it]<<"="<<CopyForRcodeWriting[it];
        if(extraScale.size()>1 && it<(extraScale.size()-1)) stream<<",";
    }
    stream<<") ## "<<endl;
    //
    //stream<<"GP$NbrMultilocusLiks<-"<<1+int(asymptoBool)<<" ## number of multilocus likelihood estimators per row (generally one, but two is possible)"<<std::endl;
    //.... but is this used ?

    stream<<"GP$profile3passesBool<-TRUE ##  should become user option ?"<<std::endl;

    stream<<"GP$ICthresh<-1e8 ## default threshold for ignoreSmallEigen"<<std::endl;

    stream<<"## (!) plotRange can contain Nb rather than latt2Ns2"<<endl;
    writeCfloatmapAsRlist(plotRange,"GP$plotRange",stream);
    //
    writeCfloatmapAsRlist(graphicPars,"GP$graphicPars",stream);
    CopyForRcodeWriting=PlotOptions; //  for stl::vectors, '=' has the nice copy constructor meaning
    provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
    stream<<"GP$plotOptions<-c(";
    ostream_vector(CopyForRcodeWriting,stream,",");
    stream<<")"<<endl;
    if (oneDimCIvars.size()>0) {
        CopyForRcodeWriting=oneDimCIvars; //  for stl::vectors, '=' has the nice copy constructor meaning
        provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
        stream<<"GP$oneDimCIvars<-c(";
        ostream_vector(CopyForRcodeWriting,stream,",");
        stream<<") ## (an empty c() value will be meaningful for nextBounts)"<<endl;
    } else { //currently provideQuotes on zero-size vector produces the string "" and in R, length("") = 1
        stream<<"GP$oneDimCIvars<-c()"<<endl;
    }
    if (spec1DProfiles.size()>0) {
        CopyForRcodeWriting=spec1DProfiles; //  for stl::vectors, '=' has the nice copy constructor meaning
        provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
        stream<<"GP$spec1DProfiles<-c(";
        ostream_vector(CopyForRcodeWriting,stream,",");
        stream<<")"<<endl;
    } else { //currently provideQuotes on zero-size vector produces the string "" and in R, length("") = 1
        stream<<"GP$spec1DProfiles<-c()"<<endl;
    }
    if (spec2DProfiles.size()>0) {
        CopyForRcodeWriting=spec2DProfiles; //  for stl::vectors, '=' has the nice copy constructor meaning
        provideQuotes(CopyForRcodeWriting);//attention ca ajoute des guillements et les test ne marchent plus...
        stream<<"GP$spec2DProfiles<-c(";
        ostream_vector(CopyForRcodeWriting,stream,",");
        stream<<")"<<endl;
    } else { //currently provideQuotes on zero-size vector produces the string "" and in R, length("") = 1
        stream<<"GP$spec2DProfiles<-c()"<<endl;
    }
    if (IBDbool) {
    //        stream<<"GV$S2factor<-"<<geoBinWidth*geoBinWidth<<" ##  the correction from condS2 in lattice unit to user's units"<<std::endl;
           stream<<"GP$S2factor<-1 ##  other values add more confusion that anything else"<<std::endl;
        if (D1IBDbool) {
            stream<<"GP$D2IBDbool<-FALSE;GP$D1IBDbool<-TRUE "<<std::endl;
            stream<<"GP$Nbfactor<-"<<geoBinWidth<<" ##  correction from 2NmS2 to 2Ds2=Nb"<<std::endl;
            if (geoUnit.size()>0)
               stream<<"GP$GeoUnit<-\""<<geoUnit<<"\" ## string that gives the Nb unit in plot"<<std::endl;
            else {
               if (geoBinWidth==1) {
                  stream<<"GP$GeoUnit<-\"lattice units\" ##  Migraine has not found a non-trivial bin width and user has not provided the unit"<<std::endl;
               } else {
                  stream<<"GP$GeoUnit<-\"\" ## Migraine has found a non-trivial bin width but user has not provided the unit"<<std::endl;
               }
            }
            stream<<"GP$mincondS2<-1"<<std::endl;
        } else { //D2IBD
            stream<<"GP$D2IBDbool<-TRUE;GP$D1IBDbool<-FALSE "<<std::endl;
            stream<<"GP$Nbfactor<-pi ##  from 2[N=D]m condS2 (where N=D in user's units) to 2piDs2 in user units "<<std::endl;
            stream<<"GP$GeoUnit<-\"\" ## string that would gives the Nb unit in 1D IBD plot"<<std::endl;
            stream<<"GP$mincondS2<-1/2"<<std::endl;
        }
        stream<<"if ( ! is.null(GP$plotRange[[\"Nb\"]])) GP$plotRange[[\"latt2Ns2\"]]<-GP$plotRange[[\"Nb\"]]/GP$Nbfactor"<<std::endl;
    } else { // not IBD
       stream<<"GP$D2IBDbool<-FALSE;GP$D1IBDbool<-FALSE "<<std::endl;
       stream<<"GP$GeoUnit<-\"\" ## string that would gives the Nb unit in 1D IBD plot"<<std::endl;
    }
    if (KrigLoci.size()>0) {
      stream<<"GP$fittedLoci <- c(";
      ostream_vector(KrigLoci,stream,",");
      stream<<")"<<endl;
    } else stream<<"GP$fittedLoci <- NULL"<<std::endl;
    stream<<"## Booleans: "<<std::endl;
    //stream<<"GP$useC<-TRUE ## no choice, should become removable"<<std::endl;
    /** selection of points**/
    //iter is the value given  as argument to Rparams.fill and thus starts from 1 for first iteration
    stream<<"GP$LikStatistic<-\""<<sequenceStat[std::min(iter,int(sequenceStat.size())-1)]<<"\" ## PAC/IS..."<<std::endl;	// CBR modified 15.02.2013, was min(iter-1,...)
    stream<<"GP$designRetain<-"<<designRetain<<" ## fraction of points in the Top retained for Kriging, removing occurrences of close points "<<std::endl;
    stream<<"GP$GCVdesignRetain<-"<<GCVdesignRetain<<" ## fraction of points in the Top retained for GCV, removing occurrences of close points "<<std::endl;
    //stream<<"# GP$maxDoubles<-"<<maxKrigedReplicatedPoints<<" ## seems obsolete"<<std::endl; //RL->FR 092016 : a enlever définitivement
    stream<<"GP$GCVptnbr<-"<<GCVptnbr<<" ## nbr of optimization steps for GCV"<<std::endl;
    stream<<"GP$GCVoptimSteps<-"<<GCVoptimSteps<<" ## nbr of optimization steps for GCV"<<std::endl;
    /** other Kriging parameters**/
    stream<<"GP$covFamily<-\""<<krigingCov<<"\" ## !!!!! covariance function for kriging"<<std::endl;
    stream<<"GP$krigmax<-"<<krigmax<<std::endl;
    stream<<"GP$kriglength<-"<<kriglength<<"## will be overwritten if krigmax<0"<<std::endl;
    stream<<"GP$krigoverlap<-"<<krigoverlap<<"## will be overwritten if krigmax<0"<<std::endl;
    stream<<"GP$minKrigPtNbr<-"<<minKrigPtNbr<<" ## nbr of optimization steps for GCV"<<std::endl;
     if (maxKrigPtNbr==-1)
        stream<<"GP$maxKrigPtNbr<-FALSE ## max pts for Kriging"<<std::endl;
    else
        stream<<"GP$maxKrigPtNbr<-"<<maxKrigPtNbr<<" ## max pts for Kriging"<<std::endl;
    stream<<"## larger bounds below => fewer boundary gcv problems but not necessarily meaningful"<<std::endl;
    stream<<"GP$GCVlowerFactor<-20 ## 2 *w*as suggested by Martin & Simpson Am. Inst. Aeron. Astron. 43: 853- (2005)"<<std::endl; // NB: maybe OK for large nu, not for nu=0.5; from GLMM experiments
    stream<<"GP$GCVupperFactor<-5 ## as suggested by ibidem"<<std::endl;
    stream<<"if (GP$covFamily %in% c(\"Matern\", \"GneitingMatern\")) {"<<std::endl;
    stream<<"\tGP$minSmoothness<-"<<minSmoothness<<std::endl;
    stream<<"\tGP$maxSmoothness<-"<<maxSmoothness<<std::endl;
    stream<<"}"<<std::endl;
    if (CovFnParam.size()>0) {
      stream<<"GP$CovFnParamInSettingsBool<-TRUE"<<std::endl;
      stream<<"GP$CovFnParam<-c(";
      ostream_vector(CovFnParam,stream,",");
      stream<<")"<<std::endl;
    } else stream<<"GP$CovFnParam <- numeric(0);GP$CovFnParamInSettingsBool<-FALSE"<<std::endl;

    stream<<"GP$metarange<-3 ## controls covariance range, except if CovFnParam vector is estimated by GCV "<<std::endl;
    /** hull parameters **/
    stream<<"GP$hullExpandFactor<-"<<hullExpandFactor<<std::endl;
    stream<<"GP$profileEdges<-c() ## will accumulate masking points encountered during profile computations"<<std::endl;
    stream<<"GP$gridstepsNbr<-"<<sequenceGridSteps[std::min(iter,int(sequenceGridSteps.size())-1)]<<std::endl;	// CBR modified 15.02.2013
    stream<<"GP$LRTlist<-list()"<<std::endl;
    /** CI, LRT, nextpoints parameters **/
    stream<<"GP$CIlevel<-"<<CILevel<<std::endl;
    stream<<"GP$NextBoundsLevel<-"<<NextBoundsLevel<<std::endl;
    stream<<"GP$nextBounds<-\""<<nextBounds<<"\""<<std::endl;
    stream<<"GP$nextPointNumber<-"<<nextPointNumber<<std::endl;
    stream<<"GP$ptSamplingSeed<-"<<_ptSamplingSeed<<" ## random seed for sampling new points"<<std::endl;
    stream<<"GP$parDigits<-"<<parDigits<<" ## digits of parameter coordinates in files"<<std::endl;
    /** one ugly numerical setting for optim() **/
    // a unit change in par/pascale should produce a unit change in fn()/fnscale
    stream<<"GP$scalefactor<-"<<scalefactor<<" ## for (fnscale,parscale in optim())"<<std::endl;

    if (ParMaxima.names.size()>0) ParMaxima.write(stream,"GP$ParMaxima <- list(");

    if (ParMinima.names.size()>0) ParMinima.write(stream,"GP$ParMinima <- list(");



    stream<<"\nwrite(\"Migraine "<<version<<"\",file=GP$cleanResu)"<<std::endl;
    stream<<"\nwrite(paste(paste(packageDescription(\"blackbox\",fields = c(\"Package\", \"Version\")), collapse=\", version \"),\"loaded\"),file=GP$cleanResu)"<<std::endl;
    stream<<"write(paste(\"R code run on \",date()),file=GP$cleanResu)"<<std::endl;
    stream<<"write(\"\",file=GP$cleanResu)"<<std::endl;
    stream<<"write(\"Data file: "<<glob_gp_filename<<"\",file=GP$cleanResu)"<<std::endl;
    stream<<"write(\"Settings file: "<<settingsfilename<<"\",file=GP$cleanResu)"<<std::endl;
    stream<<"write(\"\",file=GP$cleanResu)"<<std::endl;

    return(0);
} // end CRParameters::writeR()




double TESTformatAsInFile(float f) {
    double d,dd,ddd;
    float f2,ff= (float) 0.12345678,fff= (float) 0.12345679;
    // if one follows suggestions on the web, ther it suffices to compute
    stringstream stst;string atuer;
//                            stst.setf(ios::fmtflags(0),ios::floatfield);     // deja le default
    stst<<f;
    stst>>d;  // write up to the precision expected
    stst.str("");stst.clear();
    stst<<d;   // one can try <<f also
    stst>>f2;
    // now the subtle result, f-d is Ok but f-f2 isnot, and numbers are added to the float (beyond its meaningful digits),
    //            not to the double (because the implicit O's numbers are meaningful!)
    cout<<setprecision(10)<<f<<" "<<f2<<" "<<d<<" "<<f-f2<<" "<<f-d<<" ";getchar();
    // the strinking result is that the following doubles are equal while the float aren't !!
    stst.str("");stst.clear();
    stst<<f;
    stst>>f2;
    stst.str("");stst.clear();
    stst<<ff;
    stst>>dd;  // write up to the precision expected
    stst.str("");stst.clear();
    stst<<fff;
    stst>>ddd;  // write up to the precision expected
    cout<<(f==f2)<<" "<<(ff==fff)<<" "<<(dd==ddd);getchar(); // false, false, true....
    //so we must use doubles !!!
    return(d);
}

double formatAsInFile(double d,int digits) { // see TESTformatAsInFile() for background
    stringstream stst;
    double dd;
    stst<<setprecision(digits)<<d; // precision should match the firstprec argument of ofstream_vector_mll
    stst>>dd;  // write up to the precision expected
    //cout<<setprecision(10)<<d<<" "<<dd;getchar(); //
return(dd);
}



void findandreplace( std::string & source, const std::string & find, const std::string & replace ) {
//http://www.kbcafe.com/articles/CPP.Tricks.pdf.
int j; // et ca plante en unsigned...
for (;( j = (int) source.find( find ) ) != (int) std::string::npos;) {//CM->RL 23.01.14
	source.replace( j, find.length(), replace );
	}
}

vector<vector<double> > read_points_from_C(string filename) { // reads the parameters in a pointls file
    ifstream pointsfromC;
    pointsfromC.open(filename.c_str());
    vector<double> temp_vec;
    float temp;
    string buf;
    stringstream stst;

   	if ( ! pointsfromC.is_open()) {
   	    cout<<"Could not open file "<<filename<<". I exit.";
   	    cerr<<"Could not open file "<<filename<<". I exit.";
   	    if(cinGetOnError) cin.get();
   	    exit(-1);
    } // ELSE
    vector<vector<double> > Uniquepoints(0);
    while ( ! pointsfromC.eof()) {
       temp_vec.resize(0);
	   getline(pointsfromC,buf);
	   stst<<buf;
       while (stst>>temp && temp_vec.size()<ParameterNumber) {
           temp_vec.push_back(temp);
       }
       if (temp_vec.size()==ParameterNumber) Uniquepoints.push_back(temp_vec);
       stst.str("");
       stst.clear(); //clears the eof bit (among others)
    }
    pointsfromC.close();
    // automatically sort and keep only unique values (that's the way DesignPoint is generated too)
    set<vector<double> > s( Uniquepoints.begin(), Uniquepoints.end() );
    Uniquepoints.assign( s.begin(), s.end() );
return Uniquepoints;
}; /*fin read_points_from_C */


string krig() {
//calls migraine.R and will run the batch R file which writes into a file in Migraine's working directory
//only the .bat file knows the R directory (note it's Progra~1, not Program Files... snif
std::string callstring,batchRname,fdir;
std::string backslash="\\",forwslash="/";
std::string Rmlsname;
/**
http://pubs.opengroup.org/onlinepubs/009695399/functions/getcwd.html:
char *getcwd(char *buf, size_t size);

"If buf is a null pointer, the behavior of getcwd() is unspecified."

"On some implementations, if buf is a null pointer, getcwd() may obtain size bytes
of memory using malloc(). In this case, the pointer returned by getcwd()
may be used as the argument in a subsequent call to free(). Invoking getcwd()
with buf as a null pointer is not recommended in [IEEE Std = POSIX ] conforming applications."

see also http://stackoverflow.com/questions/2868680/what-is-a-cross-platform-way-to-get-the-current-directory
**/
    {
        char* dir=NULL; /** see above discussion. g++ seems to handle automatic reallocation**/
        // 'subtle' distinction between the 'dir' argument (*dir is where the cwd goes) and the returne value (the pointer...) :
        dir=getcwd(dir,PATH_MAX); //gets the current working directory =>char[]
        if (dir != NULL) fdir=dir; //char[]=> string
        else {
            std::cerr<<"(!) From krig(): getcwd failed. I exit."<<std::endl;
            if (cinGetOnError) cin.get();
            exit(-1);
        }
        free(dir);
    }
//converts all \ to / : all progs understand / (including windows) and R requires /
//cout<<fdir;getchar();
findandreplace(fdir,backslash,forwslash);

//cout<<fdir;getchar();
    {std::stringstream stst(std::stringstream::in | std::stringstream::out);
        stst<<"migraine_"<<_mls<<".R";
        stst>>Rmlsname;
//	sprintf(Rmlsname,"migraine_%d.R",_mls);
    }
	batchRname=Rmlsname; //name of batch R file e.g. migraine.R
	if (Rtermfullpath.size()<1) //i.e. no valid path in settings.txt
	    // --redirOut serves as a flag that stdout (but not stderr) is redirected (>) to R_out_...txt file at the OS level
	    // --args makes --redirOut be ignored by R at the C level. This prevents a warning from the R executable.
        callstring="R --vanilla --args --redirOut < \"";
	else {
		callstring=Rtermfullpath;
		callstring+=" --vanilla  --args --redirOut < \"";
        //RL->FR : RL 022017 : This call to R with --vanilla does not always work with local installation of R libraries on some Linux PC / clusters
        // using R CMD BATCH [options] my_script.R [outfile] # on Linux (GENOTOUL cluster) worked fin for me, but stdout and stderr are redirected to R_out_...txt file
        //an equivalent for windows may be "C:\Program Files\R\R-2.13.1\bin\R.exe" CMD BATCH --vanilla --slave "c:\my projects\my_script.R"
        //but the behavior of "CMD BATCH --vanilla --slave" has to be checked
	}
#ifdef WIN32
    	callstring+=fdir;
    	callstring+="/";
    	callstring+=batchRname;
    	callstring+= "\" > \"";
    	callstring+=fdir;
    	callstring+="/R_out_";
        {std::stringstream stst(std::stringstream::in | std::stringstream::out);
            stst<<_mls;
            stst>>Rmlsname;
    //    	sprintf(Rmlsname,"%d",_mls); // only for the name of the message file for R
        }
    	callstring+=Rmlsname;
    	callstring+=".txt\"";
#else
    	callstring+=batchRname;
    	callstring+= "\" > ";
    	callstring+="R_out_";
        {stringstream stst(stringstream::in | stringstream::out);
            stst<<_mls;
            stst>>Rmlsname;
    //    	sprintf(Rmlsname,"%d",_mls); // only for the name of the message file for R
        }
    	callstring+=Rmlsname;
    	callstring+=".txt";
#endif
std::cout<<std::endl<<callstring<<std::endl;//getchar();
/*With such arguments callR. bat should read input file <Migraine dir>/<R batch file>
and write output file <Migraine dir>/output_mls.txt*/
//:: The following is common to both versions
	if (0!=system(callstring.c_str())) {
		std::cout<<"\nCall to R apparently didn't work. I will exit.";if (cinGetOnError) cin.get();return("Call to R apparently didn't work");
		// FR->FR it would be nice to be able to test R call NOT through the system return call
	};
	return("normal exit");
}

int reset_PAC_IS_seq_iter(int iter) {
//pointer to member function PAC/IS: cf Primer+ p. 1071
	if (cmp_nocase(sequenceStat[iter],"PAC")==0) {
		if (sequenceStat.size()>1) std::cout<<"Iteration "<<iter+1<<": PAC Likelihood"<<std::endl<<std::flush;
		calculQused=false;
		pacbool=true; // sert pour la constante multinomiale au moins
		ChangefnPtr=&CSample::PACpi;
	} else if (cmp_nocase(sequenceStat[iter],"PACanc")==0) {
		if (sequenceStat.size()>1) std::cout<<"Iteration "<<iter+1<<": Likelihood (IS) + PAC Likelihood only for the ancestral population"<<std::endl<<std::flush;
		pacancbool=true; // sert pour la constante multinomiale au moins
		ChangefnPtr=&CSample::Change;
	} else if (cmp_nocase(sequenceStat[iter],"PACanc2id")==0) {
		if (sequenceStat.size()>1) {
			std::cout<<"Iteration "<<iter+1<<": Likelihood (IS) \n+ PAC Likelihood only for the ancestral population"<<std::endl;
			std::cout<<"+ analytical likelihood computation for the ancestral pair of genes "<<std::endl<<std::flush;
		}
		pacancbool=true;
		calculQused=true;
		ChangefnPtr=&CSample::Change;
	} else if (cmp_nocase(sequenceStat[iter],"PAC2id")==0) {
		if( ! IBDbool) std::cerr<<"(!) PAC2id option is in a test phase."<<std::endl;
		if (sequenceStat.size()>1) std::cout<<"Iteration "<<iter+1<<": PAC Likelihood + likelihood for the ancestral pair of genes"<<std::endl<<std::flush;
		pacbool=true; // sert pour la constante multinomiale au moins
        calculQused=true;
		ChangefnPtr=&CSample::PACpi;
	} else if (cmp_nocase(sequenceStat[iter],"ISstrict")==0) {
		if (sequenceStat.size()>1) std::cout<<"Iteration "<<iter+1<<": Likelihood (IS, no direct analytical computation of identity in state) "<<std::endl<<std::flush;
		calculQused=false;
		pacbool=false; // sert pour la constante multinomiale au moins
		pacancbool=false;
		ChangefnPtr=&CSample::Change;
	} else if (cmp_nocase(sequenceStat[iter],"IS")==0) {  // generic I
		if (sequenceStat.size()>1) std::cout<<"Iteration "<<iter+1<<": Likelihood (IS + analytical computation of identity in state)"<<std::endl<<std::flush;
		//if(!IBDbool) std::cerr<<"(!) IS + analytical computation of identity in state is in a test phase for non IBD models."<<std::endl;
		calculQused=true;
		pacbool=false; // sert pour la constante multinomiale au moins
		pacancbool=false;
		ChangefnPtr=&CSample::Change;
	} else {
		std::cout<<"Iteration "<<iter+1<<": Statistic or StatisticSequence specification is " << sequenceStat[iter] << " and does not correspond to any implemented option)"<<std::endl<<std::flush;
		std::cout << "Statistic or StatisticSequence is set to ISstrict." << endl;
		if (cinGetOnError) cin.get();
		sequenceStat[iter]="ISstrict";
		calculQused=false;
		pacbool=false; // sert pour la constante multinomiale au moins
		pacancbool=false;
		ChangefnPtr=&CSample::Change;
	}
return 0;
}

int CSampling::redim(const int parnbr,const int ptnbr){
    DesignPoint.resize(ptnbr);
    for(std::vector<std::vector<double> >::iterator ii=DesignPoint.begin();ii<DesignPoint.end();ii++){
        (*ii).resize(parnbr);
	}
return(0);
};

int CSampling::hypercube_sampling(const float* lb,const float* ub,int ptnbr) { // returns canonical vectors
float blockNbr,dumi,tt,xi;
double ran;
int q;
int count,expon,c,replaced;
float a, b;
std::vector<float> h(ParameterNumber);
std::vector<std::vector<float> > dx(ParameterNumber);
//cout<<"debut CSampling::hypercube_sampling";getchar();
/*this.*/redim(ParameterNumber,ptnbr);

blockNbr= (float) ptnbr;
dumi=(float) ParameterNumber;
tt=(float) pow((double)blockNbr,(double)1.0/dumi); //g++ compiler: ISO C++ says...
blockNbr=floor(tt);
q= (int) blockNbr;
    for(std::vector<std::vector<float> >::iterator ii=dx.begin();ii<dx.end();ii++) (*ii).resize(q+1);
    for(unsigned int i=0;i<ParameterNumber;i++) {
		if (samplingScale.size()>i && cmp_nocase(samplingScale[i],"LogScale")==0) h[i]=log(ub[i]/lb[i])/(blockNbr);
			else h[i]=(ub[i]-lb[i])/(blockNbr); /* divise la gamme par le nb de points par dimension */
	}
    for(unsigned int i=0;i<ParameterNumber;i++){
        if (h[i]< -0.00000001) {
            cerr<<"(!) Lower bound higher than upper bound for parameter "<<i+1<<endl;
            cerr<<"    As this is likely a typing error in the settings file, I exit."<<endl;
            if(cinGetOnError) cin.get();
            exit(-1);
        }
    }
    for(unsigned int i=0;i<ParameterNumber;i++){
        if (samplingScale.size()>i && cmp_nocase(samplingScale[i],"LogScale")==0) xi=log(lb[i]);
        else xi= lb[i];
        for(int j=0;j<q+1;j++){
        	dx[i][j]=xi; /*genere tableau avec valeurs [`a] intervalles r['e]guliers */
        	xi+=h[i];
        }
    }

    for(unsigned int i=0;i<ParameterNumber;i++) {// filling by columns...
    	count=0;
    	expon= (int) pow(blockNbr, (float) i );
    	for (int l=0;l<expon;l++){
    		  for(int k=1;k<q+1;k++){
    		    c= (int) pow(blockNbr,(dumi-i-1));
    			    for(int j=0;j<c;j++){
    			      a=dx[i][k-1];
    			      b=dx[i][k];
    			      ran= (double) a + ( b - a ) * aleah();
//if (i=01 && ran>log(10e+10)) {cout<<":"<<ran;getchar();}
//    			      DesignPoint[count][i]=formatAsInFile(ran);
    			      DesignPoint[count][i]=ran;
//if (DesignPoint[count][i]>10e+10) {cout<<"!!"<<DesignPoint[count][i]<<" "<<ran;getchar();}
    			      count+=1;
    			    }
    		  }
    	}
    }


/*le complement sont des points completement al['e]atoires dans l'espace*/
    expon = (int) pow(blockNbr,dumi);
    c=ptnbr-expon;
    if(c>0){
    	count=expon;
    	for(int j=0;j<c;j++){
    		  for(unsigned int i=0;i<ParameterNumber;i++){
//code pas optimis['e] ici mais bof
                if (samplingScale.size()>i && cmp_nocase(samplingScale[i],"LogScale")==0)
                   ran = log(lb[i])+log(ub[i]/lb[i])*aleah();
                else ran = lb[i]+(ub[i]-lb[i])*aleah();
//		        DesignPoint[count][i]=formatAsInFile(ran);
		        DesignPoint[count][i]=ran;
    		  }
    	count+=1;
    	}
    }


    /** back to canonical SCALE **/
    for(unsigned int i=0;i<ParameterNumber;i++) {
        if (samplingScale.size()>i && cmp_nocase(samplingScale[i],"LogScale")==0)
            for(std::vector<std::vector<double> >::iterator ii=DesignPoint.begin();ii<DesignPoint.end();ii++) {
//cout<<"< "<<((*ii)[i]);
            	((*ii)[i])=exp((*ii)[i]);
//cout<<"> "<<((*ii)[i]);
            }
   }
   /** now replace some random values by those in LRTmin, LRTmax
    LRTmin LRTmax are in 'user's' scpace and scale that is
    * in BoundsDefs SPACE so this must come before conversion back to canonical space
    * in canonical SCALE (ie not log) so this must come after conversion back to canonical scale
   **/
   //FR->FR maybe non ideal to do that several times within sampling loop
    for(unsigned int i=0;i<ParameterNumber;i++){
       if (ParameterNumber==1)   //FR->FR dans ces 3 lignes ParameterNumber serait a remplacer par un fitted param nbr ?
            replaced=1; /*un seul point sinon plein de valeurs identiques*/
       //int(log({1, 2, 3, 4, 5, 32, 125, 512, 1000, 5000}))={ 0., 0., 1., 1., 1., 3., 4., 6., 6., 8.}
       else replaced=max(0,min(int(DesignPoint.size())-1,int(log(DesignPoint.size()))-int(ParameterNumber)));
       if (! std::isnan(LRTmin[i])) {
          for (int it=0;it<replaced;it++) DesignPoint[aleah.randInt(DesignPoint.size()-1)][i]=max(LRTmin[i]-float(0.00001),(LRTmin[i]+lb[i])/2); // prevents a small positive value from becoming negative
       }
       if (! std::isnan(LRTmax[i])) { // typically (if only one LRT value for given variable) LRTmin=LRTmax
          for (int it=0;it<replaced;it++) DesignPoint[aleah.randInt(DesignPoint.size()-1)][i]=min(LRTmax[i]+float(0.00001),(LRTmax[i]+ub[i])/2); // and more genrally to equal or exceed lb,ub
       }
    }

    /** back to canonical SPACE **/
// maintenant on passe en echelle g si on etait en echelle sigma2
    if (BoundsDefs.size()>2 && cmp_nocase(BoundsDefs[2],"condS2")==0) {
        float S2;
        for(std::vector<std::vector<double> >::iterator ii=DesignPoint.begin();ii<DesignPoint.end();ii++) {
        	S2= (float) ((*ii)[2]);  // in lattice units, to have known bounds !
//cout<<"[ "<<((*ii)[2]);
        	((*ii)[2])= (double) groot( (float) (*ii)[2]);// computes g from sigma2|cond... not from 2sigma2...
//cout<<" ]"<<((*ii)[2]);
        	if(std::isnan(((*ii)[2]))) {
        	    cerr<<"(!) From hypercube_sampling: groot is NaN"<<endl;
        	    cerr<<"    Dispersed S2 was "<<S2<<endl;
        	    if(cinGetOnError) cin.get();
        	    exit(-1);
            }
        }

    }
// maintenant on passe en echelle 2Nm si on etait en echelle Nb (doit venir apr[`e]s la reconversion en g
    if (BoundsDefs.size()>1 && (cmp_nocase(BoundsDefs[1],"latt2Ns2")==0 || cmp_nocase(BoundsDefs[1],"Nb")==0 ) ) {
        float lattcondS2;
        for(std::vector<std::vector<double> >::iterator ii=DesignPoint.begin();ii<DesignPoint.end();ii++) {
        	lattcondS2= (float) ((*ii)[2]);  //starts from g value
        	// lattcondS2 from g:
        	if (D1IBDbool) {
              	lattcondS2=( (float) 1. + lattcondS2)/(( (float) 1. - lattcondS2)*( (float) 1. - lattcondS2));
              	if(cmp_nocase(BoundsDefs[1],"Nb")==0) {
              	    lattcondS2*= (float) geoBinWidth; // no longer really lattcondS2! Nb vs latt2Ns2 makes a difference here !
              	}
            } else if (D2IBDbool)
              	lattcondS2=( (float) 1. + lattcondS2)/(( (float) 1. - lattcondS2)*( (float) 1. - lattcondS2)*( (float) 2. - lattcondS2));
            // at this point lattcondS2 is really lattcondS2
        	(*ii)[1]/=lattcondS2;  //2'N'm is now 2Dm where D is the (density=number) per lattice step, not in user's scale
        	if (D2IBDbool) (*ii)[1]/=3.1415926535;
        }
    }


////// // IBM's xlC compiler seems buggy at this point
//////    stable_sort(DesignPoint.begin(),DesignPoint.end());  // replaced by call to set dans samplingloop

    /** format as in text file: only when its in canonical format ! **/
    for(std::vector<std::vector<double> >::iterator ii=DesignPoint.begin();ii!=DesignPoint.end();ii++)
        for(std::vector<double>::iterator jj=ii->begin();jj!=ii->end();jj++) (*jj)=formatAsInFile(*jj,parDigits);

return 0;
}; /*fin CSampling::hypercube_sampling */


int CSampling::samplingloop(const float* lb,const float* ub) {
    int missing=PointNumber;
    int it=0;

    // counts effectively variable parameters
    boundsVariableParNbr=0;
    for (unsigned int ii=0;ii<ParameterNumber;ii++) {
        if(abs(ub[ii]-lb[ii])>1e-06) boundsVariableParNbr+=1;
    }

    while (missing>0) {
           CSampling locCS;
           locCS.hypercube_sampling(lb,ub,missing);  // locCS.DesignPoint in canonical, "as in file" form
           DesignPoint.insert(DesignPoint.end(),locCS.DesignPoint.begin(),locCS.DesignPoint.end());
           if ( ( (int) DesignPoint.size() ) != PointNumber ) {
               cerr<<"(!) DesignPoint.size()!=PointNumber in CSampling::samplingloop()."<<endl;
               cerr<<"Debugging information: "<<endl;
               cerr<<DesignPoint.size()<<" "<<PointNumber;
               cerr<<"   Contact the authors. I exit";
               if (cinGetOnError) cin.get();
               exit(-1);
           }
           // if no variable parameters, then we have PointNumber replicates of the same point, as probably intented.
           // Otherwise, we want all points to be different. Then:
           if(boundsVariableParNbr>0) {
               // automatically sort and keep only unique values, with DesignPoint in canonical, "as in file" form
               set<vector<double> > s( DesignPoint.begin(), DesignPoint.end() );
               DesignPoint.assign( s.begin(), s.end() );
           }
           missing=PointNumber - (int) DesignPoint.size();
           it++;
           if (it>5) {
               cerr<<"(!) Suspicious loop in CSampling::samplingloop(): point sampling algorithm generates the same points again and again."<<endl;
               cerr<<"   Check settings. I exit";
               if (cinGetOnError) cin.get();
               exit(-1);
           }
     } // at this step the points must be sorted and in canonical form


    /** Computes pointsVariableParNbr as the observed number of variable parameters **/
    pointsVariableParNbr=0;
    for (unsigned int ii=0;ii<ParameterNumber;ii++) {
        vector<double> tmp(0);
        for (vector<vector<double> >::iterator it=DesignPoint.begin();it!=DesignPoint.end();it++) {
            tmp.push_back((*it)[ii]);
        }
        stable_sort(tmp.begin(),tmp.end()); // this only serves to count the variable parameters
        if(abs(tmp.back()-tmp.front())>1e-06) pointsVariableParNbr+=1;
    }
return(0);
} // end def sampling_loop

int CSampling::firstfill(int mls,string write,const float* lb,const float* ub) {
    DesignPoint.resize(0);
    if (cmp_nocase(write,"ReadPoints")==0) { //infor from sequenceWrite, not from BoundsSource
       stringstream stst(stringstream::in|stringstream::out);
       string ptlname;
       stst<<"nextpoints_"<<mls<<".txt";
       stst>>ptlname;
       read_points_from_R(ptlname);
    } else {
       samplingloop(lb,ub);
    }
return(0);
};

int CSampling::read_points_from_R(string filename) { // read and returns canonical vectors
    ifstream pointsfromR;
    pointsfromR.open(filename.c_str());
    vector<double> temp_vec;
    float temp;
    string buf;
    stringstream stst;

   	if ( ! pointsfromR.is_open()) {
   	    cout<<"Could not open file "<<filename<<". I exit.";
   	    cerr<<"Could not open file "<<filename<<". I exit.";
   	    if(cinGetOnError) cin.get();
   	    exit(-1);
    } // ELSE
    DesignPoint.resize(0);
    while ( ! pointsfromR.eof()) {
       temp_vec.resize(0);
	   getline(pointsfromR,buf);
	   stst<<buf;
       while (stst>>temp) {//true as long as there is a value found
           temp_vec.push_back(temp);
       }
       // a lot of trouble because R as added an empty line at the end of the file...
       if (!(temp_vec.size()==ParameterNumber || temp_vec.size()==0)) {
          cout<<"(!) From CSampling::read_points_from_R: invalid point in "<<filename<<":"<<endl;
          ostream_vector(temp_vec,cout);
          cout<<"\n I exit.";
   	      cerr<<"(!) From CSampling::read_points_from_R: invalid point in "<<filename<<":"<<endl;
          ostream_vector(temp_vec,cerr);
          cerr<<"\n I exit.";
   	      if(cinGetOnError) cin.get();
   	      exit(-1);
       }
       if (temp_vec.size()==ParameterNumber) DesignPoint.push_back(temp_vec);
       stst.str("");
       stst.clear(); //clears the eof bit (among others)
    }
    pointsfromR.close();
    if(DesignPoint.size()!=Rparams.getNextPtNbr()) {
   	    cout<<"(!) From CSampling::read_points_from_R: unexpected size of "
   	       <<filename<<" ("<<DesignPoint.size()<<" read vs. "<<Rparams.getNextPtNbr()<<" expected). I exit.";
   	    cerr<<"(!) From CSampling::read_points_from_R: unexpected size of "
   	       <<filename<<" ("<<DesignPoint.size()<<" read vs. "<<Rparams.getNextPtNbr()<<" expected). I exit.";
   	    if(cinGetOnError) cin.get();
   	    exit(-1);
    }
// should already be sorted, but let's play safe
    stable_sort(DesignPoint.begin(),DesignPoint.end());
    // counts effectively variable parameters
    pointsVariableParNbr=0;
    for (unsigned int ii=0;ii<ParameterNumber;ii++) {
        vector<double> tmp(0);
        for (vector<vector<double> >::iterator it=DesignPoint.begin();it!=DesignPoint.end();it++) {
            tmp.push_back((*it)[ii]);
        }
        stable_sort(tmp.begin(),tmp.end());
        if(abs(tmp.back()-tmp.front())>1e-06) pointsVariableParNbr+=1;
    }
return 0;
}; /*fin CSampling::read_points_from_R */

int CSampling::analyse(const int& mls) {
//ostream_vector(boundsSource,cout);getchar();
//std::cout<<"debut analyse\n pointmin= "<<pointmin;getchar();
    pointmin=max(_pointmin,1);pointmax=min(_pointmax,PointNumber);
//    pointmin=max(_pointmin,1);pointmax=min((unsigned int)(_pointmax),PointNumber);
    // for single point _pointmax=_pointmin
	if (pointmin>pointmax) {cerr<<"\a (!) Problem: pointmin>pointmax.\nI exit.";if (cinGetOnError) cin.get(); exit(-1);}
	if (pointmin>PointNumber) {cerr<<"\a (!) Problem: pointmin>PointNumber.\nI exit.";if (cinGetOnError) cin.get(); exit(-1);}
	if (pointmax>PointNumber) {cerr<<"\a (!) Problem: pointmax>PointNumber.\nI exit.";if (cinGetOnError) cin.get(); exit(-1);}
	//if (krigIterations>-1 || singlePointBool) {
        std::stringstream stst(std::stringstream::in|std::stringstream::out);
//		if (cmp_nocase(sequenceWrite[0],"Append")!=0 && ((pointmin!=1)||(pointmax!=PointNumber)))  { //FR01/2010 removed '_'pointmin/max
        //	CBR->FR: the above commented statement can now be considered as obsolete
        if (userPointMinOrMax) {
            stst<<"pointls_"<<mls<<"_"<<pointmin<<"_"<<pointmax<<".txt";
        } else {
            stst<<"pointls_"<<mls<<".txt";
        }
        stst>>ptlname;
//...et non une declaration ofstream ptls(...)	qui ets perdue quand on sort du bloc
	//} // else no pointls file
    std::string fromRname;
    std::string outputline,nextline;
    std::ifstream fromR;
    std::string atst;
    string boundsname;
    vector<vector<float> >boundsvals(ParameterNumber);
    vector<vector<int> >boundsindic(ParameterNumber);
    int at1,at2,replicatesForThisPoint;
    std::vector<float> mlee(ParameterNumber) //estimates of MLE by kriging
	                                  ,old_lb(ParameterNumber); //keeps old lower bound
    float emll/*,atfl*/,eNb,MLE,EdgeIndice;
    float *lbvar=new float[ParameterNumber];
    float *ubvar=new float[ParameterNumber];
    const float expfactor=( (float) 1. + sqrt( (float) 5.))/( (float) 2.);
    const float edgefactor= (float) 8.;
    bool boundsFromRBool=false; // indicateur local
    int everyNpoints;
// so under linux the clockare int's with maximum 2147483647 and the CLOKS_PER_SEC are 1000000.
// So EndTime cannot exceed 2147 seconds... snif.
//La solution c'est d'utiliser time() qui compte en secondes  (moins pr['e]cis, mais bof)
//clock_t _StartTime,_EndTime;sequenceWrite.size()
    time_t _StartTime=0,_EndTime=0;
    float _secs,_prevsecs=0;
    std::vector<float>mloglik(mloglikdim);
    _mls=mls;
    for (unsigned int ii=0;ii<ParameterNumber;ii++) {lbvar[ii]=lb[ii];ubvar[ii]=ub[ii];};
    int debutpt=std::max(1,pointmin);
    set_fittedpars(lbvar,ubvar);

    {std::stringstream stst(std::stringstream::in | std::stringstream::out);
            stst<<"output_"<<_mls<<".txt";
            stst>>fromRname;
//		sprintf(fromRname,"output_%d.txt",_mls);
    }


    if (!singlePointBool && !noRfileBool && fittedpars.size()>0) {
        if (cmp_nocase(sequenceWrite[0],"KrigOnly")!=0) {
            write_R_batch_file(lbvar,ubvar,Rparams);
            //writes migraine_<_mls>.R with appropriate lb,ub (before first iteration)
        }
    }
    if (sequenceWrite.size()==1 && cmp_nocase(sequenceWrite[0],"WriteROnly")==0) {
         goto quickexit; // has only written R batch file, no likelihood computation nor kriging; ad hoc files may have been created.
     }
    /// unless new replicates have been explicitly called for, we check that none are computed (protection against incorrect usage)
    if (cmp_nocase(sequenceWrite[0],"Append")==0 && (! allowNewReplicates) )  {
        /// need to read existing points and compare them to new putative points
        vector<vector<double> > points=read_points_from_C(ptlname); // initialized with unique points from pointls
        /// it's apparently necessary to 'reformat' de points despite the fact that they were read from the file... (DesigmPoint is already thus formatted, by construction)
        for(std::vector<std::vector<double> >::iterator ii=points.begin();ii!=points.end();ii++)
          for(std::vector<double>::iterator jj=ii->begin();jj!=ii->end();jj++) (*jj)=formatAsInFile(*jj,parDigits);
        points.insert(points.end(),DesignPoint.begin(),DesignPoint.end());
        set<vector<double> > s( points.begin(), points.end() );
        if (s.size()!= points.size()) { /// then there were shared values between pointls and new DesignPoint
            cerr<<"(!) 'writeSequence=Append...' appears to be used"<<endl;
            cerr<<"    to add replicates of already computed points."<<endl;
            cerr<<"    If this is really what you want to do,"<<endl;
            cerr<<"    then consider the 'addReplicates' keyword."<<endl;
            cerr<<"    But this is likely an error, where something like "<<endl;
            cerr<<"    'writeSequence=readPoints,Append' should be used."<<endl;
            cerr<<"    Check the documentation about iterative analyses. I exit";
            if (cinGetOnError) cin.get();
            exit(-1);
        }
    }





    if (__same_points_all_mls) aleah.seed(_ptSamplingSeed);


/**** Kriging and/OR likelihood computations ****/
//sequenceWrite *always* has at least one element
//si krigterations=1 ki=0, sequenceWrite[0] is used
//si krigterations=0 ki=0, sequenceWrite[0] is used
/*********************  LOOP OVER krigiterations *************************/
    for (unsigned int ki=0;ki<sequenceWrite.size();ki++) {
        /**** reset point numbers : nothing to do****/
        // this has been done either at initialization or at the end of the previous iteration
        /**** reset likelihood estimation settings ****/
	    // switches off kriging when there are no pointS to krig.
        args.nruns=sequenceNrunsPerPoint[std::min(int(ki),int(sequenceNrunsPerPoint.size()-1))];
        // le test suivant doit forc['e]ment etre vrai pour ki=0 sinon ChangefnPtr n'est pas valu'e
        // => One must have sequenceStat.size() >= 1
        if (sequenceStat.size()>ki) reset_PAC_IS_seq_iter(ki); //reset pointer according to ...[ki]
       // the || clause is evaluated only if ki=0
        /**** any likelihood to write to a file ? ****/
        if (ki>0 || cmp_nocase(sequenceWrite[0],"Append")==0 || cmp_nocase(sequenceWrite[0],"Over")==0 ) {
// excludes cases ki=0 AND (WriteROnly,WriteRKrig,KrigOnly,readBounds,readPoints...)
// calcule d'abord des vraisemblances
//cout<<ki<<" "<<sequenceWrite[0];cin.get();
        //	if (ki>0 && cmp_nocase(sequenceWrite[std::min(ki,int(sequenceWrite.size())-1)],"Append")==0) {
            if (cmp_nocase(sequenceWrite[std::min(int(ki),int(sequenceWrite.size()-1))],"Append")==0) {
        	    if (!ptls.is_open()) ptls.open(ptlname.c_str(),std::ios::out|std::ios::app);
            } else {
// saves a preexisting output file under old.txt extension.
	            int pos= (int) std::min(ptlname.find('.'),ptlname.length());
                string savename=ptlname.substr(0,pos)+".old.txt";
                if (ptls.is_open()) ptls.close();
                remove(savename.c_str()); // a preexisting 'old' file would prevent the rename()
                rename(ptlname.c_str(),savename.c_str());
                ptls.open(ptlname.c_str(),std::ios::out);
        	    if (!ptls.is_open()) {cerr<<"Could not open 'pointls' file "<<ptlname<<". I exit."; if(cinGetOnError) cin.get();exit(-1);};
            }

           if ( ( (int) DesignPoint.size() ) != PointNumber) { // should *always* be identical... (even if variableParNbr==0 in which case all its rows are identical)
               cerr<<"(!) DesignPoint.size()!=PointNumber in CSampling::analyse()."<<endl;
               cerr<<"Debugging information: "<<endl;
               cerr<<DesignPoint.size()<<" "<<PointNumber;
               cerr<<"   Contact the authors. I exit";
               if (cinGetOnError) cin.get();
               exit(-1);
           }
            //FR->FR the code for replicate points either uses DesignPoint.size or PointNumber. A bit confusing. But see test right above.
            {int PtNbrWithReplicates=sequenceAddRepl[std::min(int(ki),int(sequenceWrite.size()-1))];
        /*FR->FR petit problem si on n'a pas mis de writeSequence dans les settings, zero replicates are computed...
        */
                if (PtNbrWithReplicates==0) { /**
                    this means that seqeucneWrite did not contain the addReplicates options,
                    still, one must compute 2 replicates for some points :**/
                    if (_replicatedPoints>-1) {
                        PtNbrWithReplicates=_replicatedPoints; //explicit fixed value in settings
                    } else { //default: at least 10 and at least PointNumber/30 to 'ensure' enough replicates in GCV points
                        PtNbrWithReplicates=max(PointNumber/30,10);
                    }
                }
                if (PtNbrWithReplicates==0) { // max(...,.../0) seems fatal...
                    everyNpoints=numeric_limits<int>::max();
                } else {
                    everyNpoints=max(1,PointNumber/PtNbrWithReplicates); // should always be >0 because of "... % everyNpoints" later
                }
            } // end block PtNbrWithReplicates
//        paireIdx=debutpt*replicatedPoints/PointNumber+1;
           //	_StartTime=clock();
       	    time(&_StartTime);
            { string zut=ctime(&_StartTime); // ends with a newline...
               zut=zut.substr(0,zut.length()-1);
               ptls<< "## "<<zut;
            }
            ptls<<"; Sample: "<<mls;
            ptls<< "; # of loci: "<<nloci;
            ptls<< "; NrunsPerPoint="<<args.nruns<<";"<<std::endl;
            ptls<< "## pointmin="<<pointmin<<"  pointmax="<<pointmax <<" out of PointNumber="<<PointNumber<< std::endl;
            // calcul des points et ['e]criture cumulative dans pointls_<_mls>.txt
            //std::cout<<debutpt<<" "<<PointNumber<<" "<<pointmax;getchar();
            int localit=0;
            for (int ii=debutpt-1;ii<pointmax;ii++) {
               localit++;
            //affichage pour v['e]rification
            //cout << "\nStarting MLE calculations with "<<ParameterNumber<<" parameters for point :";getchar();
            //for(int i=0;i<ParameterNumber;i++) cout<<DesignPoint[ii][i]<<" "; //getchar();
		    //cout<<endl;
                args.point=ii;
                if (pointsVariableParNbr==-1) {
                    cerr<<"(!) From CSampling::analyse: [...]VariableParNbr has not been correctly set. I exit"<<endl;
                    if(cinGetOnError) cin.get();
                    exit(-1);
                }
                /* The above test check that we went through either Sampling or Readpoints.
                  if we went through read_points_from_R then boundsVariableParNbr is still -1 */
                if (boundsVariableParNbr==0) { //<=> gone through sampling and ub=lb <=> upperbound=lowerbound in settings or ParameterPoint given
                    /** the way to have say 10 replicates for one point is to use ParameterValue=... followed by PointNumber=10
				     replicatesForThisPoint will be 1 but the program will go 10 times over THIS code **/
                    //PointNumber operates and may give as many replicates as requested. Hence this is sufficient:
                    replicatesForThisPoint=1;
                } else {
                     if (PointNumber==1) {
                        cerr<<"pointNumber=1 appears suspect. Check settings. I exit.";
                        /** maybe that would be a valid case, but we need to identify when it is so in order to set replicatesForThisPoint **/
                        if (cinGetOnError) cin.get();
                        exit(-1);
                    } else if (sequenceAddRepl[std::min(int(ki),int(sequenceWrite.size())-1)]>0) { // option to only add replicates...
                        if ( ((ii+1) % everyNpoints)==0) { //attention, n % 0 est une cause de plantage douloureux.
                            replicatesForThisPoint=1; //assumes already one value was computed
                        } else replicatesForThisPoint=0; // not an 'everyNpoints'
                    } else if (((ii+1) % everyNpoints)==0) { //attention, n % 0 est une cause de plantage douloureux.
                        replicatesForThisPoint=2; //if variableParNbr>0 each point in the design is different; we compute some replicates
                    } else replicatesForThisPoint=1;
                }
#ifdef DEBUG
                if(DebugHypercube) cout << "ReplicatesForThisPoint=" << replicatesForThisPoint << endl;
#endif
                for (int replicate=0;replicate<replicatesForThisPoint;replicate++) { //FR 03/02/08

    /********************* here the computation of the likelihood ***************/
                	mloglik=externalfn(DesignPoint[ii],&_mls);
   /*****************************************************************************/

        	        time(&_EndTime);
        	        _secs=(float) (_EndTime- _StartTime);
                    //soit paral avec 2 Jobs soit non paral... plus affichage toutes les 2 secondes maxi. (the 1+ avoids % 0)
#ifdef DEBUG
		            if(singlePointBool && (PointNumber<100)) {
                        cout << "\nEnd of likelihood calculation with "<<ParameterNumber<<" parameters for 1 point :";
			            for(unsigned int i=0;i<ParameterNumber;i++) cout<<DesignPoint[ii][i]<<" ";
			            cout<<endl;
			            cout<<setprecision(10)<<"L="<<exp(-mloglik[0])<<" logL=";ostream_vector(mloglik,cout);
			            cout<<"; Takes approx. "<<(float) _secs/(ii-debutpt+2)<<"s . "<< endl;
			            cout << endl;
		             }
#endif
        	         if (
        	               ((MaxSimultJobs<3)&& (localit==1|| (_secs-_prevsecs>2))) //always displays the first and more than two seconds since last *display*
                        ||
                           PrintLikelihood
                        ) {
        		        std::cout<<"\r"<<ii+1<<" ";
                        //if (DesignPoint[ii].size()<4) { // protection against multiline output => \r messed.
                        ostream_vector(DesignPoint[ii],std::cout);
        		        std::cout<<" ";
                        //}
        		        std::cout<<mloglik[mloglik.size()-1-int(asymptoBool)];
                        std::cout<<" "<<_secs/(localit)<<" s per point.          "<< std::flush;
                        _prevsecs=_secs; // modified on *display*
                     }
//        	     if( (IBDbool && DesignPoint[ii][2]<0) && (std::std::isnan(mloglik.front()))||(std::std::isnan(mloglik.back()))) { //gnegBool
//                    ptls<<"#";
//        		    std::cout<<"(!) NaN -Ln(L) for some point. Commented out in output file\n"<<std::endl;
//        	     }
                 //def std::isinf: added a flag in the sh file on IFR... for std::isinf
                 //reason for test: if the R data contains inf the variable is treated as factor levels. Nonsense results
				     if((std::isinf(mloglik.front()))||(std::isinf(mloglik.back()))) {	// CBR : N.B. std::isinf() is not part of the C++ STL
                        ptls<<"#";
        		        std::cout<<"(!) infinite -Ln(L) for some point. Commented out in output file\n"<<std::endl;
        	         }
                    if((std::isnan(mloglik.front()))||(std::isnan(mloglik.back()))) {
                        ptls<<"#";
        		        std::cout<<"(!) -Ln(L) is not a number for some point. Commented out in output file\n"<<std::endl;
        		        std::cout<<"(!) This is not a normal behavior, please contact the authors\n"<<std::endl;

                    }
        	         ofstream_vector_mll(DesignPoint[ii],mloglik,ptls,parDigits);
                 } // end loop on replicate
			} //ii -> PointNumber
        } // end test on values of WriteSequence
        /**** likelihoods have been written to pointls,
         now performs kriging ****/
	    if (!noRCallBool && !(ki==0 && (cmp_nocase(sequenceWrite[0],"ReadBounds")==0 || cmp_nocase(sequenceWrite[0],"ReadPoints")==0))) {
            //kriging is explicitly disabled on first iteration if there are already bounds/points to read)
	        /**** KRIGING (call R executable) ****/
            string krigreturnstring=krig(); // calls callR.bat with appropriate arguments=> runs the migraine_<_mls>.R batch file
            if (cmp_nocase(krigreturnstring,"Call to R didn't work")==0)
            // currently -1 in return(-1) is not used. -1 just memo that this is not the standard return. But this is not exit() so the final code of main() is executed
                  return(-1); // ELSE continue
        }
        /**** Read output file whenever needed ****/
        if (
              (ki==0 && cmp_nocase(sequenceWrite[0],"ReadBounds")==0)
           ||  (ki==1 && (cmp_nocase(sequenceWrite[0],"ReadPoints") !=0) ) // NOT equal. ki=1 && ReadPoints may occur in a different directory from that where output_[....] may be
           || ki>1 // for screen display if for nothing else; we may expect the file to be there.
        ) { // but not readPoints
		// reads the ML estimates needed (in addition to the bounds_[...].txt file ) for fixing the next bounds
		    fromR.open(fromRname.c_str());
		    if (fromR.fail()) {
			    fromR.clear();
			    std::cerr<<"\nCannot open file "<<fromRname<<" (output from R call).";
			    if (cmp_nocase(sequenceWrite[0],"ReadBounds")==0) {
			       std::cerr<<"\nThis file is required with the readBounds(Only) setting!";
                }
			    std::cerr<<"\n If the file is there, it may be locked by another program.";
			    std::cerr<<"\n Hint: close any R GUI session, \n\tthen return to this window and hit any key.";
			    if (cinGetOnError) cin.get(); // ...or Remove all objects
			    fromR.open(fromRname.c_str());
			    if (fromR.fail()) {
				    fromR.clear();
				    std::cerr<<"\nCannot open file "<<fromRname<<" (output from R call). I exit.";
				    if(cinGetOnError) cin.get();
				    exit(-1);
			    };
		    };
		    std::cout.setf(std::ios::left);//std::cout<<left; pas compris par vieilles versions gcc
// maintenant on passe en echelle 2Nm si on etait en echelle Nb (doit venir apr[`e]s la reconversion en g
// display kriging variable names
            if (IBDbool) {
                if (BoundsDefs.size()>1 && (cmp_nocase(BoundsDefs[1],"latt2Ns2")==0 || cmp_nocase(BoundsDefs[1],"Nb")==0 ) ) {
    		    if(D1IBDbool) std::cout<<"  2Nmu     * 2Ds2 *   g"<<endl;
    		    if(D2IBDbool) std::cout<<"  2Nmu    * 2DPis2 *  g"<<endl;
    //           ostream_vector(BoundsDefs,std::cout); // OK si on avait compl['e]t['e] BoundsDef par vals par d['e]fault
    //           std::cout<<endl;
/* FR 16/01/2008 this bit of code does not seem to make sense here. But probably without consequence
                    float temp;
                    for(std::vector<std::vector<float> >::iterator ii=DesignPoint.begin();ii<DesignPoint.end();ii++) {
                    	temp=((*ii)[2]);
                    	temp=(1+temp)/((1-temp)*(1-temp));
                    	(*ii)[1]/=temp;
                    }
*/
                } else {
                   std::cout<<"  2Nmu      2Nm     g"<<endl;
                }
            } //endif display variable names (IBDbool else ...  )
// display upper bounds (values)
		    for (unsigned int ii=0;ii<ParameterNumber;ii++) {std::cout<<" "<<std::setw(8)<<ubvar[ii];} std::cout<<std::endl;
// now takes values from output[...].txt
// goes to last line which should contain the global optimum in case several blocs were computed
		    std::string dummystr;
		    std::stringstream strstr(std::stringstream::in|std::stringstream::out);
		    while (!fromR.eof()) {outputline=nextline;getline(fromR,nextline);}
		    fromR.clear();fromR.close();
		    strstr<<outputline;
		    strstr>>atst>>at1>>at2; // skips pointls...txt, first point, last point
            // note mlee follows canonical order
            for (unsigned int ii=0;ii<ParameterNumber;ii++) strstr>>mlee[ii];
            if (IBDbool) {
               strstr>>eNb;//neighborhood size, metric units since 16/01/2008;
            }
            strstr>>emll;//std::cout<<emll; getchar();
            if (IBDbool && BoundsDefs.size()>1 && (cmp_nocase(BoundsDefs[1],"latt2Ns2")==0 || cmp_nocase(BoundsDefs[1],"Nb")==0 )) {
            // affichage Nb [`a] la place de Nm  (& reciprocally)
    		    std::cout<<" "<<mlee[0]<<" "<<eNb;
    		    for (unsigned int ii=2;ii<ParameterNumber;ii++) std::cout<<" "<<mlee[ii];
    		    std::cout<<" 2Nm="<<mlee[1];
            } else if (IBDbool && BoundsDefs.size()>1 && (cmp_nocase(BoundsDefs[2],"condS2")==0) ) {
                float lattcondS2=0;
    		    for (unsigned int ii=0;ii<ParameterNumber-1;ii++) std::cout<<" "<<mlee[ii];
                if (D1IBDbool) {
                    lattcondS2=( (float) 1.+mlee[2])/(( (float) 1.-mlee[2])*( (float) 1.-mlee[2]));
                } else if (D2IBDbool) {lattcondS2=( (float) 1.+mlee[2])/(( (float) 1.-mlee[2])*( (float) 1.-mlee[2])*( (float) 2.-mlee[2]));}
    		    std::cout<<" "<<lattcondS2;
            } else { //
    		    for (unsigned int ii=0;ii<ParameterNumber;ii++) std::cout<<" "<<mlee[ii];
    		    if(D1IBDbool) std::cout<<" 2Ds2="<<eNb;
    		    if(D2IBDbool) std::cout<<" 2DPis2="<<eNb;
            }
            std::cout<<" LogL="<<emll<<" (kriging iteration: "<<ki+1<<")"<<std::endl;
//displays lower bounds
		    for (unsigned int ii=0;ii<ParameterNumber;ii++) {std::cout<<" "<<std::setw(8)<<lbvar[ii];}
            std::cout<<std::endl<<std::endl;
//if (mlee[0]>	ubvar[0]) {std::cout<<"zut";getchar();}
		// computes new bounds (bounds[...].txt should be in 'user' parameter space, I guess -- 13/02/2008 )
        } //end if (complex test) read output...
        /**** Output file has been read whenever needed ****/
        /**** Now read bounds file whenever needed ****/
        boundsFromRBool=false;
        /*first subtest so that with ReadBounds, Migraine doesn't bother whether the user has written
        NextBounds=From1DCI rather than NextBounds=,From1DCI*/
	    if ( (ki==0 && cmp_nocase(sequenceWrite[0],"ReadBounds")==0)
		    || cmp_nocase(boundsSource[std::min(int(ki),int(boundsSource.size()-1))],"From1DCI")==0
            || cmp_nocase(boundsSource[std::min(int(ki),int(boundsSource.size()-1))],"From2DCI")==0) { //kriging has produced potentially useful bounds
            boundsFromRBool=true; // indicateur local
            {stringstream bdst;bdst<<"bounds_"<<mls<<".txt";bdst>>boundsname;}
            ifstream boundstream(boundsname.c_str());
    	    if (boundstream.fail()) {
    		    boundstream.clear();
    		    std::cout<<"\nCannot open file "<<boundsname<<" (output from R call). I exit.";
                if (cinGetOnError) {cin.get();}
                exit(-1);
            };
    	    for (unsigned int ii=0;ii<ParameterNumber;ii++) {
        	    boundsvals[ii].resize(2,0);boundsindic[ii].resize(2,0);
                boundstream>>boundsindic[ii][0]>>boundsindic[ii][1]>>boundsvals[ii][0]>>boundsvals[ii][1];
                // [0] and [1] are indicators: 0 => use the value (previous range too large); 1 => opposite
            }
            boundstream.close();
        }
        /**** Bounds, if any, have been read ****/
        /**** now fills next DesignPoints and generates next R file****/
	    if (ki+1<sequenceWrite.size()) { // 31/03/2007
	        float zou;
// be careful that variables such as pointNumber must be reset when next _sample_ is analyzed (code further below)
/* Here we should reset PointNumber to le latest value.

   If we give explicit pointmin or
   pointmax in settings it may be that we want a job to compute likelihoods for a block after a first iteration.
   hence we should not reset pointmin and pointmax to PointNumber in that case.

   If we do not reset these values here they keep the values of the previous iteration. But they may have been set to the
   previous pointNumber, i.e. the previous sequencePointNumber value, and these values are no longer correct.

   Hence we want to use the original info from the pointmin and pointmax settings, originally stored in _pointmin and _pointmax.
   That is, _pointmin and _pointmax should serve to store these original values, not for other purposes:
   => changed code in various places

   (Commentaire Faux a ce point : in particular if _pointmin and _pointmax were at their default negative values they must stay so!)
   */
            PointNumber=sequencePointNumber[std::min(int(ki+1),int(sequencePointNumber.size()-1))];
            if(PointNumber<=0) {
    		    std::cout<<"(!) From CSampling::analyse: PointNumber<=0. Check whether this make sense; hypercube_sampling should not be called.";
    		    std::cerr<<"   I exit.";
                if (cinGetOnError) {cin.get();}
                exit(-1);
            }
            if (userPointMinOrMax)
            	continue;
            else
            	pointmax = PointNumber;
//            DesignPoints.resize(ParameterNumber,PointNumber);
//cout<<std::min(ki,int(boundsSource.size())-1)<<" ...  "<<boundsSource[std::min(ki,int(boundsSource.size())-1)];getchar();
//         reads the FIRST element of boundsSource at the SECOND iteration
            if (cmp_nocase(boundsSource[std::min(int(ki),int(boundsSource.size()-1))],"pointsFromR")==0) {
               /*** Read points from R file ***/
               stringstream stst(stringstream::in|stringstream::out);
               string ptlname;
               stst<<"nextpoints_"<<mls<<".txt";
               stst>>ptlname;
               DesignPoints.read_points_from_R(ptlname);
               // saves the actual nextpoints_X.txt file under the name nextpoints_X.old.txt
                // it avoid Migraine to read the same nextpoint.txt file when the R analysis chrashed
                int pos= (int) std::min(ptlname.find('.'),ptlname.length());
                string savename=ptlname.substr(0,pos)+".old.txt";
                remove(savename.c_str()); // a preexisting 'old' file would prevent the rename()
                rename(ptlname.c_str(),savename.c_str());
            } else {
            /*** set bounds from output and bounds files previously read ***/
            //setting lower bounds LBvar
                for (unsigned int ii=0;ii<ParameterNumber;ii++) {
    		       old_lb[ii]=lbvar[ii];
    		       if(samplingScale.size()>ii && cmp_nocase(samplingScale[ii],"LogScale")==0) {
                      MLE=log(mlee[ii]);
                      EdgeIndice=(log(lbvar[ii])+(log(ubvar[ii])-log(lbvar[ii]))/edgefactor);
                   } else {
                      MLE=mlee[ii];
                      EdgeIndice=(lbvar[ii]+(ubvar[ii]-lbvar[ii])/edgefactor);
                   }
    		       if (MLE<EdgeIndice ||
                          (boundsFromRBool && boundsindic[ii][0]==1) //==1 : don't use this bound
                      ) {
                            lbvar[ii]/=expfactor; //plus proche zero
 		 	            } else if (boundsFromRBool && boundsindic[ii][0]==0) {
 		 	                lbvar[ii]=boundsvals[ii][0];
                        } else {
                            lbvar[ii]=(mlee[ii]+lbvar[ii]*expfactor)/(expfactor+1); // moy pond['e]r['e]e
                        }
                }
               //setting upper bounds UBvar (code unifie' IBD/autres)
               // in each test ubvar still has its old value hence (ubvar[0]-old_lb[0]) is the old range
               for (unsigned int ii=0;ii<ParameterNumber;ii++) {
                   if (IBDbool && BoundsDefs.size()>ii && (cmp_nocase(BoundsDefs[1],"latt2Ns2")==0 || cmp_nocase(BoundsDefs[1],"Nb")==0 ))
                       zou=eNb; //vraiment Nb
                  else zou=mlee[ii];
               if(samplingScale.size()>ii && cmp_nocase(samplingScale[ii],"LogScale")==0) {
                    MLE=log(zou);
                    EdgeIndice=(log(ubvar[ii])-(log(ubvar[ii])-log(old_lb[ii]))/edgefactor);
               } else {
                   MLE=zou;
                   EdgeIndice=(ubvar[ii]-(ubvar[ii]-old_lb[ii])/edgefactor);
               }
    		   if (MLE>EdgeIndice ||
                    (boundsFromRBool && boundsindic[ii][1]==1) //==1 : don't use this bound
                  ) { // extrapolate out of already explored range
                        if (IBDbool && ii==2)
                            /*
                            note BoundsDefs.size()>ii && cmp_nocase(BoundsDefs[ii],"g")==0
                            pas bien car FALSE si BoundsDefs[...] est "Canonical"...
                            */
                            ubvar[ii]=(ubvar[ii]*expfactor +  (float) 1.)/(expfactor +  (float) 1.); // so that g remains below 1
                         else ubvar[ii]*=expfactor;
               } else if (boundsFromRBool && boundsindic[ii][1]==0)
                               ubvar[ii]=boundsvals[ii][1];
                          else ubvar[ii]=(ubvar[ii]*expfactor+zou)/(expfactor+1); // bracketing within  explored range
               }
               cout << "New Bounds are :" << endl;
               for (unsigned int ii=0;ii<ParameterNumber;ii++) cout << lbvar[ii] << " ";
               cout << endl;
               for (unsigned int ii=0;ii<ParameterNumber;ii++) cout << ubvar[ii] << " ";
               cout << endl;
               // new points (not the firstfill before any likelihood computation)
                      samplingloop(lbvar,ubvar); //call from within analyse =within fn membre de CSampling!
            } //endif pointsfromR else...
            /**** next DesignPoint has been filled ****/
            /**** Now write next R file ****/
            Rparams.fill(ki+1);
            if (!(ki==0 && cmp_nocase(sequenceWrite[0],"KrigOnly")==0)) {
                write_R_batch_file(lbvar,ubvar,Rparams);
                //writes migraine_<_mls>.R with appropriate lb,ub (before further iteration)
            }
        } //if ki+1<krigIterations
    }//krigIterations

    // at this point all kriging iterations have been performed
    if ( ! noRCallBool && false) { //FR inactivated this code 12/2010. The output.txt file is not useful [cf aussi remove("output.txt") supprim'e au d'ebut de main() ]
        //krig has been called hence the results are in fromRname...
        // copy final result in output.txt
	    std::ofstream globalout("output.txt",std::ios::out|std::ios::app);
	    std::ifstream mlsout(fromRname.c_str());
	    while (!mlsout.eof()) {outputline=nextline;getline(mlsout,nextline);}
	    mlsout.clear();mlsout.close();
	    globalout<<outputline<<std::endl;
	    globalout.close();
     }
    ptls << "## Finished pointls." << endl;
    ptls.close();
    quickexit:
    delete[] lbvar;
    delete[] ubvar;
#ifdef UNIX_PARAL
   if (AvailableProcessors>1)
		std::cout <<"Child for mls="<< mls<< " finished."<<std::endl;
   else
		std::cout <<"Run for mls="<< mls<< " finished."<<std::endl;
#else
		std::cout <<"Run for mls="<< mls<< " finished."<<std::endl;
#endif
	std::cout.flush();

//cout<<"fin CSampling::analyse";getchar();
return(0);
}; // CSampling::analyse


CAnalysePARAL::CAnalysePARAL(CSampling& DesignPoints,const int& MaxSimultJobs): CSampling(DesignPoints) {
this->MaxSimultJobs=MaxSimultJobs;
};

