/***************************************************************************
© F. Rousset, R. Leblois 2007

francois.rousset@univ-montp2.fr
raphael.leblois@mnhn.fr

This file is part of Migraine. This software is a computer program
whose purpose is to perform statistical analyses.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

 ***************************************************************************/
#ifndef H_LOCUS
#define H_LOCUS

#include <cmath>
#include <limits>
#include "genepop.h"
//#include "migraine.h" //only for D2IBDbool could be removed later ?

class CgeoInfo {
// an interface between Genepop files and variables used by the likelihood algos
// info common to all loci in genepop file
public :
    int xBins,yBins;
    float llx,lly,dx,dy;
    float BinWidth,cosrot,sinrot;
    CgeoInfo() {};
	int fill(std::vector<float> geoBoundingBox,int geoBinNbr) {
	    //geoBinNbr is maximum one dimensional # of bins
	    //2D and (ultimately 1D via IBDcrunchLocTable
	   float ratio;
	   llx=geoBoundingBox[0];lly=geoBoundingBox[1];dx=geoBoundingBox[2];dy=geoBoundingBox[3];
	   cosrot= (float) cos(geoBoundingBox[4]); // note that angle has been converted to radians when HabitatPars was read
	   sinrot= (float) sin(geoBoundingBox[4]);
       if (dx>dy) {
          ratio=dy/dx;
          BinWidth=dx/geoBinNbr;
          xBins=geoBinNbr;yBins=std::max(1,int(ceil(ratio*geoBinNbr)));
       } else {
          ratio=dx/dy;
          BinWidth=dy/geoBinNbr;
          yBins=geoBinNbr;xBins=std::max(1,int(ceil(ratio*geoBinNbr))); // FR 08/2014; was not previously consistent with doc  "The largest dimension of the habitat is divided by the given number of bins (here, geoBinNbr=...)"
       } // such that ratio<=1
 //std::cout << dx<<" "<<dy<<" "<< ratio <<" "<< xBins <<" "<< yBins ; std::cin.get();
       return(0);
    };
	int fill(int geoBinNbr) {//called when geoBinWidth=0
       BinWidth=std::numeric_limits<float>::quiet_NaN();xBins=geoBinNbr;yBins=1;
       return(0);
    };
    int print() {
        std::cout<<" "<<BinWidth<<" "<<xBins<<" "<<yBins<<" "<<cosrot<<" "<<sinrot;
		return(0);
    }
	virtual ~CgeoInfo() {};//destructeur should be virtual (Primer+ p. 665)
};



class CProcessedLocus {
// an interface between Genepop files and variables used by the likelihood algos
// infos from several Genepop files to be stored in a global vector<CProcessedFile>
public :
	std::vector<int> alleleIndices;
    std::vector<std::vector<int> >alleleCounts;
    int locus; // basic info for finding the right element of final SMMstepSizes and for checking
    std::string fileName; //idem
    std::string locusName;	// CBR (03.09.2013) stores the locus name as given in the Genepop file
    std::map<std::string, int> mapIndLab2State; // CBR (30.08.2013) for a given locus maps every individuals genepop label to its allelic state
	CProcessedLocus() {};//constructeur
	CProcessedLocus(CFichier_genepop *gp,int iLoc);//constructeur
	virtual ~CProcessedLocus() {};//destructeur should be virtual (Primer+ p. 665)
	int copyTable(std::vector<std::vector<int> >counts);
	int copyTable(std::vector<std::vector<std::vector<int> > >counts);
	int fillTable(CFichier_genepop *gp);
	int fillGrid(CFichier_genepop *gp,unsigned int latticeSize,std::vector<unsigned int> samPos);
	int writeAdHocFile(std::ofstream& stream);
};

//decalred in migraine.cpp
extern std::vector<CProcessedLocus>ProcessedLoci;
extern CgeoInfo geoInfo;
#endif
