/***************************************************************************
© F. Rousset, R. Leblois 2007

francois.rousset@univ-montp2.fr
raphael.leblois@mnhn.fr

This file is part of Migraine. This software is a computer program
whose purpose is to perform statistical analyses.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

 ***************************************************************************/
#ifndef H_WRITER
#define H_WRITER

#include "hypercube.h"


template <typename argtype>
int write_R_batch_file(argtype lbloc[], argtype ubloc[],  CRParameters& Rparams) {
/******* write R batch file  to analyze the output *******/
/*            IMPORTANT SUGGESTIONS to keep this clear
use ## to mark explanations in R code;
use # to mark commented out R _code_;
use std::endl rather then \n to show newlines in R code;

For warnings in english,  set language=en
in (e.g.) C:\Program Files\R\R-2.12.2\etc\Rconsole
*/
   std::string krigName;
    //int jj=0;

   if (fittedpars.size()==0) {
      std::cout<<"(!) No parameter to estimate; check parameter ranges.";
       if (cinGetOnError) std::cin.get();
       exit(-1);
   }
   if ( ! singlePointBool) {
        std::stringstream stst(std::stringstream::in | std::stringstream::out);
        //FR 11/2008: match the R file name to the pointls file name is the only safe way:
        int pos= (int) std::min(ptlname.find('_'), ptlname.length());
        std::string tempstr=ptlname.substr(pos);
        pos= (int) std::min(tempstr.find('.'), tempstr.length()); // assumes no "." before the ".txt" !
        tempstr=tempstr.substr(0, pos);
        stst<<"migraine"<<tempstr<<".R";
        stst>>krigName;
    }
    std::ofstream krigR(krigName.c_str(), std::ios::out);
    /** ONLY COMMENTS **/
    /* Problem if file is edited with Notepad :
    http://stackoverflow.com/questions/3255993/how-do-i-remove-i-from-the-beginning-of-a-file
       Three words for you:
       Byte Order Mark (BOM)
       That's the representation for the UTF-8 BOM in ISO-8859-1. You have to tell your editor to not use BOMs or use a different editor to strip them out.

    http://www.ktuk.net/phpBB3/viewtopic.php?f=32&t=3835
       For some reason Microsoft has designed [Notepad and Wordpad] to automatically insert the BOM when files are saved as UTF-8 and as such they should not be used.
    */
    krigR<<"## This is free software written by a free software"<<std::endl;
    krigR<<"## covered by the CeCILL licence (GPL compatible)."<<std::endl;
    krigR<<"## © F. Rousset & R. Leblois 2007- ; contributor: C.R.Beeravolu"<<std::endl;
    krigR<<"## Migraine version "<<version<<"."<<std::endl;
    if(IBDbool) {
        krigR<<"##  rosglobal$latt2Ns2 : estimate of 2{Nh}m condS2 where Nh is haploid number per lattice step and condS2 is is lattice units"<<std::endl;
    }
    krigR<<"## ..."<<std::endl;
    krigR<<std::endl;
    /** CODE BEGINS HERE **/
    krigR<<"if (interactive()) {options(error=recover)} else {"<<std::endl;
    krigR<<"  options(echo = FALSE)"<<std::endl;
    krigR<<"  options(error = quote(dump.frames(paste(\"dump\",blackbox.getOption(\"jobSampleNbr\"),sep=\"\"), TRUE)))"<<std::endl;
    krigR<<"}"<<std::endl;
    krigR<<"rm(list=ls())"<<std::endl;
    krigR<<"chk <- try(library(blackbox))"<<std::endl;
    krigR<<"if (class(chk)==\"try-error\") stop(\"(!) Failed to load the 'blackbox' library\")"<<std::endl;
    krigR<<"chk1 <- try(library(doSNOW))"<<std::endl;
    krigR<<"if (class(chk1)==\"try-error\") message(\"(!) Failed to load the 'doSNOW' library, parallelization will use 'pbapply()' instead \")"<<std::endl;


    /** write Rparams and other model dependent parameters to the krigR file,  including useC
    The idea is towrite safe code by allowing the C variables to affect the R code only through one entry point:
    Once some info is passed (once!)  from C to the R code,  it should be  processed entirely within the C code.

    IBDbool or DemographicModel dependend R code are annoying exceptions to this rule
    **/
    Rparams.write_R_GP_code(krigR); /// ici je dois avoir l'essentiel de GP$ mais petit residu:

    // FR->FR ??? essayer de le depalcer dans write_R_GP_code

    if(testPointList.size()>0) {
        krigR<<"## Building list of points to be tested; list(list(...)) stack sublists rather than appends a single list"<<std::endl;
        krigR<<"   GP$testPointList <- list()"<<std::endl;
        for (std::vector<Ctestpoint>::iterator it=testPointList.begin();it!=testPointList.end();it++) {
            if ((*it).names.size()>0) {
                (*it).write(krigR, "  tmplist <- c("); // appends testPointList with a sublist for each testpoint.
                krigR<<"   GP$testPointList <- c(GP$testPointList, list(tmplist))"<<std::endl;
            }
        }
        krigR<<"## ...list of points done"<<std::endl;
        /// krigR<<fillminmax1DtestvaluesString<<"## constructs a list of minimal and maximal tested values in one-dim profile LRT's"<<std::endl; // semble obsolete
    } else {
        krigR<<"GP$testPointList <- NULL"<<std::endl;
    }

    krigR<<"# GP$interactiveGraphics <- FALSE # Uncomment this to switch to file graphic output in interactive session"<<std::endl;
    krigR<<"GP <- preprocessbboptions(optionList=GP)"<<std::endl;

    krigR<<"do.call(blackbox.options, GP)"<<std::endl;




    // krigR<<nottoomuchfn<<std::endl; keep it,  but obsolete

    if (RcodeStString.str().size()>0) {
        krigR<<"## Explicit R code in Migraine settings:"<<std::endl;
        krigR<<RcodeStString.str();
        krigR<<std::endl;
    }


    krigR<<"if (!blackbox.getOption(\"interactiveGraphics\"))  {providePlotFile(blackbox.getOption(\"basicRplotsfile\"))}"<<std::endl;


    /** FR->FR this needs some description... **/
    krigR<<"## *************************"<<std::endl;
    krigR<<"## **** Processing data ****"<<std::endl;
    krigR<<"## *************************"<<std::endl;



    std::string locstring="\n\
pointls <- buildPointls(dataFile=blackbox.getOption(\"dataFile\"), \n\
                        respCols=NULL,\n\
                        subsetRows=NULL,\n\
                        ycolname=\"-ln(L)\", \n\
                        cleanResu=blackbox.getOption(\"cleanResu\"))\n\
\n\
FONKgpointls <- buildFONKgpointls(pointls=pointls) # includes blackbox.options(FONKgpointls = FONKgpointls)\n\
\n\
GCVblob <- calcGCV(sorted_data=FONKgpointls, \n\
                   CovFnParam=blackbox.getOption(\"CovFnParam\"),\n\
                   GCVptnbr=blackbox.getOption(\"GCVptnbr\"), \n\
                   topmode=\"dLnL\",\n\
                   verbose=blackbox.getOption(\"verbosity\"), \n\
                   cleanResu=blackbox.getOption(\"cleanResu\"), \n\
                   optimizers=blackbox.getOption(\"optimizers\") \n\
                   )\n\
\n\
provideDevice(bbdefaultPars=TRUE) ## one window for _each_ graphic\n\
\n\
fitobject <- calcPredictorOK(FONKgpointls=FONKgpointls, \n\
                   minKrigPtNbr=blackbox.getOption(\"minKrigPtNbr\"), \n\
                   krigmax=blackbox.getOption(\"krigmax\"), \n\
                   topmode=\"dLnL\", \n\
                   rawPlots=TRUE, \n\
                   cleanResu=blackbox.getOption(\"cleanResu\")) ## includes rawProfiles\n\
\n\
";
    krigR<<locstring<<std::endl;  //computations

    krigR<<"primarymax <- maximizeOK(fitobject=blackbox.getOption(\"fitobject\"), cleanResu=blackbox.getOption(\"cleanResu\"))"<<std::endl;

    /** 1D profiles**/
    if(ParameterNumber>1) krigR<<"calc1Dprofiles()"<<std::endl; //computation of *1D* profile grids

    /** Slice plots,  2D profiles...**/
    // Slice plots come after a first profiling so that the isoline goes to the best possible plot
    // it would be statistically better to compute them after the 2D profiles,  but some early plotting is also useful
    locstring="\n\
\n\
calc2D3Dplots() ## **** plain or slice plots of predicted likelihood ****\n\
\n\
## **** More graphics : 2D profiles ****\n\
calcProfileLR(varNames=blackbox.getOption(\"fittedNames\"), cleanResu=blackbox.getOption(\"cleanResu\")) \n\
\n\
## regression observed/ kriging prediction\n\
provideDevice(bbdefaultPars=TRUE) ## one window for _each_ graphic\n\
plot(blackbox.getOption(\"fitobject\"), which=1) ## Diagnostic plot lnL Estimates vs lnL predicted values\n\
\n\
## **** Likelihood-ratio tests ****\n\
LRTs <- calcLRTs(testPointList=blackbox.getOption(\"testPointList\"),\n\
         cleanResu=blackbox.getOption(\"cleanResu\"))\n\
\n\
## **** 1D CIs ****\n\
if (length(blackbox.getOption(\"oneDimCIvars\"))>0 \n\
           || blackbox.getOption(\"nextBounds\") %innc% c(\"from1dci\", \"frompoints\")) {\n\
  CIpointsList <- calc1DCIs(oneDimCIvars=blackbox.getOption(\"oneDimCIvars\"), \n\
                     FONKgNames=blackbox.getOption(\"FONKgNames\"), \n\
                     fittedNames=blackbox.getOption(\"fittedNames\"), \n\
                     CIlevel=blackbox.getOption(\"CIlevel\"), \n\
                     nextBounds=blackbox.getOption(\"nextBounds\"), \n\
                     boundsOutfile=paste(\"bounds_\", blackbox.getOption(\"jobSampleNbr\"), \".txt\", sep=\"\"), \n\
                     dataString=blackbox.getOption(\"dataName\"), \n\
                     cleanResu=blackbox.getOption(\"cleanResu\")\n\
  ) \n\
  blackbox.options(CIpointsList=CIpointsList) ## important for not yet transparent code\n\
} \n\
\n\
## **** Sampling design for next iteration ****\n\
if (blackbox.getOption(\"nextBounds\") %==nc% \"pointsfromr\") {\n\
  goodpoints <- sampleByResp(size=blackbox.getOption(\"nextPointNumber\"), \n\
                  outfile = blackbox.getOption(\"nextPointsf\"), \n\
                  useEI=blackbox.getOption(\"useEI\"), \n\
                  rnd.seed=blackbox.getOption(\"ptSamplingSeed\")\n\
  )\n\
}  \n\
\n\
writeFinalInfo(cleanResu=blackbox.getOption(\"cleanResu\"))\n\
# close(blackbox.getOption(\"estimOutf\")) ## ! leaves connection open if interactive run !\n\
";
    krigR<<locstring<<std::endl;  //computations

    krigR.close();
return 0;
}


#endif
