/***************************************************************************
© F. Rousset, R. Leblois 2007

francois.rousset@univ-montp2.fr
raphael.leblois@mnhn.fr

This file is part of Migraine. This software is a computer program
whose purpose is to perform statistical analyses.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author, the holder of the
economic rights, and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading, using, modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate, and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and, more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

 ***************************************************************************/

#define reg register

#include <cstdlib>
#include <ctime>
#include <numeric>//pour accumulate
#include "migraine.h"
#include "hypercube.h" //ParameterNumber
// alea is defined latin.cpp
#include "mainDG.h"
#include "dg.h"
#include "dgCMigCMut.h"
//#include "computeId.h"
#include "dgCSample.h"
#include "myutils.h"
#include "nexus.h"
#include "PCL_SISR.h"

bool golubbool,runsTempFileBool=false,calcQibdtableDoneBool=false,UsePCL_SISRbool=false,ForceGivenKbool=false;



using namespace std;


namespace NS_calclogl { // 'local' variables from calclogl that need to be visible from member function of CSample etc.
  //FR->RL expliquer chaq variable et en profiter pour nettoyer le code...
  int MutParamNbr,RelPopSizeParamNbr,MigParamNbr,ChangeTimeParamNbr,RelAncPopSizeParamNbr,AncThetaParamNbr,VarSizeShapeParamNbr;
    //MutParamNbr : number of mutatinal parameters = 1=Theta if not GSM, 2=pGSM and Theta if GSM, or ThetaSEQ and Theta(MIC) for multiple markers and not GSM
  	 										//		 3=(only if multipleMarkersBool=true) ThetaSEQ, pGSM, Theta(MIC)
	//RelPopSizeParamNbr=number of parameters for relative population size i.e. the q's (=0 for IBD and OnePop, = pop nber -1 for Npop and IM)
	//MigParamNbr=number of migration parameters : 2 for IBD, NPop*(NPop-1) for NPop and IM
	//ChangeTimeParamNbr=number of demographic change time parameters (=NPop-1 for IM, 1 for ONePop with VarPopSize and 0 for other models)
	//RelAncPopSizeParamNbr= number of relative ancestral population size parameters (=NPop-2 for IM because the last one is ancestral theta, 0 for other models)
    //AncThetaParamNbr= number of ancestral theta parameters (=Npop-1 for IM because there is a new theta at each split event, 0 for other models)
  vector<parstype> Disp_Pars;
  vector<parstype> _PopSizeTab, _AncPopSizeTab,_AncThetaTab;
  vector<scaledTimeType> _ChangeTimeTab;
  parstype _VarSizeShape;
  parstype _Theta; // copies local de (possibly scaled) args.pars[0]
  parstype _pGSM; // copies local de args.pars[1] for GSM
  parstype _ThetaMIC; // used to assign MutMod1->AncThetaDIG for the ISM model when multipleMarkersBool = true
  int MutScaling;
  parstype Temp;
  long double LnLikelihood,MultinomialCoeff,LnMeanLikelihood;
  vector <int> LineagesPerPopAtDivTime;
  int PopMerged;
  bool Changedbool;//if changes occured in FctnPtr (PiCalc and IS/PAC) and Mig and Mut models in the split model
  int ancEventIt; //iterator over ancestral events
  bool beforeLastEvent; //fast test whether ancestral events should be looked at.
  int localallelenbr; // info sur nombre d'alleles connu ou a deduire d'un fichier de donnees
}


/*-----------ln(gamma(xx))--(Numerical recipe)--------------------------*/
double gammln(double xx)
{double xxx,tmp,ser;
 static double cof[6]={76.18009173,-86.50532033,24.01409822,-1.231739516,0.120858003e-2,-0.53363382e-5};
 xxx=xx-1.0;
 tmp=xxx+5.5;
 tmp-=(xxx+0.5)*log(tmp);
 ser=1.0;
 for(int j=0;j<5;j++)
	{xxx+=1.0;
	 ser+=cof[j]/xxx;
	}
 return -tmp+log(2.50662827465*ser);
}
/*-----------------------------------------------------------------------------*/


double CSample::LnMultinomialCoeff() {
// produit de coeffs multinomiaux, un par demes
// chaque coeff multinomial est n_{deme,+}!/{\prod_alleles n_{deme,allele}}
double lnproba=0;
int s;
//if (Size==0) return 1; //FR 08/2008 semble une bug. Devrait etre 0. Suite doit etre correcte.
	for(int a=0;a<PopSampledOrNot;a++) {
		s=PopEffectif[a];
		//cout << "s=PopEffectif[a]=" << s << endl;
		for(int all=0;all<Sam_AlleleNbr;all++) {
			for(int jj=1;jj<=Effectif[a][all];jj++) {
                //cout << "s=" << s << "; jj=" << jj << "; lnproba before = " << lnproba << endl;
                //cout << "log(float s / jj)=" << log(((float)s)/jj) << endl;
				lnproba=lnproba+log(((float)s)/jj); s--;
				//cout << "lnproba after + = log(((float)s)/jj)" << lnproba << endl;
			}
		}
	}
return lnproba;
}

double CSample::LnHypergeom2tables(CProcessedLocus* PLptr) {
/** computes the \prod of hypergeometric distributions
there is one hypergeom for each deme
For each deme, the hypergeometric describes the distribution of cell counts
of a K*2 table (K alleles, 2 samples merged for each deme)
One sample is the CSample, the other is the ProcessedLocus.
Thus the numerator of each hypergeometric is a product of binomial terms;
and the denom of each geom is the multinomial coeff for total size per deme
partitioned wrt allelic counts per deme.
           [\prod_{j=CS,PL} (margins per allele over alleles! / \prod cells!)]
\prod_demes[-----------------------------------------------------------------]
           [(total (per deme as others)!) / (margins per CS/PL over alleles!)]
*/
double lnproba=0;
int totalperdeme,alleleMarginOverCSPL;
    for (unsigned int deme=0;deme<Effectif.size();deme++) {// loop over demes
        totalperdeme=0;
        for (unsigned int allele=0;allele<Effectif[deme].size();allele++) {// loop over alleles
            alleleMarginOverCSPL=Effectif[deme][allele]+PLptr->alleleCounts[deme][allele]; //marginal sum CS+PL
            totalperdeme+=alleleMarginOverCSPL;
            //numerator
    		for(int jj=1;jj<=Effectif[deme][allele];jj++) {
    			lnproba=lnproba+log(float(alleleMarginOverCSPL)/jj); alleleMarginOverCSPL--;
    		} // was the binomial term
        }
        // denominator
        for(int jj=1;jj<=PopEffectif[deme];jj++) {
			lnproba=lnproba-log(float(totalperdeme)/jj); totalperdeme--;
		} // was division by binomial term
    }
return lnproba;
}

// commun [`a] calclogl et change
//vector < vector <pirestype> > PIs;

//#ifdef ASYMPTO
double CSample::LnPr_sample_from_island_KAM(CMigMod *Mig,CMutMod *Mut) {
double lnproba=0;
double DeuxNuNm=Mut->ThetaDIG*(Sam_AlleleNbr-1)/(PopSampledOrNot*Sam_AlleleNbr)
					+1/(Mig->limit_FSurUnMoinsF); // not direct but....
//cout<<DeuxNuNm<<endl;
double DeuxNuNmx=DeuxNuNm/Sam_AlleleNbr; // avant on a seulement corrig['e] theta
// si utile : ces constantes pourraient etre calc quand Fsur... l'est
int s;
//cout<<endl<<Mut->ThetaDIG<<" "<<DeuxNeMu<<" ";getchar();
//bins everything as a single deme
if (Size==0) return 1;
//all the gamma functions
for(int a=0;a<PopSampledOrNot;a++) {
	lnproba+=gammln(DeuxNuNm)-gammln(DeuxNuNm+PopEffectif[a]);
//if (Size==2) {cout<<exp(lnproba)<<endl;getchar();}
	for(int all=0;all<Sam_AlleleNbr;all++)
		{lnproba+=gammln(DeuxNuNmx+Effectif[a][all])-gammln(DeuxNuNmx);
//     if(std::isnan(lnproba)) {cout<<Mig->InvAsymptoNe<<" "<<Sam_AlleleNbr<<endl;
//     Print(Mut); getchar();}
// if (Size==2) {cout<<all<<" "<<AllEffectif[all]<<" "<<exp(lnproba)<<endl;getchar();}
    }
// the multinomial coeff
	s=PopEffectif[a];
for(int all=0;all<Sam_AlleleNbr;all++)
	{for(int jj=1;jj<=Effectif[a][all];jj++)
		{
		lnproba=lnproba+log((float)s)-log((float)jj); s--;
		}
	}
// if (Size==2) {cout<<exp(lnproba)<<endl;getchar();}
//cout<<exp(lnproba)<<endl;getchar();
}
return lnproba;
}

double CSample::LnPr_sample_from_coal_KAM(CMigMod *Mig,CMutMod *Mut) {
double lnproba=0;
double DeuxNeMu=Mut->ThetaDIG*(1+1/(Mig->limit_FSurUnMoinsF))*(Sam_AlleleNbr-1)/(Sam_AlleleNbr);
//cout<<DeuxNuNm<<endl;
double DeuxNeMux=DeuxNeMu/Sam_AlleleNbr;
// si utile : ces constantes pourraient etre calc quand Fsur... l'est
int s;
//cout<<endl<<Mut->ThetaDIG<<" "<<DeuxNeMu<<" ";getchar();
//bins everything as a single deme
if (Size==0) return 1;
//all the gamma functions
	lnproba+=gammln(DeuxNeMu)-gammln(DeuxNeMu+Size);
vector<int>AllEffectif(Sam_AlleleNbr,0);
// effectifs alleliques cumul['e]s
for(int a=0;a<PopSampledOrNot;a++)
	for(int all=0;all<Sam_AlleleNbr;all++) AllEffectif[all]+=Effectif[a][all];
//coefficients gamma alleliques
for(int all=0;all<Sam_AlleleNbr;all++)
	lnproba+=gammln(DeuxNeMux+AllEffectif[all])-gammln(DeuxNeMux);
//cout<<lnproba<<endl;
// the multinomial coeff
for(int a=0;a<PopSampledOrNot;a++) {
	s=PopEffectif[a];
	for(int all=0;all<Sam_AlleleNbr;all++) {
		for(int jj=1;jj<=Effectif[a][all];jj++) {
			lnproba=lnproba+log((float)s)-log((float)jj); s--;
			}
		}//all
// if (Size==2) {cout<<exp(lnproba)<<endl;getchar();}
//cout<<exp(lnproba)<<endl;getchar();
	}//a
return lnproba;
}



/* obsolete car use Nhaploid

double CSample::Pr_sample_from_coal_KAM(CMigMod *Mig,CMutMod *Mut) {
double lnproba=0;
double DeuxNeMu=Mut->ThetaDIG*Mig->InvAsymptoNe/(PSON*Nhaploid); // not direct but....
//cout<<endl<<Mut->ThetaDIG<<" "<<DeuxNeMu<<" ";getchar();
//bins everything as a single deme
if (Size==0) return 1;
vector<int>AllEffectif(Mut->Mut_AlleleNbr,0);
	for(int a=0;a<PSON;a++) for(int all=0;all<Mut->Mut_AlleleNbr;all++) AllEffectif[all]+=Effectif[a][all];
//all the gamma functions
lnproba+=gammln(DeuxNeMu)-gammln(DeuxNeMu+Size);
//if (Size==2) {cout<<exp(lnproba)<<endl;getchar();}
for(int all=0;all<Mut->Mut_AlleleNbr;all++)
	{lnproba+=gammln(DeuxNeMu/Mut->Mut_AlleleNbr+AllEffectif[all])-gammln(DeuxNeMu/Mut->Mut_AlleleNbr);
//     if(std::isnan(lnproba)) {cout<<Mig->InvAsymptoNe<<" "<<Mut->Mut_AlleleNbr<<endl;
//     Print(Mut); getchar();}
// if (Size==2) {cout<<all<<" "<<AllEffectif[all]<<" "<<exp(lnproba)<<endl;getchar();}
    }
// the multinomial coeff
int s=Size;
for(int all=0;all<Mut->Mut_AlleleNbr;all++)
	{for(int jj=1;jj<=AllEffectif[all];jj++)
		{
		lnproba=lnproba+log((float)s)-log((float)jj); s--;
		}
	}
// if (Size==2) {cout<<exp(lnproba)<<endl;getchar();}
return exp(lnproba);
}
*/

//#endif

string finalize_PCG_golubbool(int pson) {// This updates the solver using new info...
//cout<<IBDbool<<D1IBDbool<<PiSolverFromSettingsInfo;getchar();
       // PiSolverFromSettingsInfo is the value deduced from the information in settings (PiSolver+DemographicModel+MutationModel...)
    string locPiSolver=PiSolverFromSettingsInfo; //default
     // then we handle incomplete specifications in settings
     if (PiSolverFromSettingsInfo=="") {
         if (IBDbool) {
               // default linear solver chosen according to PSON
               if (pson > 59) {
                 if (D1IBDbool) locPiSolver="PCG_penta"; else locPiSolver="PCG_island";
              } else locPiSolver="SymDirect";
        }
    } else { // non-empty PiSolverFromSettingsInfo
           // only warnings
           if ((pson > 59) && cmp_nocase(PiSolverFromSettingsInfo.substr(0,3),"PCG")!=0) {
               cout<<"\n "<<pson<<" populations: letting PiSolver=PCG[...] might be useful.\n";
           }
           if ((pson < 60) && cmp_nocase(PiSolverFromSettingsInfo.substr(0,3),"PCG")==0) {
               cout<<"\n "<<pson<<" populations: PiSolver=PCG[...] might _not_ be useful.\n";
           }
    }
    // only setting golubbol
    if (cmp_nocase(locPiSolver.substr(0,3),"PCG")==0) {
              golubbool=true;
    } else golubbool=false;
return(locPiSolver);
}

int getParams(CSample* Sample1) {
using namespace NS_calclogl;

	if (multipleMarkersBool) {
		if (mutModelBool[GSM] && mutModelBool[ISM])
			MutParamNbr = 3;	// ThetaSEQ, pGSM and Theta(MIC)
		else if (mutModelBool[GSM])
			MutParamNbr = 2;	// // pGSM and Theta
		else if (mutModelBool[ISM])
			MutParamNbr = 2;	// ThetaSEQ and Theta(MIC)
	}
	else {
		if(mutModelBool[SNP])
			MutParamNbr = 0;	// pas de Theta (arbitrarily fixed to 1.0 or 0.0 later)
        else if (mutModelBool[GSM])
			MutParamNbr = 2;	// pGSM and Theta
		else
			MutParamNbr = 1;	// Theta
	}

	if(DemographicModel=="Npop") {
        if(Sample1->PopSampledOrNot==1) {
            //cout<<args.pars.size()<<" "<<MutParamNbr<<" "<<RelPopSizeParamNbr<<" "<<MigParamNbr<<" "<<ChangeTimeParamNbr<<" "<<RelAncPopSizeParamNbr<<" "<<AncThetaParamNbr;getchar();
    	   cout << "Wrong number of populations in sample file;" << endl;
    	   cout << "Check your input file for consistency between number of parameters and demographic model;" << endl;
    	   cout << "I exit." << endl;
    	   if (cinGetOnError) cin.get();
    	   exit(-1);
        }
		RelPopSizeParamNbr=Sample1->PopSampledOrNot-1;
    	MigParamNbr=Sample1->PopSampledOrNot*(Sample1->PopSampledOrNot-1);
     	ChangeTimeParamNbr=0;
     	RelAncPopSizeParamNbr=0;
     	AncThetaParamNbr=0;
		VarSizeShapeParamNbr=0;
    } else if(DemographicModel=="IM") {
			if(Sample1->PopSampledOrNot==1) {
				//cout<<args.pars.size()<<" "<<MutParamNbr<<" "<<RelPopSizeParamNbr<<" "<<MigParamNbr<<" "<<ChangeTimeParamNbr<<" "<<RelAncPopSizeParamNbr<<" "<<AncThetaParamNbr;getchar();
				cout << "Wrong number of populations in sample file;" << endl;
				cout << "Check your input file for consistency between number of parameters and demographic model;" << endl;
				cout << "I exit." << endl;
				if (cinGetOnError) cin.get();
				exit(-1);
			}
			RelPopSizeParamNbr=Sample1->PopSampledOrNot-1;
			MigParamNbr=Sample1->PopSampledOrNot*(Sample1->PopSampledOrNot-1);
			ChangeTimeParamNbr=Sample1->PopSampledOrNot-1;
			RelAncPopSizeParamNbr=Sample1->PopSampledOrNot-2;
			AncThetaParamNbr=Sample1->PopSampledOrNot-1;
			VarSizeShapeParamNbr=0;
	} else { // IBD or OnePop
    	RelPopSizeParamNbr=0;
        RelAncPopSizeParamNbr=0;
        if(DemographicModel.substr(0,6)=="OnePop") { //RL a simplifier
            MigParamNbr=0;
            if(DemographicModel=="OnePopVarSize") {
                ChangeTimeParamNbr=1;//le moment du changement demo, discret
                AncThetaParamNbr=1;//le theta ancestral ou le Nratio pour les SNP
				if(VarSizeFunction=="Logistic") {
					VarSizeShapeParamNbr=1;
					ChangeTimeParamNbr=1;//le d['e]but du changement demo
				} else if(VarSizeFunction=="Exponential") {
					VarSizeShapeParamNbr=0;
					ChangeTimeParamNbr=2;//le d['e]but du changement demo et sa dur['e]e
				} else if(VarSizeFunction=="Linear") {
					VarSizeShapeParamNbr=0;
					ChangeTimeParamNbr=2;//le d['e]but du changement demo et sa dur['e]e
				}
            } else if(DemographicModel=="OnePopFounderFlush") {
                AncThetaParamNbr=2;//le theta Founder et le theta ancestral
				if(VarSizeFunction=="Logistic") {
					VarSizeShapeParamNbr=1;
					ChangeTimeParamNbr=1;//le d['e]but du changement demo
				} else if(VarSizeFunction=="Exponential"){
					VarSizeShapeParamNbr=0;
					ChangeTimeParamNbr=2;//le d['e]but du changement demo et sa dur['e]e
				} else if(VarSizeFunction=="Linear") {
					VarSizeShapeParamNbr=0;
					ChangeTimeParamNbr=2;//le d['e]but du changement demo et sa dur['e]e
                } else if(VarSizeFunction=="Discrete") {
					cerr << "Discrete change in population size and OnePopFounderFlush are uncompatible." << endl;
                    cerr << "Change settings according to the user manual. I have to exit." << endl;
                    if(cinGetOnError) cin.get();
                    exit(-1);
                }
            } else {//OnePop
                    ChangeTimeParamNbr=0;
                    AncThetaParamNbr=0;
					VarSizeShapeParamNbr=0;
                }
        } else {//pour IBD
            MigParamNbr=2;
            ChangeTimeParamNbr=0;
            AncThetaParamNbr=0;
			VarSizeShapeParamNbr=0;
        }
    }
    Disp_Pars.resize(MigParamNbr);
	_PopSizeTab.resize(RelPopSizeParamNbr+1);
	_ChangeTimeTab.resize(ChangeTimeParamNbr);
    _AncPopSizeTab.resize(RelAncPopSizeParamNbr);
    _AncThetaTab.resize(AncThetaParamNbr);
#ifdef DEBUG
	if(DebugDG || PrintLikelihood) {
		//affichage des param[`e]tres non mis a l'['e]chelle globale
		cout<<args.pars.size()<<" unscaled Parameters :";
		ostream_vector(args.pars,cout);cout << endl;
	}
#endif
    //check parameter number according to demographic model:
    if(int(args.pars.size())!=(MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr+AncThetaParamNbr+VarSizeShapeParamNbr)) {
//cout<<args.pars.size()<<" "<<MutParamNbr<<" "<<RelPopSizeParamNbr<<" "<<MigParamNbr<<" "<<ChangeTimeParamNbr<<" "<<RelAncPopSizeParamNbr<<" "<<AncThetaParamNbr;getchar();
    	cout << "Wrong number of parameters;" << endl;
    	cout << "Check your input file for consistency between number of parameters and demographic model;" << endl;
    	cout << "I exit." << endl;
    	if (cinGetOnError) cin.get();
    	exit(-1);
    }
#ifdef LOCALSCALING
     //this is the case where the parameters manipulated by the user are scaled by deme size...
     // hence the immigration rate and mutation rates need to be rescaled for
     // further use in MakeMatrix (the Mab matrix)
	copy(args.pars.begin()+MutParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr,_PopSizeTab.begin());//copy les pop size
	Temp=0.0;
	for(int i=0;i<int(_PopSizeTab.size())-1;i++) Temp+=_PopSizeTab[i];
	_PopSizeTab[_PopSizeTab.size()-1]=1.0-Temp;
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr,Disp_Pars.begin());//copy les "points" migration
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr,_ChangeTimeTab.begin());//copy les points temps de divergence et/ou d['e]but du changement d['e]mo et/ou sa dur['e]e
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr,_AncPopSizeTab.begin());//copy les points Ancestral pop size
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr+AncThetaParamNbr,_AncThetaTab.begin());//copy les points Ancestral theta
	if(DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush") _VarSizeShape=args.pars[MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr+VarSizeShapeParamNbr];

    //ajuste diferents parametres : les taux de mig et theta car algorithme considere des taux globaux = Ntot*X et pas Ndeme*X, et le temps du second changement demo
	if(DemographicModel=="Npop" || DemographicModel == "IM") {
		//	CBR -> RL,FR commenting out the old code
		//	for(int ii=0;ii<MigParamNbr;ii++) Disp_Pars[ii]/=_PopSizeTab[(int) floor( (float) ii/(Sample1->PopSampledOrNot-1))];//on divise par les popsize pour avoir le taux global
		for (int ii = 0; ii < MigParamNbr; ii++)
			Disp_Pars[ii] = Disp_Pars[ii] / _PopSizeTab[ii];
		MutScaling=1;//ce n'est pas necessaire avec theta quand NPop car theta est toujours globale sinon impossible d'avoir un theta avec des tailles de pops diff['e]rentes...
	} else if(DemographicModel == "LinearIBD" || DemographicModel == "PlanarIBD") { //FR: correction BUG entroduite en fevrier 2014: == "IBD"
			Disp_Pars[0]*=Sample1->PopSampledOrNot;//equivalent car en IBD  : popsize = 1/Nbre de pops
			MutScaling=Sample1->PopSampledOrNot;
	} else if((DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush") && (VarSizeFunction=="Exponential" || VarSizeFunction=="Linear")) _ChangeTimeTab[1]+=_ChangeTimeTab[0]; //add the duration of the event to its starting time to obtain a second absolute time

	if (multipleMarkersBool) {
		if (mutationModel == "GSM") {
			if (mutModelBool[ISM]) {
				_pGSM = args.pars[1];	//copy pGSM
				_Theta = args.pars[2] * MutScaling;	//copy le theta
			}
			else {
				_pGSM = args.pars[0];	//copy pGSM
				_Theta = args.pars[1] * MutScaling;	//copy le theta
			}
		}
		else if (mutationModel == "SMM") {
			if (mutModelBool[GSM] && mutModelBool[ISM])
				_Theta = args.pars[2] * MutScaling;	//copy le theta
			else
				_Theta = args.pars[1] * MutScaling;	//copy le theta
		}
		else {	// ISM
			if (mutModelBool[GSM]) {
				_Theta = args.pars[0] * MutScaling;	//copy le theta(SEQ)
				_ThetaMIC = args.pars[2] * MutScaling;  // used to assign MutMod1->AncThetaDIG}
			}
			else {
				_Theta = args.pars[0] * MutScaling;	//copy le theta(SEQ)
				_ThetaMIC = args.pars[1] * MutScaling;  // used to assign MutMod1->AncThetaDIG}
			}
		}
	}
	else {
		if (mutationModel == "SNP")
			_Theta = 0.0;	//pas de theta en SNP, remplacé par 1.0 dans certaines formules pour assurer la mutation unique qui peut arriver
		else  if (mutationModel == "GSM") {
			_pGSM = args.pars[0];	//copy pGSM
			_Theta = args.pars[1] * MutScaling;	//copy le theta
		} else _Theta = args.pars[0] * MutScaling;	//copy le theta
	}

#ifdef DEBUG//affichage des param[`e]tres mis a l'[‘e]chelle ou non plus loin
	if(PrintLikelihood) cout<<args.pars.size()<<" scaled parameters from localscaling input :"<<endl;
#endif

#else //!localscaling
	copy(args.pars.begin()+MutParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr,_PopSizeTab.begin());//copy les pop size
	Temp=0.0;
	for(int i=0;i<_PopSizeTab.size()-1;i++) Temp+=_PopSizeTab[i];
	_PopSizeTab[_PopSizeTab.size()-1]=1.0-Temp;
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr,Disp_Pars.begin());//copy les "points" migration
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr,_ChangeTimeTab.begin());//copy les points temps de divergence
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr,_AncPopSizeTab.begin());//copy les points Ancestral pop size
	copy(args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr,args.pars.begin()+MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr+AncThetaParamNbr,_AncThetaTab.begin());//copy les points Ancestral theta
	_VarSizeShape=args.pars[MutParamNbr+RelPopSizeParamNbr+MigParamNbr+ChangeTimeParamNbr+RelAncPopSizeParamNbr+VarSizeShapeParamNbr];
	if((DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush") && (VarSizeFunction=="Exponential" || VarSizeFunction=="Linear")) _ChangeTimeTab[1]+=_ChangeTimeTab[0]; //add the duration of the event to its starting time to obtain a second absolute time
	if (mutationModel == "SNP")
        _Theta=0.0;//pas de theta pour SNP, remplace par 1.0 dans certaines formules pour assurer la mutation unique qui peut arriver
    else if(gsmbool)   {
		_Theta=args.pars[1];//copy le theta
		_pGSM=args.pars[0];//copy pGSM
	} else _Theta=args.pars[0];//copy le theta



#ifdef DEBUG
if(PrintLikelihood) cout<<args.pars.size()<<" scaled parametres from globalscaling input :"<<endl;
#endif

#endif//end ifdef localscaling else

#ifdef DEBUG//affichage des param[`e]tres mis a l'[‘e]chelle ou non plus loin
if(PrintLikelihood)  {
    //affichage des points Theta, PopSize, Change/divTimes, Mig, ancTheta

    if (mutationModel == "SNP") {
        cout<<"One pseudo Theta : " << _Theta << endl;
    } else if (mutationModel=="GSM") {
        cout<<"pGSM=" << _pGSM << "and One Theta : " << _Theta << endl;
    } else {
        cout<<"One Theta : " << _Theta << endl;
    }

    if(_PopSizeTab.size()>1) {
        cout<<_PopSizeTab.size() -1.0 <<" PopSize point(s) are: ";
        ostream_vector(_PopSizeTab,cout);
        cout << endl;
    }
    if(Disp_Pars.size()>0) {
        cout<<Disp_Pars.size()<<" Mig points are: ";
        fflush(stdout);
        ostream_vector(Disp_Pars,cout);
        cout << endl;
    }
    if(DemographicModel=="IM") {
        cout<<_ChangeTimeTab.size()<<" Divergence Time points are: ";
        ostream_vector(_ChangeTimeTab,cout);
        cout << endl;
        if(Sample1->PopSampledOrNot>2) {
            cout<<_AncPopSizeTab.size()<<" Ancestral PopSize points are: ";
            ostream_vector(_AncPopSizeTab,cout);
            cout << endl;
        }
        cout<<_AncThetaTab.size()<<" Ancestral Theta points are: ";
        ostream_vector(_AncThetaTab,cout);
        cout << endl;
    }
    if(DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush") {
        cout<<_ChangeTimeTab.size()<<" Demographic change Time point(s) are: ";
        ostream_vector(_ChangeTimeTab,cout);
        cout << endl;
        if (mutationModel == "SNP") cout<<_AncThetaTab.size()<<" Ancestral pseudo Theta point(s) ( = Nratio) are: ";
         else cout<<_AncThetaTab.size()<<" Ancestral Theta point(s) are: ";
        ostream_vector(_AncThetaTab,cout);
        cout << endl;
        if(VarSizeFunction=="Logistic") cout << "and Logistic Growth Rate equals to : " << _VarSizeShape << endl;
    }
    cout << endl;
    fflush(stdout);

}
#endif

return(0);
} // end getParams()


/**********************************************************************************/
/*******************           CALCLOGL ( -ln(L) for one locus )  *****************/
vector<float> calclogl(const Arguments args,CMigMod *MigMod1) {
using namespace NS_calclogl;

//the type of arguments if always float so template cannot be used here
// only one definition of calclogl is given so explicit instantiations cannot be used

string String,tempfilename;
string _SampleFileName=args.SampleFileName;
localallelenbr=0;
long double TempLnLik;
vector<long double>Likvec(mloglikdim),SumL(mloglikdim,0),MeanL(mloglikdim);
vector<long double>TMRCAvec(mloglikdim),SumTMRCA(mloglikdim,0),MeanTMRCA(mloglikdim);//pour calcul ds TMRCA
vector<long double>LikPerTreevec;
changetype TempChangeResult;
MutScaling=1;
bool pacChangebool=false;//if changes occured in FctnPtr (PiCalc and IS/PAC) and Mig and Mut models in the split model
Changedbool=false;

CMutMod *MutMod1;
CSample *Sample1,*CurrentSample;
CNexus nexusSample;
int rdeja;
static string PiSolverFromData;

#ifdef NGIVEN
	static CN_known N_known=CN_known();
#endif

    //pour genepop files iLoc est accessible dans GetSample et GivenK peut alors etre lu dedans
	//la lecture de fichier genepop SMM ne marchait pas sans ca
    if(true) {// was if(readAdHocFileBool)  //!! 14/01/11
               if(sequenceGivenK.size()>0) localallelenbr=sequenceGivenK[std::min(args.locus-1,int(sequenceGivenK.size())-1)];
               else localallelenbr=0;
    }

    if (firstCountPSON) { // if first call to calclogl for first locus of dataset; then we determine solver etc. (one for each multilocus sample)
       int locPSON;
       if (readAdHocFileBool) {

         { // ici pb for ad hoc files uniquement, on determine le nb de pop par une lecture bidon
           CSample *DummySample;
           DummySample=new CSampleWout();
           if(0!=DummySample->GetSample(_SampleFileName,localallelenbr)){ //reads PopSampledOrNot here
    		     cout << "\n Can't read sample file (1) : " << _SampleFileName << ". Exit." << endl;
    		     if (cinGetOnError) cin.get();
    		    exit(-1);
           }

           // * en 1D SANS fichier Genepop, on n'a pas encore execut['e] geoInfo.fill(...), on le fait ici
           locPSON=DummySample->PopSampledOrNot;
           if (D1IBDbool) {
                 geoInfo.fill(locPSON); // D1 && NOT Genepop (quick patch)
           }
           delete DummySample;
         } //block DummySample

    } else { //Genepop file => ProcesssedLoci has been filled
        locPSON= (int) ProcessedLoci[args.locus-1].alleleCounts.size();//CM->RL 23.01.14
    } // endif adhoc else Genepop
    PiSolverFromData=finalize_PCG_golubbool(locPSON);
  }


// now we have set golubbool so
  if (golubbool) Sample1=new CSampleWithGolub(); else Sample1=new CSampleWout();

  // no we have allocated Sample1 so
    if (readAdHocFileBool) {
  	  if(0!=Sample1->GetSample(_SampleFileName,localallelenbr)){ //evaluates PopSampledOrNot here if not previously done
  		  cout << "\n Can't read sample file (2) : " << _SampleFileName << ". Exit." << endl;
  		  if (cinGetOnError) cin.get();
  		  exit(-1);
  	  }
    } else {
  	  Sample1->GetSample(&(ProcessedLoci[args.locus-1]),localallelenbr,alleleNbrChecked);
    }

  Sample1->locusNum=args.locus;

  if (mutationModel == "ISM") {
	  if (firstCallOfLocus[args.locus - 1]) {
		  // checks for and reads the locus-specific NEXUS file
	      nexusSample.cleanNEXUSFile(args.locus);
	      nexusSample.readNEXUSFile(Sample1);
	      if (nexusSample.hapCount != ProcessedLoci[args.locus-1].alleleCounts[0].size()) {
				cerr << "\n(!) Inconsistency at locus number " << args.locus << " (name : " << ProcessedLoci[args.locus - 1].locusName
					 << ") between the specified number \nof allelic states in the genepop file (allele count = " << ProcessedLoci[args.locus-1].alleleCounts[0].size()
                     <<  ") and their corresponding haplotype sequences (hapCount = " <<  nexusSample.hapCount << ")."
					 << "\nPlease cross-check the Genepop file : " << ProcessedLoci[args.locus - 1].fileName
					 << "\nand the Nexus file : " << nexusSample.fileName << " respectively.\n"
					 << "\nAborting Migraine...Press any key to exit.\n" << endl;
				if (cinGetOnError)
					cin.get();
				exit(-1);
	      }

	      // Checks if the sequence data is compatible with the ISM assumptions and suggests modifications to the user
	      nexusSample.checkIfDataISMCompatible(Sample1);

	      // keeps a copy of the CSample info for the present sequence locus
	      storeSeqCSamples[args.locus] = nexusSample.CopyCatNexus;

	      firstCallOfLocus[args.locus - 1] = false;
	  }
	  else
		  // restores the stored CSample copy of the corresponding sequence lous
		  storeSeqCSamples[args.locus].getCSample(Sample1);
  } else   if (mutationModel == "SNP") {
	  if (firstCallOfLocus.size()==0 or firstCallOfLocus[args.locus - 1]) {
		  // fill the haplotype tables without reading Nexus file
	      nexusSample.FillHaplotypeInfoFromSNP(Sample1);
	      if (nexusSample.hapCount != ProcessedLoci[args.locus-1].alleleCounts[0].size()) {
              cerr << "\n(!) Inconsistency at locus number " << args.locus << " (name : " << ProcessedLoci[args.locus - 1].locusName
              << ") between the specified number \nof allelic states and their corresponding haplotype sequences."
              << "\nPlease cross-check the Genepop file : " << ProcessedLoci[args.locus - 1].fileName
              << "\nand the Nexus file : " << nexusSample.fileName << " respectively.\n"
              << "\nAborting Migraine...Press any key to exit.\n" << endl;
              if (cinGetOnError)
                  cin.get();
              exit(-1);
	      }

	      // Checks if the sequence data is compatible with the ISM assumptions and suggests modifications to the user
	      nexusSample.checkIfDataISMCompatible(Sample1);

	      // keeps a copy of the CSample info for the present sequence locus
	      storeSeqCSamples[args.locus] = nexusSample.CopyCatNexus;

//	      if (firstCallOfLocus.size()<=args.locus) {firstCallOfLocus.push_back(false);firstCallOfLocus.push_back(true);}//RL a revoir
//              else
          if(firstCallOfLocus.size()>0) firstCallOfLocus[args.locus - 1] = false;
	  }
	  else
		  // restores the stored CSample copy of the corresponding sequence lous
		  storeSeqCSamples[args.locus].getCSample(Sample1);
  } else if(multipleMarkersBool)
      firstCallOfLocus[args.locus - 1] = false;

// no we have read the new sample so
    if (firstCountPSON) { // if first call to calclogl for first locus of dataset; then we determine solver etc.
       firstCountPSON=false; // do not go twice through this code
       firstPSON=Sample1->PopSampledOrNot;
    } else if (Sample1->PopSampledOrNot!=firstPSON) {
        cout << "\n Inconsistent number of populations (sampled or not) among loci. Exit." << endl;
		cout << "firstPSON=" << firstPSON <<  " and Sample1->PopSampledOrNot=" << Sample1->PopSampledOrNot << endl;
        if (cinGetOnError) cin.get();
        exit(-1);
  } //endif firstCountPSON else

// this could come earlier
	PicalcfnPtr=setPiCalcFnPtr(PiSolverFromData);

    Sample1->checkOptions();  // checks compatibility between different options

#ifdef DEBUG
    if(PrintLikelihood) Sample1->Print();
#endif

    getParams(Sample1);
	Sample1->setMigMod(MigMod1);
	MutMod1=new CMutMod();
	Sample1->setMutMod(MutMod1,MigMod1);

    if (calculQused && !(DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush")) { //IBDbool-> symmetric KAM, for OnePopVarSize & OnePopFounderFlush & IM  this intialization is done once the ancestral stable population is reach (in demographicChange())
       // preparation for computation of probabilities of identity by standard matrix algebra
      // Here only those computations needed only once for a given parameter point.
      if (calculQ.newPars()) {
         calculQ.setNh(MigMod1,MutMod1);
         calculQ.initMig(MigMod1);
         calculQ.eigenanalysis();
         picalctype mufactor=MutMod1->Mut_AlleleNbr/(MutMod1->Mut_AlleleNbr-1.); // symmetric KAM
         calculQ.initMut(MutMod1,mufactor,mutationModel);
         calculQ.newPars(false); // prevents further recomputation until a new call to externalfn()
      }
	}
	calcQibdtableDoneBool=false;

	if(DemographicModel!="IM")    Sample1->new_L_etc();//pas possible ici en IM a cause de changements temporels changeant le nombre de pops
    long double LnMnomialcoeff=Sample1->LnMultinomialCoeff();
    if (golubbool) CurrentSample=new CSampleWithGolub(Sample1);
     else CurrentSample=new CSampleWout(Sample1);
    //  Checks whether some part of the computation has been saved in a file
    rdeja=0;
    if (runsTempFileBool) { // looks whether part of the computation was saved
            stringstream strstr;
            strstr<<_SampleFileName<<"_"<<args.point<<".tmp";
            strstr>>tempfilename;
            ifstream tempfile(tempfilename.c_str());
            if (tempfile.is_open()) { // some computations were saved
                tempfile>>rdeja;
                for (vector<long double>::iterator ii=SumL.begin();ii<SumL.end();ii++)
    		        tempfile>>(*ii);
                rdeja++;
                tempfile.close();
       		    std::cout<<"Restart at run "<<rdeja<<" from file "<<tempfilename<<endl<<flush;
            } else tempfile.clear(); //not anormal
    }

    if(UsePCL_SISRbool){
        //calcul avec resampling =  RL + CM + PP 07112013
      //cout << "PCL_SISR_UseIndivNbResamplingbool = " << PCL_SISR_UseIndivNbResamplingbool << endl; // CM 25.03.14
        LnMeanLikelihood=Calclogl_PCL_SISR(Sample1->Effectif,Sample1->Sam_AlleleNbr,_Theta, _AncThetaTab[0], _ChangeTimeTab[1],
                                args.nruns,PCL_SISR_UseIndivNbResamplingbool,PCL_SISR_IndivNbResampling, PCL_SISR_MaxEventNbInPopAnc, PCL_SISR_ESSMinFall, PCL_SISR_StartESSAtEvent, PCL_SISR_Alpha, PCL_SISR_Beta,JobMin); //CM 13.02.14  // CM 25.03.14
      //cout << " ICI " << JobMin << endl; //CM 23.04.14
      //cout << " Appel de Calclogl_PCL_SISR, args.point = " << args.point<< " args.locus = "<< args.locus <<endl;// CM 12.03.14"

    } else {//else du if(UsePCL_SISR)

//Calcul classique sans resampling

//  BOUCLE SUR NRUNS...
LikPerTreevec.resize(args.nruns,0);
for(long int r=rdeja;r<args.nruns;r++){ // MAIN ITERATION


    //alea.seed(_EstimationSeed);
    //cout << "alea.seed(_EstimationSeed); pour avoir tuojours les memes séquences a chaque arbre; a virer absolument" << endl;

         if (ancestralSuffixes.size()>0) beforeLastEvent=true; else beforeLastEvent=false;
         ancEventIt=0; //iterator over ancestral events
    	if(DemographicModel=="IM") Sample1->new_L_etc();//pour changement demo impliquant un changemet du nombre de pops
    	if(pacancbool) ChangefnPtr=&CSample::Change;
    	// the condition below allows us to use m'sat and seq markers irrespective of the IS statistic used for the m'sats
    	if (mutationModel == "ISM" || mutationModel == "SNP") { //RL adding SNP is probably not necessary unless we use multipleùmarkers with SNP (not planned now 11/03/20143)
        	if (HobolthISM)
        		ChangefnPtr = &CSample::calcHobolthChange;
        	else
        		ChangefnPtr = &CSample::Change;
    	}
    	*CurrentSample=*Sample1; //this also resets generation,disp_coal_events...

        if(mutationModel=="SMM" or mutationModel=="GSM") {//SMM or GSM // RL necessaire ici car reinitialisation de
            CurrentSample->GetAllelicRange(0);
            CurrentSample->GetAllelicRange(1);
        }

        if(Changedbool) {
    //if a split or any demographic change occured, we have to reinitialize demographic model and Pis computation because it could have been changed in a previous run
			pacChangebool=false;
            MutMod1->ThetaDIG=_Theta*MutMod1->ThetaFactor;
			if(DemographicModel=="IM") {
                MigMod1->PopMergedNbr=0;
			    MigMod1->MakeMatrixNpop(CurrentSample->PopSampledOrNot,Disp_Pars);
			    MigMod1->MakePopSize(_PopSizeTab);
            }
			MigMod1->ResetChangeTime(_ChangeTimeTab);
			PicalcfnPtr=setPiCalcFnPtr(PiSolverFromData);
			Changedbool=false;
			if(piOnePopSDbool) CurrentSample->CalcMatSD(MutMod1,MigMod1,MutMod1->ThetaDIG,false);//pour PiOnePopSD in the actual population
        }
#ifdef DEBUG
		/*if(DebugSplit) {
			cout << "\n****Start Run " << r << " : resettings Initial Configuration because a change occured..." << endl;
			MigMod1->PrintMig();
			MigMod1->PrintPopSize();
			cout << "ThetaDIG=" << MutMod1->ThetaDIG << endl;
			cout << "Algorithm:" ;
			if (sequenceStat.size()==1) {
   				if (cmp_nocase(sequenceStat[0],"PAC")==0) cout << "PAC-likelihood" << endl;
					else cout << "Importance Sampling" << endl;
			}
			if (sequenceStat.size()>1) {
				cout <<"Sequence =  ";
			   	for(int i=0;i<sequenceStat.size();i++) cout << sequenceStat[i] ;
			   	cout << endl;
			}
			if (PicalcfnPtr==&CSample::PiCalcPimDG2002) cout << "PiPim version" << endl;
					else if (PicalcfnPtr==&CSample::PiCalcOnePop) cout << "PiOnePop version" << endl;
                            else if (PicalcfnPtr==&CSample::PiCalcTwoPopSmmDG2002) cout << "PiTwoPopSMM version" << endl;
			cout << "****End of configuration after resettings\n" << endl;
		}*/
#endif

    	if (pacbool) LnLikelihood=LnMnomialcoeff; else LnLikelihood=0.0;
     //init_file_et_critere est infini si pas Asympto et scaled_generation doit rester zero sauf en Asympto
     //cout<<MigMod1->init_file_et_critere;getchar();
    	while(CurrentSample->Size>0 && (CurrentSample->scaled_generation<=MigMod1->init_file_et_critere)){
    // nota certaines valeurs de Change sont >1, seul leur produit est <1
    //call through pointer to member function, see Primer+ p. 1071
    // points to PACpi or to Change

    /************ here the core computation ********************************/
             if (CurrentSample->Size==2 && calculQused && mutationModel!="ISM" && mutationModel!="SNP" && //CBR (12042011) RL (11/03/2014) ISM & SNP condition needs to be reviewed
                 // additionally check whether we are in the stationary ancestral population for time-inhomogeneous models.
				 (!(DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush"  || DemographicModel=="IM") || MigMod1->InAncStablePopbool)  ) {

                //check if the last two genes are of different types in the same deme
//                 explications :
//                 * les deux gènes sont identiques dans le même dème
//                 => calculQ  dans computeIIS() produit iis/'K', c'est bien la proba de l'échantillon "le
//                 premier gène de type 'a" et le deuxième de type 'a'"
//
//                 * les deux gènes sont différents dans le même dème. Le test (d1==d2) dans computeIIS() est
//                 vrai donc on calcule pr*=2.;
//                 => calculQ produit 2*(1-iis)/(K(K-1)), il faut rediviser par deux pour
//                 avoir la séquence "le premier gène de type 'a" et le deuxième de type 'b'"
//                 (car pour K alleles, on a K(K-1) sequences de ce type, par K(K-1)/2)
//
//                 Ensuite pour des dèmes différents :
//
//                 * les deux gènes sont identiques dans deux dèmes différents
//                 => calculQ produit iis/'K', c'est bien la proba de l'échantillon "le
//                 premier gène de type 'a" et le deuxième de type 'a'" (iis/K n'est pas la
//                 proba de ["1e gène de type 'a" et 2e de type 'a'" plus l'inverse])
//
//                 * les deux gènes sont differents dans deux dèmes différents. Le test
//                 (d1==d2) est faux donc cette foix calculQ a calculé (1-iis)/(K(K-1))
//                 c'est bien la proba de l'échantillon "le premier gène de type 'a" (1/K) et
//                 le deuxième de type 'b'(1/(K-1))". La séquence distincte "le premier gène
//                 de type 'b" et le deuxième de type 'a'" serait un autre terme de proba
//                 (1-iis)/(K(K-1)).
//                 => il ne faut rien rediviser.

                bool DiffAllWithinPop=false;
                if((pacbool or pacancbool)) for(int a=0;a<CurrentSample->PopSampledOrNot;a++) {
                    if(CurrentSample->PopEffectif[a]==2) { //intrapop
                         for(unsigned int i=0;i<CurrentSample->Effectif[a].size();i++) {
                             if (CurrentSample->Effectif[a][i]==1) {
                                 DiffAllWithinPop=true;
                                 break;
                             }
                         }
                     }
                     if(DiffAllWithinPop) break;
                 }

                 //CurrentSample->Print();

                 if (C_c_bool) cerr<<endl<<"call to computeIIS(...)";
                 TempChangeResult=CurrentSample->computeIIS(&calculQ,MutMod1); // computes identity in state for the pair of genes
                 //cout << "ln(computeIIS)=" << log(TempChangeResult) << endl;

                 //correction when the two last genes are not identical and in the same deme, likelihood is half the prob IIS
                 //only for PAC
                 if((pacbool or pacancbool) and DiffAllWithinPop) {
                     TempChangeResult/=2;
                     //cout << "(pac or pacanc) and foundDiffAll ln(TempChangeResult)/=2" <<  endl;
                 }
                 //cout << "ln(TempChangeResult)=" << log(TempChangeResult) <<  endl;
                 TempChangeResult/=CurrentSample->Sam_AlleleNbr;
                 //cout << "ln(1/Sam_AlleleNbr)=" << log(1.0/CurrentSample->Sam_AlleleNbr) << endl;
                 //cout << "ln(TempChangeResult/Sam_AlleleNbr)=" << log(TempChangeResult) << endl;

                 if (C_c_bool) {cerr<<endl<<"returned from call to computeIIS(...)";}

				  //ajout RL pour problemes de calculs numériques en GSM dans computeIIS() = probleme de somme alternées + et - dans la transformé de Fourier inverse
				  if(mutationModel=="GSM" && TempChangeResult<=0.0) {
					  if (C_c_bool) cerr<<endl<<"call to CurrentSample->*ChangefnPtr(...)";
					  TempChangeResult=(CurrentSample->*ChangefnPtr)(MigMod1,MutMod1);//IS or PAC
					  //cout << "Change Result=" << TempChangeResult << "; ln(TempChangeResult)=" << log(TempChangeResult)<< endl;
					  if (C_c_bool) {cerr<<endl<<"returned from call to CurrentSample->*ChangefnPtr(...)";}
				  }
				  if(mutationModel!="GSM" && TempChangeResult<=0.0) {
					  cerr << "(!) computeIIS() returned " << TempChangeResult << "<=0.0, contact the authors or rerun without 2id options" << endl;
                      cerr << "I must exit." << endl;
					  if(cinGetOnError) cin.get();
					  exit(-1);
				  }
                 //cout << "Final (IIS) TempChangeResult=" << TempChangeResult << "; ln(TempChangeResult)=" << log(TempChangeResult) <<  endl;

            } else {
                  if (C_c_bool) cerr<<endl<<"call to CurrentSample->*ChangefnPtr(...)";
                TempChangeResult=(CurrentSample->*ChangefnPtr)(MigMod1,MutMod1);//IS or PAC
                //cout << "Change Result=" << TempChangeResult << "; ln(TempChangeResult)=" << log(TempChangeResult)<< endl;
                  if (C_c_bool) {cerr<<endl<<"returned from call to CurrentSample->*ChangefnPtr(...)";}
            }
    /***********************************************************************/

            if(TempChangeResult==-1) {
                    cout << "unexpected bug in Change function, end of the program." << endl;
                    if(cinGetOnError) cin.get();
                    exit(-1);
            } else TempLnLik=log(TempChangeResult);
            if (std::isnan(TempLnLik)) {
                cout<<"TempLnLik=log("<<TempChangeResult<<") is NaN..." << endl;
                if(cinGetOnError) cin.get();
            }

            //if demographic change


    		if((DemographicModel=="IM" || DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush") && !std::isinf(MigMod1->MinChangeTime)
                     && CurrentSample->scaled_generation>=MigMod1->MinChangeTime) {//RL : to be replaced by !Mig->InAncStablePopbool
                 CurrentSample->DemographicChange(MigMod1,MutMod1);

    		} else if (beforeLastEvent && CurrentSample->scaled_generation>ancestralScaledTimes[ancEventIt]) {// other event

    		    CurrentSample->EssaiSequentialSample();
            } else {//no event occurred
//cout<<beforeLastEvent<<" "<<CurrentSample->scaled_generation<<" "<<ancEventIt<<" ";
//ostream_vector(ancestralScaledTimes,cout);getchar();
#ifdef DEBUG
    			if(DebugDG) {
                    cout << "LnLikelihood before(+=TempLnLik) is" << LnLikelihood << endl;
                    cout << "Likelihood before(+=TempLnLik) is" << exp(LnLikelihood) << endl;
                }
#endif

    		 	LnLikelihood+=TempLnLik;   // i.e. sum of log() over steps in the tree
//cout<<" "<<CurrentSample->Size<<" "<<TempLnLik<<" "<<LnLikelihood;getchar();

				//pour passer en PACanc si on est dans la zone "stationnaire proche de ThetaAnc" de la logistique, les autres passages en PACanc se font dans DemographicChange
				if(VarSizeFunction=="Logistic" && pacancbool && !pacChangebool && fabs((MigMod1->PopSize[0]-MutMod1->AncThetaDIG/MutMod1->ThetaDIG)/(MutMod1->AncThetaDIG/MutMod1->ThetaDIG))<0.01) {
					if (multipleMarkersBool && (mutationModel == "ISM"))
						MutMod1->ThetaDIG=(1.0*MutMod1->ThetaFactor)*(_AncThetaTab[0] * _Theta / _ThetaMIC); // Here ThetaDIG for ISM needs to be rescaled when multipleMarkersBool=true
					else
						MutMod1->ThetaDIG=(1.0*MutMod1->ThetaFactor)*_AncThetaTab[0];
					MigMod1->PopSize[0]=_PopSizeTab[0];
					if(piOnePopSDbool) CurrentSample->CalcMatSD(MutMod1,MigMod1,MutMod1->ThetaDIG,false);//pour PiOnePopSD in the ancestral population
					//if(mutationModel=="SMM" or mutationModel=="GSM") {//SMM or GSM// RL sans doute aps necessaire si on le fait apres chaque mutation.
					//	CurrentSample->GetAllelicRange(MigMod1,MutMod1,0);
					//	CurrentSample->GetAllelicRange(MigMod1,MutMod1,1);
					//}
					ChangefnPtr=&CSample::PACpi;
					//cout << "LnLikelihood before switching to PAC = " << LnLikelihood << endl;
					//cout << "LnMultinomialCoeff->" << CurrentSample->LnMultinomialCoeff() << endl;
					LnLikelihood+=CurrentSample->LnMultinomialCoeff();//for PAC on ancestral population
					//cout << "LnLikelihood after addition of LnMultinomialCoeff for PAC = " << LnLikelihood << endl;
					pacChangebool=true;
				}


//cout<<"\n"<<CurrentSample->Size<<" "<<TempLnLik<<" "<<LnLikelihood; //frfrfrfr

			//cout << "LnLikelihood+=TempLnLik" << TempLnLik << "->LnLikelihood=" << LnLikelihood << endl;
#ifdef DEBUG
    			if(DebugDG){
                    //cout << "LnLikelihood after adding TempLnLik =" << TempLnLik << ")->" << LnLikelihood << endl;
                    //cout << "Likelihood after adding TempLnLik =" << exp(TempLnLik) << ")->" << exp(LnLikelihood) << endl;
                    cout << endl;
    				if(CurrentSample->Size>0){
    					cout << "-> New ";
    					CurrentSample->Print();
    				}
    			}
#endif

        		if (C_c_bool) {
        		   cout<<"File: "<<args.SampleFileName<<"\n Parameters: ";
        		   ostream_vector(args.pars,cout);
        		   cout<<"\nRun: "<<r<<" ";
        		   cout<<"Size: "<<CurrentSample->Size<<" ";
        		   cout<<"Scaled time: "<<CurrentSample->scaled_generation<<" ";
        		   cout<<"LnLik: "<<LnLikelihood<<"\n";
        		      for(int a=0;a<CurrentSample->PopSampledOrNot;a++) {
        			 for(int all=0;all<CurrentSample->Sam_AlleleNbr;all++)
        				cout<<CurrentSample->Effectif[a][all]<<" ";
        			 cout<<endl;
        		      }
        		   cout<<"*** Note Control-c interrupt disabled; use Control-z instead. ***\n";
//    getchar();
        		   C_c_bool=false; //inactiver cet affichage jusqu'au prochain [C-c... => call C_c_handler=> C_cbool=true
        		}//end if (C_c_bool)
        	}//end if no demographic change else if ... else ...
         }//end while(CurrentSample->Size>0&&CurrentSample->scaled_generation<MigMod1->init_file_et_critere)

// Likvec contain the single tree likelihood
    	 if (asymptoBool) {
    // Estimations pour les 2 approximations pour chaque ancestral tree
    		Likvec.front()=exp(LnLikelihood+CurrentSample->LnPr_sample_from_coal_KAM(MigMod1,MutMod1));
    		Likvec.back()=exp(LnLikelihood+CurrentSample->LnPr_sample_from_island_KAM(MigMod1,MutMod1));
    	} else {
            (*Likvec.begin())=exp(LnLikelihood);
    		(*TMRCAvec.begin())=CurrentSample->scaled_generation;
    }
#ifdef DEBUG
    if(CurrentSample->Size>0){
        cout << "Problem : sample size is not nul at the end of the run but is" << CurrentSample->Size << endl;
        if (cinGetOnError) cin.get();
        exit(-1);
    }
	if(PrintLikelihood) {
      cout << setprecision(10) << "Point "<<args.point<<" locus "<<args.locus<<", Tree " << r << " gives Likelihood = " << exp(LnLikelihood) << ";" ;
      cout << setprecision(10) << "and LnLikelihood = " << LnLikelihood << ";" ;
		if(timeNeededBool || TMRCABool) {
			cout << setprecision(10) <<" & TMRCA=" << *TMRCAvec.begin()<< endl;
			ofstream TMRCAFile("TMRCA.txt",ios::out|ios::app);
			if(!TMRCAFile.is_open()) {
				cerr << "Unable to open file TMRCA.txt in calclog();" << endl;
				if (cinGetOnError) cin.get();
				TMRCAFile.clear();
				exit(-1);
			}
			TMRCAFile << setprecision(10) << *TMRCAvec.begin() << endl;
		} else cout << endl;
      if(DemographicModel=="IM" && Changedbool)
            cout << "\tThis tree has "<< ostream_vector(LineagesPerPopAtDivTime,cout)<< " lineages in each population at Divergence time" << endl;
       fflush(stdout);
       cout << "\n********************************" << endl;
       cout << "\n********************************\n" << endl;
    }
    if(PrintSuiviLikelihood) {
    	ofstream Suivi("SuiviL.txt", ios::out | ios::app);
    	if (!Suivi.is_open()) {
    		cerr << "Unable to open file SuiviL.txt in calclog();" << endl;
    		if (cinGetOnError)
    			cin.get();
    		Suivi.clear();
    		exit(-1);
    	}
    	Suivi << setprecision(15) << exp(LnLikelihood) << endl;
    }
#endif
    // ci dessous on somme les estimations (Likvec pour chaque ancestral tree) dans SumL

    //(chaque element [i] de SumL est sum de l'element [i] des vecteurs Likvec successifs)
    	{vector<long double>::iterator aa=Likvec.begin();
    		for (vector<long double>::iterator ii=SumL.begin();ii<SumL.end();ii++,aa++)
    		(*ii)+=(*aa);
		//cout<<r<<" ";ostream_vector(SumL,cout);cout<<endl;//getchar();////////////////
		vector<long double>::iterator bb=TMRCAvec.begin();
		for (vector<long double>::iterator jj=SumTMRCA.begin();jj<SumTMRCA.end();jj++,bb++)
		(*jj)+=(*bb);
    	} // block aa local
        if (runsTempFileBool) { // saves current state of computation in temp file
            ofstream tempfile(tempfilename.c_str());
            if (tempfile.is_open()) {
                tempfile<<r;
                for (vector<long double>::iterator ii=SumL.begin();ii<SumL.end();ii++)
    		        tempfile<<" "<<(*ii);
                tempfile.close();
            } else {
               cerr<<"Unable to open file "<<tempfilename<<" for writing"<<endl;
               if (cinGetOnError) cin.get();
               exit(-1);
            }
        }

        if(DemographicModel=="IM") {if(!Changedbool) Sample1->delete_L_etc(); else CurrentSample->delete_L_etc();}


    }//fin boucle sur repet
}// fin else du if(UsePCL_SISR)


// si on enl[`e]ve ca il ne saura plus quels sont les loci achev['e]s.
//    if (runsTempFileBool) remove(tempfilename.c_str());

    if(piOnePopSDbool || piOnePopEigenbool) {
	//cout << "LA delete MatriceSD" << endl;
    	for(int i=0;i<=Sample1->Size;i++) {
            //cout << Sample1->Size << " " << Sample1->Sam_AlleleNbr ;cout.flush();
    		for(int j=0;j<Sample1->Sam_AlleleNbr;j++) delete[] MatriceSD[i][j];
    		delete[] MatriceSD[i];
    	}
    	delete[] MatriceSD;//pour calcul des Pi SD OnePop
		delete[] SDpivec;
		for(int j=0;j<Sample1->Sam_AlleleNbr;j++) delete[] MatA[j];
		delete[] MatA;

    }
    delete CurrentSample;
    if(DemographicModel!="IM") Sample1->delete_L_etc();
    delete MutMod1;
    delete Sample1;

    if(!UsePCL_SISRbool){
        vector<long double>::iterator aa=SumL.begin();
        for (vector<long double>::iterator ii=MeanL.begin();ii<MeanL.end();ii++,aa++)
        	(*ii)+=(*aa)/((long double) (args.nruns)); // otherwise warning with some compilers ?

        vector<long double>::iterator bb=SumTMRCA.begin();
        for (vector<long double>::iterator jj=MeanTMRCA.begin();jj<MeanTMRCA.end();jj++,bb++)
            (*jj)+=(*bb)/((long double) (args.nruns)); // otherwise warning with some compilers ?

    }// bloc aa local = maintenant fin du if(!UsePCL_SISRbool){

    vector<float> zut(mloglikdim);
    if(UsePCL_SISRbool){
        vector<float>::iterator ii=zut.begin();
        (*ii)=-LnMeanLikelihood;
        ii++;
        for (;ii<zut.end();ii++) (*ii) -=  (float) log(0);//!! RL -> FR
    } else {
        vector<long double>::iterator aa=MeanL.begin();
   		for (vector<float>::iterator ii=zut.begin();ii<zut.end();ii++,aa++) (*ii)+=-log((*aa));
    }// bloc aa local = maintenant fin du else if(UsePCL_SISRbool)

#ifdef DEBUG
if((/*singlePointBool ||*/ PrintLikelihood) && (TMRCABool || DemographicModel=="IM")) {
    cout<<setprecision(10)<<"Mean TMRCA for this locis at this point is "<<MeanTMRCA[0]<< endl;
	ofstream TMRCAFile("TMRCA.txt",ios::out|ios::app);
	if(!TMRCAFile.is_open()) {
		cerr << "Unable to open file TMRCA.txt in calclog();" << endl;
		if (cinGetOnError) cin.get();
		TMRCAFile.clear();
		exit(-1);
	}
	TMRCAFile << endl;
	TMRCAFile << setprecision(10) << MeanTMRCA[0] << endl;

    //cout<<"Proportion of skiped trees due to non exact SMM  computations= "<<SkipedTrees/TotalTrees<< endl;
    //if(cinGetOnError) cin.get();
}
#endif

//ostream_vector(zut,cout);cin.get(); //frfrfrfr
return zut;
}
/***************   FIN CALCLOGL (1 Sample=1 locus)  ********************************************/





//#define TEMPS

double Euclidian(vector<pirestype> &veca,vector<pirestype> &vecb) {
double dist=0;
vector<pirestype>::iterator jt=vecb.begin();
for (vector<pirestype>::iterator it=veca.begin();it<veca.end();it++) {
	dist+=(*it-*jt)*(*it-*jt);
	jt++;
	}
return sqrt(dist);}


changetype CSample::computeIIS(CcomputeId* loccalculQ,CMutMod* MutMod) {

// directly computes the probability of the remaining pair of genes
int d1=-1,d2=0;// the demes where the genes are
int t1=-1,t2=0;// the allelic types
int c;
vector<int>::iterator it;
pirestype ibd,iis,pr=0.;

#ifdef DEBUG
    if(DebugIIS) {
        cout << "\nSample configuration when entering ComputeIIS()" << endl;
        Print();
    }
#endif
        for(int a=0;a<PopSampledOrNot;a++) {
            if(PopEffectif[a]==2) { //intrapop
                d1=d2=a;
                for(unsigned int i=0;i<Effectif[a].size();i++) {
                    c=Effectif[a][i];
                    if (c==2) { // si identiques
                        t1=t2=i;
                        Effectif[a][i]=0;
                    } else if (c==1){ //if different
                        if (t1==-1) { // si first type not yet identified
                            t1=i;
                            Effectif[a][i]=0;
                        } else {
                            t2=i;
                            Effectif[a][i]=0;
                            break;
                        }
                    }
                }
            } else if (PopEffectif[a]==1){ //interpop
                if (d1==-1) {
                    d1=a;
                    it=find(Effectif[a].begin(),Effectif[a].end(),1);
                    t1=int(it-Effectif[a].begin());
                    Effectif[a][t1]=0;
                } else {
                    d2=a;
                    it=find(Effectif[a].begin(),Effectif[a].end(),1);
                    t2=int(it-Effectif[a].begin());
                    Effectif[a][t2]=0;
                }
            }
        }
        if ( mutationModel=="PIM" or (mutationModel=="GSM" and MutMod->pGSM>0.99)) {
            if (C_c_bool) cerr<<endl<<"call to CcomputeId::IdbyD(...)";
            ibd=calculQ.IdbyD(d1,d2); // assumes symmetric migration matrix
            if (C_c_bool) {cerr<<endl<<"returned from call to CcomputeId::IdbyD(...)";C_c_bool=false;}
            iis=ibd+(1-ibd)/Sam_AlleleNbr; //symmetric K allele model
            if (t1==t2) {
                pr=iis/Sam_AlleleNbr;
            } else {
                pr=(1.-iis)/(Sam_AlleleNbr*(Sam_AlleleNbr-1));
                if (d1==d2) pr*=2.; //pr's sum to 1 over all n(n-1)/2 distinct a1,a2 pairs within one deme
                // otherwise pr's sum to 1 over all n(n-1) distinct a1,a2 pairs within two demes
            }
        } else if( mutationModel=="SMM" or mutationModel=="GSM") { // circular  stepwise mutation model, SMM and GSM
           
                //New code, RL 02/2017
                long double posTerms=0.,negTerms=0.,temp,allTerms=0.,maxIBD=0.;
                bool use1minusQ=false;
                picalctype factor=2.*PI/Sam_AlleleNbr;
                
                //For OnePop models, computation of ibdTable is done for the first run of nrunsperpoints.
                // Need to be extended for models with more than one population
                if(!calcQibdtableDoneBool || DemographicModel.substr(0,6)!="OnePop") {
                    picalctype muIBD=loccalculQ->mu; // mu IIS
                    
                    //vector<picalctype> ibdTable(locallelenbr);//RL now resized in CcomputeId::initMut()
                    for (int allele=0;allele<Sam_AlleleNbr; allele++) {
                        if( mutationModel=="SMM") loccalculQ->mu=muIBD*(1.0-cos(factor*allele));
                        else loccalculQ->mu=muIBD*(1.0- ( (cos(factor*allele)-MutMod->pGSM)*(1-MutMod->pGSM)/(1.-2.*MutMod->pGSM*cos(factor*allele)+MutMod->pGSM*MutMod->pGSM)) );
                        
                        //Computation of Q
                        if(loccalculQ->mu==0.0) loccalculQ->ibdTable[allele]=1.0;
                            else if(DemographicModel.substr(0,6)=="OnePop") {
                                loccalculQ->gamma=pow( (1.-loccalculQ->mu),( picalctype ) 2. );
                                loccalculQ->ibdTable[allele]=loccalculQ->gamma/(loccalculQ->Nhdeme*(1.-loccalculQ->gamma)+loccalculQ->gamma);
                            } else loccalculQ->ibdTable[allele]=loccalculQ->IdbyD(d1,d2);
                    }
                    loccalculQ->mu=muIBD;
                    calcQibdtableDoneBool=true;
                } //end if(!calcQibdtableDoneBool || DemographicModel.substr(0,6)!="OnePop")
                
                
                int stepnumber=abs(t1-t2);
            
                //compute max of IBD Table to check if it is close to 0.99,
                //in which case further computations are more precise when based on 1-Q instead of Q
                if(*max_element(loccalculQ->ibdTable.begin(), loccalculQ->ibdTable.end())>0.99) use1minusQ=true;
                
                //gain de temps possible : enregistrer le dernier stepnumber, et recalculer uniquement si il a changé, reinitialiser pour chaque point
                for (int allele=0;allele<Sam_AlleleNbr; allele++) {
                    if(use1minusQ) temp=cos(factor*stepnumber*allele)*(1.-loccalculQ->ibdTable[allele]);
                         else temp=cos(factor*stepnumber*allele)*loccalculQ->ibdTable[allele];
                    if(temp>0) posTerms+=temp; else negTerms+=temp;
                    allTerms+=temp;
                }
                //Just to check, to be removed, RL 022013,
                //RL 042014 not sur it should be removed because this often hapens when there is lots of mutations resulting in potentialy very divergent final allelic types
                if(fabs((posTerms+negTerms-allTerms)/allTerms)>0.001) {
                    cout << "\n!! In ComputeIIS for SMM or GSM:" << endl;
                    cout << "there is a difference between the sum of positive and negative terms and the summ of all terms!! " << endl;
                    cout << "If this message often appears for a specific data set, it may be due to potentialy very high mutation rates (i.e. high theta values) " << endl;
                    cout << "or acumulation of many mutations of a single branch of the tree." << endl;
                    cout << "In such cases, exact computation of the probability of the last pair of genes (2id) may not be appropriate." << endl;
                    cout << "Then you should better use the option statisticSequence=ISstrict or PAC or PACanc, but not 2id." << endl;
                }
                
                if(use1minusQ) {
                    if(stepnumber==0) pr=(Sam_AlleleNbr-allTerms)/Sam_AlleleNbr; else pr=-allTerms/Sam_AlleleNbr;
                } else pr=allTerms/Sam_AlleleNbr;
                
                if(stepnumber!=0 && d1==d2) pr*=2.;
            
        } else { // not PIM nor SMM/GSM
			cerr << "Mutation model = " << mutationModel << " is not compatible with analytical computation of identity in state CalculQ()/ComputeIIS()" << endl;
			cerr << "I exit." << endl;
			if(cinGetOnError) cin.get();
			exit(-1);

		}
        if(pr>=0.0) {
			Size=0;
			PopEffectif[d1]=0;
			PopEffectif[d2]=0;
		} else {
			Effectif[d1][t1]++;
			Effectif[d2][t2]++;
		}
      return(pr);
}

changetype CSample::Change(CMigMod* Mig,CMutMod* Mut){
// for PAC likelihood see CSample::PACpi();
//cout<<"debut Change"<<flush;

int a,j; //pas locales aux boucles
int b,i,Lasti/*,k*/; // locales aux boucles mais gagne epsilon de temps si def globalement
bool firstPassbool,localSDComputationbool;//Boolean to chack if ChoosePop&Type is called many times (e.g. when a demo change is found instead of a new event or when Pija is null)
changetype Weight,ChooseRes;//pour resultat de choosepoptype si changement demo
cumultype Random,Cumul;
//modification de pascumulype en pirestype car ce sont des pi et denominator est utilis['e] pour cumul et ne doit pas etre nul 18122008
    pirestype Denominator;//,AlternativeDenominator
pirestype PIja,PIjb,PIia;
pascumultype StationAll=1.0/Sam_AlleleNbr; //FR->FR code pas clair, ce n'est pas generalement le nombre d'alleles observ'es
const parstype ThetaDIG=Mut->ThetaDIG;
const int PSON=PopSampledOrNot;
vector<int>::iterator siteJ;	// CBR (29.04.2012) specifies which mutation of the chosen singleton gene needs to be eliminated
bool RechoosePopType;	//	CBR (02.05.2012) specifies if a new Pop'n'Type needs to be chosen for ISM
bool singletonMigration;	//	CBR (02.05.2012) helps in checking the event compatibility for singleton genes capable of ONLY immigration
DIGEventType singleGeneISM;

#ifdef DEBUG
	if (DebugDGISM) {
		cout << "-----------------------------------------------------------------------------------\n";
		cout << "Mutation paths before CSample::Change():\n";
		PrintSeq();
		cout << "Pops before CSample::Change:\n";
		Print();
	}
#endif

events++;

//	code for returning the weight for the last individual of the MRCA type
if ((mutationModel == "ISM" || mutationModel == "SNP") && Size == 1) {
	Size = 0;
	return 1.0;//RL -> FR, RL -> CBR, ca me parait bizarre, distrib sttationnaire des haplotypes?? sans doute pas très important
}

//RL sans doute pas necessaire si on le fait apres chaque mutation.
//if(mutationModel=="SMM" or mutationModel=="GSM") {GetAllelicRange(Mig,Mut,0);GetAllelicRange(Mig,Mut,1);}

//si il reste un seul genes alors le poids est la proba de l'allele dans la distrib stationnaire
if((Size==1) && (mutationModel != "ISM") && (mutationModel != "SNP")){
	for(a=0;a<PSON;a++)
		if(PopEffectif[a]>0)
			for(vector<int>::iterator i=Effectif[a].begin();i<Effectif[a].end();i++)
				if((*i)==1){
					Weight=StationAll;
#ifdef DEBUG
					if(DebugDG) cout << "\n Final Step, probability of state XXX(not printable iterator) is " << Weight << endl;
#endif
					Size=0;
					(*i)=0;
					PopEffectif[a]=0;
                    if (Weight>1 || Weight <= 0) {
                       cout<<"Problem in weightStation computation! a v['e]rifier -> "<<Weight << endl;
           				//getchar();
                    }
                    //cout << "WeightStation=" << Weight << endl;
					return Weight;
					}
	}

firstPassbool=true;// for non equilibrium models, allows to go through choosePop&Type without computing a new time if it is not a new event.

int countSeq = 0;
while (true) { //RL : pour éviter les problemes de somme de cumul << 1, probleme a résoudre! = adapter les modeles finis de mutation pour SMM et GSM.

//si il reste plus d'un gene
//choose a pop and a type (eq. 2.12 DIG04ii)
PIja=0.0;
int compt=0;
countSeq++;
PIjabool=true;//PIjabool only for alternativeIS
while (PIja<=0.0 || (PIja-1.0)>epsilonPiGSM) { //ne pas remplacer par <=epsilon.... (14/01/11)
/** With this test,  when Pija ~0 we're acting as if the proposal term was 0.
    i.e PIia << PIja; cf mutation term in Table 2 of DIG04ii... Here is some reason for doing so:

    We have problems in the ONe Pop SMM case, [where the hat pi's are computed by the explicit formulas 3.9, 3.10 in DGLR05];
    in the case where only mutation occurs (one copy of all the alleles present), all alleles present are non-adjacent, and theta is small
    In that case we see from the explicit formulas that PIja and PIia are both small but their ratio is
    PIia/PIja=(rho/(1+sqrt(1-rho^2))) with rho=theta/(n+theta)~0 hence the proposal term ~ PIia/PIja ~0
    **/
	compt++;
	localSDComputationbool=false;
	RechoosePopType = false;
	ChooseRes=ChoosePopType(&a,&j,Mig,Mut,firstPassbool);
	firstPassbool=false;
    if(ChooseRes==-1.0) {
	   cout << "\n\nProblem in Choose Pop & Type, result is -1" << endl;
	   cout << "\nI exit" << endl;
	   if (cinGetOnError) cin.get();
	   exit(-1);
    } else if (ChooseRes!=-666) return ChooseRes;//si on est a MinChangeTime
#ifdef DEBUG
    if(DebugChoosePopType || DebugDG) cout << "\nChoose Pop=" << a << " and type=" << j << endl;
#endif
    //on enleve ce gene de l'echantillon
    PopEffectif[a]--;
    Effectif[a][j]--;
    StateEffectif[j]--;
    Size--;

    //vérifications ISM pour savoir quels evenements sont possibles
    if (mutationModel == "ISM"  || mutationModel == "SNP") {
		//	playing it safe
		singleGeneISM = noDIGEvent;

		//	under ideal conditions nZero remains PopEffectif[a] + 1
		nZero[a] = PopEffectif[a] + 1;

		//	CBR: several checks and event assignments for singleton genes capable of mutation/ONLY immigration events
		//	1°	identifies a site and assigns ISMtype1 mutation when the chosen gene contains at least 2 unique mutation sites
		//	2°	identifies a site and assigns ISMtype2 mutation when the chosen gene contains 1 unique mutation site
		//	3°	corrects nZero (i.e. no mutation event) when the chosen singleton gene contains no unique mutation sites
		//	4°	assigns migOnly (i.e. capable of ONLY immigration) when the chosen gene is a local singleton
		for (int k=0; k < Sam_AlleleNbr; k++) {
			//	"GLOBALLY singleton" check for EVERY HAPLOTYPE accounting for the gene taken out of the sample
			if (((k != j) && (Effectif[a][k] == 1) && (StateEffectif[k] == 1))
					|| ((k == j) && (Effectif[a][k] == 0) && (StateEffectif[k] == 0))) {
				vector<int>::iterator it1, it2;
				//	looping over the list of segregated sites for each haplotype
				for (it1 = mutPathMatrix[k].begin(); *it1 != 0; it1++) {
					//	IF a unique mutation is found for the haplotype
					if (siteMutCountVec[*it1 - 1] == 1) {
						//	IF the haplotype is of the chosen type
						if (k == j) {
							//	continue loop
							for (it2 = it1+1; *it2 != 0; it2++) {
								//	check for a second mutation
								if (siteMutCountVec[*it2 - 1] == 1) {
									//	found a second unique mutation site = migPlusISMtype1
									siteJ = it1;
									singleGeneISM = migPlusISMtype1;
									break;
								}
							}
							if (*it2 == 0) {
								//	found only a single unique mutation site = migPlusISMtype2
								siteJ = it1;
								singleGeneISM = migPlusISMtype2;
							}
						}
						//	break loop as the current haplotype corresponds to a possible event
						break;
					}
				}
				//	no unique mutation sites found = no possible events for this haplotype
				if (*it1 == 0) {
					--nZero[a];
					//	IF this haplotype is the chosen type then re-choose a new Pop'n'Type
					if (k == j) {
						RechoosePopType = true;
						PIja = 0.0;	//	to ensure that the (PIja<=0.0) while loop keeps running
						break;
					}
				}
			}
			//	CBR: correct nZero and count singleton genes capable of ONLY immigration
			singletonMigration = false;
			//	"LOCALLY singleton" check for EVERY HAPLOTYPE accounting for the gene taken out of the sample
			if (((k != j) && (Effectif[a][k] == 1) && (StateEffectif[k] > 1))
					|| ((k == j) && (Effectif[a][k] == 0) && (StateEffectif[k] > 0))) {
				//	loop over second pop
				for (b = 0; b < PSON; b++) {
					//	IF second pop is different from chosen + backward migration possible + chosen type exists in second pop
					if ((b != a) && (Mig->Mab[a][b] != 0) && (Effectif[b][k] > 0)) {
						//	THEN the current haplotype is capable of IMMIGRATING FROM at least one other pop
						singletonMigration = true;
						break;
					}
				}
				//	IF the chosen type can ONLY immigrate THEN set event
				if ((singletonMigration) && (k == j))
					singleGeneISM = migOnly;
				//	IF the chosen type couldn't have immigrated into the chosen pop THEN re-choose a new Pop'n'Type
				else if ((!singletonMigration) && (k == j)) {
					RechoosePopType = true;
					PIja = 0.0;	//	to ensure that the (PIja<=0.0) while loop keeps running
					break;
				}
				//	IF the current haplotype couldn't have immigrated into the chosen pop THEN modify nZero
				else if ((!singletonMigration) && (k != j))
					--nZero[a];
			}
		}
	}

//on calcul le Pi pour ce gene
//cout<<"avant PicalcPIM";
//	CBR: we only need to calculate PI's for ISM when the chosen gene (j) is not a singleton
	if ((mutationModel != "ISM" && mutationModel != "SNP") || ((mutationModel == "ISM" || mutationModel == "SNP") && (Effectif[a][j] > 0))) {
		chosenPop = a;	//	CBR: needed for CSample::PiCalcISM
		(this->*PicalcfnPtr)(Mig,Mut,j); //Pi for the gene taken out; more calls to PicalcfnPtr for mutation...
		//cout<<"apr[`e]s PicalcPIM";
		//JAMA doesn't use vectors, so we really need the a,j indices here
		//if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) PIja=altPIs[j][a];
		// else
		PIja=PIs[j][a];
		if(piOnePopEigenbool and PIja<epsilonPiGSM){
            if(false) cout << "PiOnePop SD will be used instead of Eigen because PIja=" << PIja << " < epsilonPiGSM=" << epsilonPiGSM << endl;
			localSDComputationbool=true;

			CalcMatSD(Mut,Mig,ThetaDIG,false);//pour PiOnePopSD in current population
			picalctype mab,pib=0.0;
			for(int k=0;k<Sam_AlleleNbr;k++) {
				mab=MatriceSD[Size][k][j];
				if(Effectif[0][k]>0.0) pib+=((Effectif[0][k]) * mab )/ (Size);
			}
			PIja=pib;

		}

    if (PIja<=0.0 || (PIja-1.0)>epsilonPiGSM) { //can occur numerically for SMM, very unlikely mutation events, at least, probably due to very low values converted to 0, should be worth with float calc.
        if (C_c_bool) cerr<<endl<<"PIja="<< PIja << "=0 or >1 in CSample::Change()";
#ifdef DEBUG
		cerr << "\nPIja="<< PIja << "(<=0 or >1) in CSample::Change()";
		if(true) {
			Print();
			if(localSDComputationbool) cerr << "SD algorithm used"<<endl;
			cerr << "\tChoose pop=" << a << " allele="<<j<<endl;
			cerr << "\tPopEffectif[a] = " << PopEffectif[a] << endl;
			cerr << "\tMig->PopSize[a] = " << Mig->PopSize[a] << endl;
			cerr << "\tMig->Ma[a] = " << Mig->Ma[a] << endl;
			cerr << "\tThetaDIG = " << ThetaDIG << endl;
			//		if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr << "PopSize=AncThetaDIG/ThetaDIG = " << AncThetaDIG/ThetaDIG << endl;
			cerr << "\tDenominator is " << Denominator << endl;
			//		if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr << "AlternativeDenominator is " << AlternativeDenominator << endl;
			cerr << "\tPIja is" << PIja << endl;
			cerr << "\tDenom.remainder (PopEffectif[a]/Mig->PopSize[a]+Mig->Ma[a]+ThetaDIG) is " << (PopEffectif[a]/Mig->PopSize[a]+Mig->Ma[a]+ThetaDIG) << endl;
			//		if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr << "AlternativeDenom.remainder (PopEffectif[a]/(AncThetaDIG/ThetaDIG)+Mig->Ma[a]+ThetaDIG) is " << (PopEffectif[a]/(AncThetaDIG/ThetaDIG)+Mig->Ma[a]+ThetaDIG) << endl;
			if(PSON>1) cout << "\tMab[1][0]="<< Mig->Mab[1][0] << endl;
			//if(true or AlternativeIS_FRExpMixAncActPopSizebool) DebugPiSMM=true;
			//PrintPi(Mig,Mut);
			//if(true or AlternativeIS_FRExpMixAncActPopSizebool) DebugPiSMM=false;
			//cout<<'\a'<<'\a';//<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a';
			//cout << endl;
			//cout << "Press any key to resume." << endl;
			if (DebugSigmaPi)
				PrintPi(Mig,Mut);
			//fflush(stdout);
			//if (cinGetOnError) cin.get();
		}
#endif
	   //on remet le gene pr['e]c['e]denmment enlev['e] de l'echantillon
			PopEffectif[a]++;
			Effectif[a][j]++;
			StateEffectif[j]++;
			Size++;
		}
		if(compt==100 || compt==200 || compt==300) cout << compt << " tries to compute a not too small PIja" << endl;
		if(compt==300) {
			cout << "Too many tries to compute PIja, I exit." << endl;
			if(cinGetOnError)  cin.get();
			exit(-1);
		}
	}

	if (RechoosePopType) {
		//	restoring the gene taken out of the sample
		++Effectif[a][j];
		++PopEffectif[a];
		++StateEffectif[j];
		++Size;
    }
    else
    	if(mutationModel=="ISM"  || mutationModel == "SNP") break;	//	RL -> CBR if(ISM) added ... CBR: IF no need to calculate PI's (for ISM) or no event possible then break from the PIja<=0.0 while loop
    // at this point either PIja=0 and the while loop iterates until a good j and a are found
    // or we exit the loop
} // fin while PIja==0
PIjabool=false;
#ifdef DEBUG
if(DebugDG) {
	/*if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool)
			cout << "\nChange()->altPIja[" << j << "][" << a << "]=" << PIja << "; (and Pi=" << PIs[j][a] << ")" << endl;
	 else*/ cout << "\nChange()->PIja[" << j << "][" << a << "]=" << PIja << endl;
}
//verif du PIja
if(PIja<0.0 || (PIja-1.0)>epsilonPiGSM){
	cout << "------------------------------" << endl;
	cout << "Bug pi("<< j << "," << a << ")=" << PIja << endl;
	cout << "numeric_limits<picalctype>::min()=" << numeric_limits<picalctype>::min() << endl;
	cout << "numeric_limits<picalctype>::epsilon()=" << numeric_limits<picalctype>::epsilon() << endl;
	Print();
	PrintPi(Mig,Mut);
	if (cinGetOnError) cin.get();
}

#endif
//choose the event
//note that the sample has already decreased for type j,a
while(true) { // repeats until a return is executed...

	if (C_c_bool) cerr<<endl<<"One loop of  while(true) in CSample::Change()";

	Random=alea();
	Cumul=0.0;

#ifdef DEBUG
	if(DebugDG2) cout << "Random=" << Random << endl;
	if(DebugProposal) {
		cout << "Random is set to 1.2, so that no event can occurs, to get all proposal probabilties" << endl;
		Random=1.2;
	}
#endif

	//calcul D(n)(p439 DIG04ii)  (110629012 d'eplac'e : )

	//if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) AlternativeDenominator=(PopEffectif[a]/(AncThetaDIG/ThetaDIG)+Mig->Ma[a]+ThetaDIG)*PIja;

	Denominator=(PopEffectif[a]/Mig->PopSize[a]+Mig->Ma[a]+ThetaDIG);
	if (mutationModel != "ISM" && mutationModel != "SNP") Denominator*=PIja;//a un facteur pres car on a PopEffectif[a] deja choisi la pop et le type
													//on multliplie ensuite par PIja en allelique car toujours come ca dans les proposal col1 tab2 DIG04ii
		else Denominator*=(PopEffectif[a] + 1); //	CBR: see pg.447, DIG'2004b, note that Denominator = Ra in GD04ii


//CBR->RL,FR: warning: ‘AlternativeDenominator’ may be used uninitialised in this function
//if(DebugDG2 or DebugDGISM or Denominator<=0.0 || std::isnan(Denominator) or  ( (AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) and (AlternativeDenominator<=0.0 || std::isnan(AlternativeDenominator) ) ) ) {
if(DebugDG2 or DebugDGISM or Denominator<=0.0 || std::isnan(Denominator) or  ( (AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) ) ) {
			if(Denominator<=0.0 || std::isnan(Denominator)) cout << "Problem in Change; Denominator(<=0.0 or > 1.0) =" << Denominator << endl;
#ifdef DEBUG
		cout << "\tscaled_generation = " << scaled_generation << endl;
		cout << "\tChoose Pop=" << a << " and type=" << j << endl;
		cout << "\tPopEffectif[a] = " << PopEffectif[a] << endl;
		cout << "\tMig->PopSize[a] = " << Mig->PopSize[a] << endl;
		cout << "\tMig->Ma[a] = " << Mig->Ma[a] << endl;
		cout << "\tThetaDIG = " << ThetaDIG << endl;
		if (mutationModel == "ISM" || mutationModel == "SNP") {
			cout << "\tSize = " << Size << endl;
			cout << "\tnZero = " << nZero[a] << endl;
		}
		//if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) {
		//	cout << "\tMig->PopSize[a] = (AncThetaDIG/ThetaDIG)" << (AncThetaDIG/ThetaDIG) << endl;
		//	cout << "\tAncThetaDIG = " << AncThetaDIG << endl;
		//	cout << "AlternativeDenominator= " << AlternativeDenominator << endl;
		//}
		cout << "\tPIja = " << PIja << endl;
		Print();
		if(!DebugDG2) {
			DebugPiSMM=true;
			PrintPi(Mig,Mut);
			fflush(stdout);
			DebugPiSMM=false;
		}
		if (cinGetOnError) cin.get();
#endif
			if(Denominator<=0.0 || std::isnan(Denominator) or
				   ( (AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) ) )  return(-1.0);
//						(AlternativeDenominator<=0.0 || std::isnan(AlternativeDenominator)) ) )  return(-1.0);

			}

/*******  coalescence*/
if(Effectif[a][j]>0){
		//facteur de la proposal col1 tab2 DIG04ii
		/*if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) Cumul=(Effectif[a][j]/(AncThetaDIG/ThetaDIG))/AlternativeDenominator;
		else*/
		if (mutationModel != "ISM" && mutationModel != "SNP")  Cumul=(Effectif[a][j]/Mig->PopSize[a])/Denominator;
		 else Cumul = (Effectif[a][j] / Mig->PopSize[a] * nZero[a]) / (Denominator * PIja);// RL->CBR comprends pas, ne correspond a ce qui est écrit dans le tableau 4 de DG2004ii. effet de choose pop and type?

		//cumul accumulates the terms of the proposal distribution col1 tab2 DIG04ii.
		//Ideally the sum (over coal, mut from j, mig from beta, or alpha, i fixed) will sum to one.
		//SMM with boundariy: If the continuous approximation to the discrete Fourier expression for mutation is inaccurate the mutation probability may be too low or too high
#ifdef DEBUG
	if(DebugDG2 && !(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool)) cout << "Change()/coa->Cumul= (Effectif[a][j]/Mig->PopSize[a])/Denominator=" << Cumul << endl;
//	if(DebugDG2 && (AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool)) cout << "Change()/coa->Cumul= (Effectif[a][j]/(AncThetaDIG/ThetaDIG))/AlternativeDenominator=" << Cumul << endl;
#endif

		if(Random<Cumul){
			//calcul le poid pour une coalescence col1 tab2 DIG04ii
			Weight=PIja*(PopEffectif[a]+1.0)/(Effectif[a][j]+1.0);
	//		if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) Weight*=(AncThetaDIG/ThetaDIG)/Mig->PopSize[a]*(AlternativeDenominator/Denominator);
	//		  else if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) Weight*=proposalCorrectionFactor;//to correct the cumul probabilities so that they sum to 1, 19012011
#ifdef DEBUG
		if(DebugDG || DebugDGISM){
			cout << "\t *coalescence in pop " << a ;
			cout << ", of type " << j ;
            //printf("\n\t Weight=%g ",Weight);
            cout << "\t Weight=" << Weight << "\n\n" << endl;
		}
#endif
	//cout<<"fin coal"; cout.flush();
	//if (Weight>1) {cout<<"fin coal "<<Weight; getchar();}
	//		disp_coal_events++;
			if (/*Weight>1 ||*/ Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
				cerr << "Numerical errors -> Weight=" << Weight << "Cumul=" << Cumul << endl;
                if(!(piOnePopSDbool || piOnePopEigenbool) and Sam_AlleleNbr<200) {
                    cerr << "Using PinonePopSolver=DG with a small number of allelic state (e.g. <200) causes some numerical errors." << endl;
                    cerr << "I will continue the computation but you may try piOnePopSolver=SD or Eigen instead of the default DG." << endl;
                }
                if((piOnePopSDbool || piOnePopEigenbool) and Sam_AlleleNbr>200) {
                    cerr << "Using PinonePopSolver=SD or Eigen with a large number of allelic state (e.g. >200)" << endl;
                    cerr << "may be very slow and may causes some numerical errors." << endl;
                    cerr << "I will continue the computation but you may try piOnePopSolver=DG instead of SD or Eigen." << endl;
                }

#ifdef DEBUG
           cerr<<"\n\n(!) Problem in weight or cumul computation when looking for a coalescent event" << endl;
            /*if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cout<<"Weight=PIja*(PopEffectif[a]+1.0)/(Effectif[a][j]+1.0)*AlternativeDenominator/Denominator "<<Weight << endl;
			else*/ cerr<<"Weight=PIja*(PopEffectif[a]+1.0)/(Effectif[a][j]+1.0) = "<<Weight << endl;
            /*if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cout<<"Cumul=(Effectif[a][j]/(AncThetaDIG/ThetaDIG))/AlternativeDenominator; "<<Cumul << endl;
			else*/ cerr<<"Cumul=(Effectif[a][j]/Mig->PopSize[a])/Denominator =  "<<Cumul << endl;
            cerr<<"Mig->PopSize[a]="<<Mig->PopSize[a] << endl;
            //if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cout<<"AlternativeDenominator="<< AlternativeDenominator << endl;
			cerr<<"Denominator="<< Denominator << endl;
            cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
            cerr << "\tPopEffectif[a] = " << PopEffectif[a] << endl;
            cerr << "\tEffectif[a][j] = " << Effectif[a][j]<< endl;
            if(mutationModel == "ISM" || mutationModel == "SNP") cout << "\tnZero = " << nZero[a] << endl;
	        cerr << "\tPIja = " << PIja << endl;
	        Print();
            //DebugPiSMM=true;
            //PrintPi(Mig,Mut);
            fflush(stdout);
			//DebugPiSMM=false;
			if (DebugSigmaPi)
				PrintPi(Mig,Mut);
			if (cinGetOnError) cin.get();
#endif
			}
		//cout << "WeightCoa=" << Weight << endl;
		return Weight;
		}//end if(Random<Cumul)
}//end if(Effectif[a][j]>0)


/************ mutation*/
    factortype factor; /** This needs to be long double, because a (pirestype=double) can be below <double>::min(),
    but cannot be above <double>::max()
    Denominator can be below min() and then factor is above <double>::max **/
	/*if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) factor=ThetaDIG/AlternativeDenominator;
	else*/ factor=ThetaDIG/Denominator;
	if(std::isinf(factor)) {
		cerr<<"Problem in Change(): 'factor' is infinite = " << factor << endl;
		cerr<<"This is a numerical issue. Try higher Theta values." << endl;
		cerr<<"If the problem persists, report it to the authors." << endl;
		if (cinGetOnError) cin.get();
	}

if(mutationModel=="SMM") {

	for(int k=0;k<2;k++){
		i=j-1+2*k;
		if (/*false or */!(piOnePopSDbool || piOnePopEigenbool) || (i>=0 && i<Mut->Mut_AlleleNbr)) {//// SMM soit circulaire (DG), soit "absorbant" (SD Eigen)

			if (/*false or */!(piOnePopSDbool || piOnePopEigenbool)) {// SMM circulaire, pris en compte dans PiCalcOnePop()
				if(i>=Mut->Mut_AlleleNbr) i-=Mut->Mut_AlleleNbr;
				else if(i<0) i+=Mut->Mut_AlleleNbr;
			}
			Lasti=i;

			//RL -> RL, RL -> FR
			//PIia sometimes very small because eigenvector and eigen values computations may not be very precise,
			//no clean solution for the moment
			if (i!=j) {


				if(piOnePopEigenbool and localSDComputationbool){
#ifdef DEBUG
                    if(false) cout << "PiOnePop SD used instead of Eigen because localSDComputationbool is true;" << endl;
#endif
					picalctype mab,pib=0.0;
					for(int k=0;k<Sam_AlleleNbr;k++) {
						mab=MatriceSD[Size][k][i];
						if(Effectif[0][k]>0.0) pib+=(picalctype) ((Effectif[0][k]) * mab )/ (Size);
					}
					PIia=pib;

				} else {
					(this->*PicalcfnPtr)(Mig,Mut,i);
					PIia=PIs[i][a];
				}
			}
			if(PIia<0.0 || PIia>1.0 || std::isnan(PIia)) {
				cout << "Problem in Change for mutation SMM PIia=" << PIia << endl;
				if (cinGetOnError) cin.get();
			}
			//calcul le facteur de la proposal col1 tab2 DIG04ii
			//Pij est toujours 0.5 en SMM, non uniquelent en SMM curculaire
			Cumul+=factor*Mut->Pij[i][j]*PIia;// pas logique du tout de Pji, revoir avec FR.
#ifdef DEBUG
			if(DebugDG2 || Cumul>1.001 || std::isnan(Cumul) || std::isnan(factor)) {
                thisMessage1++;
                if(thisMessage1<=50){
                    cerr << "\nChange()/mutSMM->Cumul+=ThetaDIG/Denominator*Mut->Pij["<< i << "][" << j << "]*PIia=" << Cumul << endl;
                    cerr << "\tThetaDIG=" << ThetaDIG << endl;
                    cerr << "\tDenominator=" << Denominator << endl;
                    cerr << "\tMut->Pij[i][j]="<< Mut->Pij[i][j] << endl;
                    cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
                    cerr << "\tEffectif[a][i] = " << Effectif[a][i] << endl;
                    cerr << "\tEffectif[a][j] = " << Effectif[a][j]<< endl;
                    if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool)
                        cerr << "\taltPIia=" << PIia << " (and PIia=" << PIs[i][a] << ")" << endl;
                    else cerr << "\tPIia=" << PIia << endl << endl;
                    cerr << "\tPIja = " << PIja << endl;
                    if (DebugSigmaPi)
                        PrintPi(Mig,Mut);
   
                }
                if(thisMessage1==50){
                    cerr << "\n Migraine will continue the computation but this message will not be further printed out. \nTry to increase the state number of the SMM (e.g. +200) to remove this problem." << endl;
                }
            }
#endif
			if(Random<Cumul){
				//calcul le poid de cette mutation col1 tab2 DIG04ii
				Weight=(Effectif[a][i]+1)*PIja;
				Weight/=(PIia*(Effectif[a][j]+1));
//				if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) Weight*=(AlternativeDenominator/Denominator);
//					else if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) Weight*=proposalCorrectionFactor;//to correct the cumul probabilities so that they sum to 1, 19012011
				if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) Weight*=proposalCorrectionFactor;//to correct the cumul probabilities so that they sum to 1, 19012011
				//ajust the allelic range if changed
				if(i<AllelicRange[0]) AllelicRange[0]=i;
				if(i>AllelicRange[1]) AllelicRange[1]=i;
				//cout << "AllelicRange[0]=" << AllelicRange[0] << endl;
				//cout << "AllelicRange[1]=" << AllelicRange[1] << endl;
				//change les caracteristiques de l'echantillon apres cette mutation
				Effectif[a][i]++;
				PopEffectif[a]++;
				StateEffectif[i]++;
				Size++;
#ifdef DEBUG
                if(AllelicRange[0]<0 || AllelicRange[1]>(Sam_AlleleNbr-1))
                    cout << "Problem in SMM state Number :  min=" << AllelicRange[0] << " and max=" << AllelicRange[1] << endl;
				if(DebugDG){
					cout << "\t *SMM Mutation back in pop ";
					cout << a << ": type " << j << "->" << i << ", " ;
					if(DebugDG2) cout << " cumulate=" << Cumul;
					//printf("\n\t Cumulate=%g",Cumul);
					if(DebugDG2) cout << ", Random=" << Random << endl;
					//printf("; Random=%g ;",Random);
					//printf("\n\t Weight=%g ",Weight);
					cout << "\t Weight=" << Weight << endl << endl;
				}
#endif
				if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
                    thisMessage1++;
                    if(thisMessage1<=50){
                        cerr << "Numerical errors -> Weight=" << Weight << "  Cumul=" << Cumul << endl;
                        if(!(piOnePopSDbool || piOnePopEigenbool) and Sam_AlleleNbr<200) {
                            cerr << "Using PinonePopSolver=DG with a small number of allelic state (e.g. <200) causes some numerical errors." << endl;
                            cerr << "I will continue the computation but you may try piOnePopSolver=SD or Eigen instead of the default DG." << endl;
                        }
                        if((piOnePopSDbool || piOnePopEigenbool) and Sam_AlleleNbr>200) {
                            cerr << "Using PinonePopSolver=SD or Eigen with a large number of allelic state (e.g. >200)" << endl;
                            cerr << "may be very slow and may causes some numerical errors." << endl;
                            cerr << "I will continue the computation but you may try piOnePopSolver=DG instead of SD or Eigen." << endl;
                        }
                        
                    }
                    if(thisMessage1==50){
                        cerr << "\n Migraine will continue the computation but this message will not be further printed out. \nTry to increase the state number of the SMM (e.g. +200) to remove this problem." << endl;
                    }
#ifdef DEBUG
                    if(thisMessage1<=50){
                        Effectif[a][i]--;
                        PopEffectif[a]--;
                        StateEffectif[i]--;
                        Size--;
                        cerr << "(!) Problem in weightSMM or Cumul computation " << endl;
                        //					if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr<<"Weight=AlternativeDenominator/Denominator*(Effectif[a][i]+1)*PIja/(PIia*(Effectif[a][j]+1)) ="<<Weight << endl;
                        //					 else
                        cerr << "Weight=(Effectif[a][i]+1)*PIja/(PIia*(Effectif[a][j]+1)) "<<Weight << endl;
                        if(DebugDG2 || Cumul>1.001 || std::isnan(Cumul) || std::isnan(factor)) {
                            cerr << "\tCumul+=factor*Mut->Pij[i][j]*PIia ="<<Cumul << endl;
                            cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
                            cerr << "\tThetaDIG=" << ThetaDIG << endl;
                            cerr << "\tDenominator=" << Denominator << endl;
                            cerr << "\tMut->Pij[i][j]="<< Mut->Pij[i][j] << endl;
                            cerr << "\tPIia = " << PIja << endl;
                            cerr << "\tPIja = " << PIja << endl;
                        }
                        Print();
                        //DebugPiSMM=true;
                        //PrintPi(Mig,Mut);
                        //DebugPiSMM=false;
                        fflush(stdout);
                        if(cinGetOnError) cin.get();
                        Effectif[a][i]++;
                        PopEffectif[a]++;
                        StateEffectif[i]++;
                        Size++;
                        if (DebugSigmaPi)
                            PrintPi(Mig,Mut);
                    }
#endif
					if(Weight <= 0 || std::isnan(Weight)) {
						cerr << "(!) Weight is set to numeric_limits<changetype>::epsilon()\n " << endl;
						Weight=numeric_limits<changetype>::epsilon();//RL->FR : bidouille pour r['e]soudre un probleme qui arive parfois en SMM, voir l'effet du SMM circulaire.
					}
				}
				//cout << "WeightSMM=" << Weight << endl;
				return Weight;
			}//fin if(random>..
		}// fin if (!(piOnePopSDbool || piOnePopEigenbool) || (i>=0 && i<Mut->Mut_AlleleNbr)) {

	}//fin for k=0;;;

} else if(mutationModel=="GSM" /*and Mut->pGSM<=0.99*/) {
	int compt=0;
	for(i=0;i<Sam_AlleleNbr;i++) if(Mut->Pij[i][j]>0) {

		//RL -> RL, RL -> FR
		//PIia sometimes very small because eigenvector and eigen values computations may not be very precise,
		//especially when pGSM is very low, no clean solution for the moment
		if (i!=j) {


			if(piOnePopEigenbool and localSDComputationbool){
#ifdef DEBUG
                if(false) cout << "PiOnePop SD used instead of Eigen because localSDComputationbool is true;" << endl;
#endif
				picalctype mab,pib=0.0;
				for(int k=0;k<Sam_AlleleNbr;k++) {
					mab=MatriceSD[Size][k][i];
					if(Effectif[0][k]>0.0) pib+=((Effectif[0][k]) * mab )/ (Size);
				}
				PIia=pib;

			} else {
				(this->*PicalcfnPtr)(Mig,Mut,i);
				PIia=PIs[i][a];
			}
		}


		if( PIia < 0.0 || (PIia-1.0)>numeric_limits<pirestype>::epsilon() || std::isnan(PIia)) {
			if(DebugDG2 or compt==Sam_AlleleNbr) {
				cout << "\nProblem in Change for mutation GSM PIia=" << PIia << " with i=" << i << "and a=" << a << endl;
				if(localSDComputationbool) cout << "\tSD algorithm used" << endl;
				cout << "\tChoose Pop=" << a << " and type=" << j << endl;
				cout << "\tThetaDIG=" << ThetaDIG << endl;
				cout << "\tpGSM=" << Mut->pGSM << endl;
				cout << "\tDenominator=" << Denominator << endl;
				cout << "\tMut->Pij[i][j]="<< Mut->Pij[i][j] << endl;
				cout << "\tPIia = " << PIia << endl;
				cout << "\tPIja = " << PIja << endl;
				//cout << "numeric_limits<double>::min()=" << numeric_limits<double>::min() << endl;
				//cout << "numeric_limits<long double>::min()=" << numeric_limits<long double>::min() << endl;
				Print();//print sample
				//DebugPiSMM=true;
				PrintPi(Mig,Mut);
				//DebugPiSMM=false;
				fflush(stdout);
				if(cinGetOnError) cin.get();
			}
			compt++;
			Lasti=i;
			//PIia=numeric_limits<pirestype>::epsilon();
		} else {
			Lasti=i;
			//calcul le facteur de la proposal col1 tab2 DIG04ii
			Cumul+=factor*Mut->Pij[i][j]*PIia;
#ifdef DEBUG
			if(DebugDG2 || Cumul>1.001 || std::isnan(Cumul) || std::isnan(factor)) {
				cout << "\nChange()/mutGSM -> ThetaDIG/Denominator*Mut->Pij[" << i << "][" << j << "]*PIia=" << factor*Mut->Pij[i][j]*PIia << endl;
				cout << "Cumul+=factor*Mut->Pij[i][j]*PIia; "<<Cumul << endl;
				cout << "\tChoose Pop=" << a << " and type=" << j << endl;
				cout << "\tThetaDIG=" << ThetaDIG << endl;
				cout << "\tpGSM=" << Mut->pGSM << endl;
				cout << "\tDenominator=" << Denominator << endl;
				cout << "\tMut->Pij[i][j]="<< Mut->Pij[i][j] << endl;
				cout << "\tPIia = " << PIia << endl;
				cout << "\tPIja = " << PIja << endl;
				//cout << "numeric_limits<double>::min()=" << numeric_limits<double>::min() << endl;
				//cout << "numeric_limits<long double>::min()=" << numeric_limits<long double>::min() << endl;
				Print();//print sample
				//DebugPiSMM=true;
				//PrintPi(Mig,Mut);
				//DebugPiSMM=false;
				if (DebugSigmaPi)
					PrintPi(Mig,Mut);
				fflush(stdout);
				if(cinGetOnError) cin.get();

		}
#endif
		if(Random<Cumul){
			//compute the weight of this mutation event col1 tab2 DIG04ii
			Weight=(Effectif[a][i]+1)*PIja;
			Weight/=(PIia*(Effectif[a][j]+1));
			//ajust the allelic range if changed
			if(i<AllelicRange[0]) AllelicRange[0]=i;
			if(i>AllelicRange[1]) AllelicRange[1]=i;
			//change sample configuration according to the mutation event
			Effectif[a][i]++;
			PopEffectif[a]++;
			StateEffectif[i]++;
			Size++;
#ifdef DEBUG
            if(AllelicRange[0]<0 || AllelicRange[1]>(Sam_AlleleNbr-1))
                cout << "Problem in SMM state Number :  min=" << AllelicRange[0] << " and max=" << AllelicRange[1] << endl;
			if(DebugDG){
				cout << "\t *GSM Mutation back in pop ";
				cout << a << ": type " << j << "->" << i << ", " ;
				if(DebugDG2) {
					cout << " cumulate=" << Cumul;
					cout << ", Random=" << Random << endl;
				}
				cout << "\t Weight=" << Weight << endl << endl;
			}
#endif
			if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
				cerr << "Numerical errors -> Weight=" << Weight << "  Cumul=" << Cumul << endl;
				if(!(piOnePopSDbool || piOnePopEigenbool) and Sam_AlleleNbr<200) {
                    cerr << "Using PinonePopSolver=DG with a small number of allelic state (e.g. <200) causes some numerical errors." << endl;
                    cerr << "I will continue the computation but you may try piOnePopSolver=SD or Eigen instead of the default DG." << endl;
                }
                if((piOnePopSDbool || piOnePopEigenbool) and Sam_AlleleNbr>200) {
                    cerr << "Using PinonePopSolver=SD or Eigen with a large number of allelic state (e.g. >200)" << endl;
                    cerr << "may be very slow and may causes some numerical errors." << endl;
                    cerr << "I will continue the computation but you may try piOnePopSolver=DG instead of SD or Eigen." << endl;
                }

#ifdef DEBUG
					Effectif[a][i]--;
					PopEffectif[a]--;
					StateEffectif[i]--;
					Size--;
					cerr << "(!) Problem in weightGSM or Cumul computation "<< endl;
					cerr << "Weight=(Effectif[a][i]+1)*PIja/(PIia*(Effectif[a][j]+1)) "<<Weight << endl;
					if(!(DebugDG2 || Cumul>1.05 || std::isnan(Cumul) || std::isnan(factor))) {
						cerr << "Cumul+=factor*Mut->Pij[i][j]*PIia; "<<Cumul << endl;
						cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
						cerr << "\tThetaDIG=" << ThetaDIG << endl;
						cerr << "\tpGSM=" << Mut->pGSM << endl;
						cerr << "\tDenominator=" << Denominator << endl;
						cerr << "\tMut->Pij[i][j]="<< Mut->Pij[i][j] << endl;
						cerr << "\tPIia = " << PIia << endl;
						cerr << "\tPIja = " << PIja << endl;
						Print();
						//DebugPiSMM=true;
						//PrintPi(Mig,Mut);
						//DebugPiSMM=false;
						fflush(stdout);
					}
					if (DebugSigmaPi)
						PrintPi(Mig,Mut);
					if(cinGetOnError) cin.get();
					Effectif[a][i]++;
					PopEffectif[a]++;
					StateEffectif[i]++;
					Size++;
					//}
#endif
					if(Weight <= 0 || std::isnan(Weight)) {
						cerr << "(!) Weight is set to numeric_limits<changetype>::epsilon()\n " << endl;
						Weight=numeric_limits<changetype>::epsilon();//RL->FR : bidouille pour r['e]soudre un probleme qui arive parfois en SMM, voir l'effet du SMM circulaire.
					}
				}
				//cout << "WeightSMM=" << Weight << endl;
				return Weight;
			}//fin if(random>..
		}//fin 	if( PIia < 0.0 || (PIia-1.0)>numeric_limits<pirestype>::epsilon() || std::isnan(PIia))
	}//fin for(i=0;i<Sam_AlleleNbr;i++)

} else if (mutationModel == "PIM"){

	for(i=0;i<Sam_AlleleNbr;i++){//i= ancestral state, j is descendant state
		if (i!=j) {// else PIja has already been computed for the coalescence case
			(this->*PicalcfnPtr)(Mig,Mut,i); // one call for each of Sam_AlleleNbr-1 alleles /* no obvious simplification as the systems solved differ for each i*/
		}
		if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) PIia=altPIs[i][a];
		else PIia=PIs[i][a];
		Lasti=i;
#ifdef DEBUG
		if(DebugDG2) cout << "Change()/mutKAM->PIia=" << PIia << endl;
#endif

// mutation j -> j is possible because PIs[j][a] is non zero
Cumul+=factor*Mut->Pij[i][j]*PIia;

#ifdef DEBUG
		if(DebugDG2) cout << "Change()/mutKAM->Cumul=" << Cumul << endl;
#endif
		if(Random<Cumul){
			//calcul le poid correpsondant a cette mutation col1 tab2 DIG04ii
			Weight=(Effectif[a][i]+1)*PIja;
			Weight/=(PIia*(Effectif[a][j]+1));
//			if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) Weight*=AlternativeDenominator/Denominator;
//			  else if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) Weight*=proposalCorrectionFactor;//to correct the cumul probabilities so that they sum to 1, 19012011
			if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) Weight*=proposalCorrectionFactor;//to correct the cumul probabilities so that they sum to 1, 19012011
			//effectue les changement sur l'echantillon
			Effectif[a][i]++;
			PopEffectif[a]++;
			StateEffectif[i]++;
			Size++;
#ifdef DEBUG
			if(DebugDG){
				cout << "\t *Mutation back in pop ";
				cout << a << ": type " << j << "->" << i << ", " ;
				if(DebugDG2) cout << " cumulate=" << Cumul;
				//printf("\n\t Cumulate=%g",Cumul);
				if(DebugDG2) cout << ", Random=" << Random << endl;
				//printf("; Random=%g ;",Random);
				//printf("\n\t Weight=%g ",Weight);
				cout << "\t Weight=" << Weight << endl;
			}
#endif
			if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
				cerr << "Numerical errors -> Weight=" << Weight << "  Cumul=" << Cumul << endl;
#ifdef DEBUG
	            cerr<<"(!) Problem in weightPIM or Cumul computation "<< endl;
//			    if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr<<"Weight=AlternativeDenominator/Denominator*(Effectif[a][i]+1)*PIja/(PIia*(Effectif[a][j]+1)) = "<<Weight << endl;
//			    else
				cerr<<"Weight=(Effectif[a][i]+1)*PIja/(PIia*(Effectif[a][j]+1)) = "<<Weight << endl;
                cerr<<"Cumul+=factor*Mut->Pij[i][j]*PIia = "<<Cumul << endl;
                cerr<<"factor="<<factor << endl;
                cerr<<"Mut->Pij[i][j]="<< Mut->Pij[i][j] << endl;
                cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
	            cerr << "\tEffectif[a][i] = " << Effectif[a][i] << endl;
	            cerr << "\tEffectif[a][j] = " << Effectif[a][j]<< endl;
    	        cerr << "\tPIia = " << PIia << endl;
    	        cerr << "\tPIja = " << PIja << endl;
    	        Print();
                //DebugPiSMM=true;
                //PrintPi(Mig,Mut);
                fflush(stdout);
				//DebugPiSMM=false;
				if (DebugSigmaPi)
					PrintPi(Mig,Mut);
	            if(cinGetOnError) cin.get();
#endif
			}
			//cout << "WeightPIM=" << Weight << endl;
			return Weight;
		}//end if(Random<Cumul)
	}//end for(i=0;i<Sam_AlleleNbr;i++)
}//fin else PIM
else if (mutationModel == "ISM" || mutationModel == "SNP"){
	if (singleGeneISM == migPlusISMtype1) {//into a NEW type, donc pas en SNP

#ifdef DEBUG
        //avirer RL->RL
        if(mutationModel == "SNP") {
            cerr << "Probleme with ISM/SNP mtations 1" << endl;
            if(cinGetOnError) cin.get();
            exit(-1);
        }
#endif

		Cumul += ThetaDIG / (Mig->Ma[a] + ThetaDIG);
		if (Random < Cumul) {
			Weight = nZero[a] * (Mig->Ma[a] + ThetaDIG) / Denominator;

			//	accounting for a Type 1 mutation
			siteMutCountVec[*siteJ - 1] = 0;
			mutPathMatrix[j].erase(siteJ);

			//	restoring the gene (now having mutated into a NEW type) taken out of the sample
			++Effectif[a][j];
			++PopEffectif[a];
			++StateEffectif[j];
			++Size;

#ifdef DEBUG
			if(DebugDGISM){
				cout << "\t *migPlusISMtype1 (mutation into inexistent type) back in pop ";
				cout << a << ": old type " << j << "-> new type " << j << "; Weight=" <<  Weight << "\n\n" << endl;
			}
#endif

			if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
				cerr << "Numerical errors -> Weight=" << Weight << "Cumul=" << Cumul << endl;
#ifdef DEBUG
				cerr<<"\n\n(!) Problem in weight or cumul computation when looking for an ISM Type1 mutation event" << endl;
				cerr<<"Weight = nZero[a] * (Mig->Ma[a] + ThetaDIG) / Denominator = "<<Weight << endl;
				cerr<<"Cumul = ThetaDIG / (Mig->Ma[a] + ThetaDIG) =  "<<Cumul << endl;
				cerr<<"Mig->PopSize[a] = "<<Mig->PopSize[a] << endl;
				cerr<<"Denominator = "<< Denominator << endl;
				cerr << "\tChoose Pop = " << a << " and type = " << j << endl;
				cerr << "\tPopEffectif[a] = " << PopEffectif[a] << endl;
				cerr << "\tEffectif[a][j] = " << Effectif[a][j]<< endl;
				cerr << "\tnZero = " << nZero[a] << endl;
				Print();
				fflush(stdout);
				if (DebugSigmaPi)
					PrintPi(Mig,Mut);
				if (cinGetOnError) cin.get();
#endif
			}
			return Weight;
		}
	}
	else if (singleGeneISM == migPlusISMtype2) {// soit mutation vers type existant, soit vers type non existant comme ci dessus...
		if(mutationModel=="SNP") Cumul += 1.0 / (Mig->Ma[a] + 1.0); // RL->CBR pourquoi il n'y a pas r[a]/(Denom*nzero[a]), fait dans choosepopandType()?
            else Cumul += ThetaDIG / (Mig->Ma[a] + ThetaDIG);// RL->CBR pourquoi il n'y a pas r[a]/(Denom*nzero[a]), fait dans choosepopandType()?
		if (Random < Cumul) {
			//	accounting for a Type 2 mutation site
			siteMutCountVec[*siteJ - 1] = 0;
			mutPathMatrix[j].erase(siteJ);

			//	searching for the existing haplotype into which the chosen type has mutated
			size_t pathCounter;

			//	IF the existing type is mutating into the MRCA then look for an already existing MRCA
			if (mutPathMatrix[j][0] == 0) {
				for (pathCounter = 0; pathCounter < mutPathMatrix.size(); pathCounter++)
					if ((pathCounter != (unsigned int) j) && (mutPathMatrix[pathCounter] == mutPathMatrix[j]) && (Effectif[a][pathCounter] > 0))
						break;
			}
			//	ELSE look for a similar non-MRCA type
			else {
				for (pathCounter = 0; pathCounter < mutPathMatrix.size(); pathCounter++)
					if ((pathCounter != (unsigned int) j) && (mutPathMatrix[pathCounter] == mutPathMatrix[j]))
						break;
			}

			//	CBR: in this case a migPlusISMtype2 mutation gives rise to a previously
			//	inexistent type in the sample (like for migPlusISMtype1)
			if (pathCounter == mutPathMatrix.size()) {

#ifdef DEBUG
                //avirer RL->RL
                if(mutationModel == "SNP") {
                    cerr << "Probleme with ISM/SNP mtations 2" << endl;
                    if(cinGetOnError) cin.get();
                    exit(-1);
                }
#endif

				Weight = nZero[a] * (Mig->Ma[a] + ThetaDIG) / Denominator;

				//	restoring the gene (now having mutated into a NEW type) taken out of the sample
				++Effectif[a][j];
				++StateEffectif[j];
				++PopEffectif[a];
				++Size;
#ifdef DEBUG
				if (DebugDGISM) {
					cout << "\t *migPlusISMtype2 (BUT mutation into inexistent type!!) back in pop ";
					cout << a << ": old type " << j << "-> new type " << j << "; Weight=" <<  Weight << "\n\n" << endl;
				}
#endif
			}
			//	CBR: this case corresponds to the classic ISM Type 2 mutation (i.e. from type x -> type y)
			else {
				//	eliminates all traces of the chosen type's segregation counts
				for (vector<int>::iterator it = mutPathMatrix[j].begin(); *it != 0; it++)
					--siteMutCountVec[*it - 1];

				//	assign zero to indicate that this type is non-existent from now onwards
				mutPathMatrix[j].assign(1,0);
                if(mutationModel=="SNP") Weight = ((Effectif[a][pathCounter] + 1.0) * nZero[a] * (Mig->Ma[a] + 1)) / Denominator;//RL->CBR pourquoi /Denom et pas /r[a] comme dans tableau 4 DG2004ii
				  else Weight = ((Effectif[a][pathCounter] + 1.0) * nZero[a] * (Mig->Ma[a] + ThetaDIG)) / Denominator;//RL->CBR pourquoi /Denom et pas /r[a] comme dans tableau 4 DG2004ii

				//	restoring the gene (now having mutated into an ALREADY EXISTING type) taken out of the sample
				++Effectif[a][pathCounter];
				++StateEffectif[pathCounter];
				++PopEffectif[a];
				++Size;
#ifdef DEBUG
				if (DebugDGISM) {
					cout << "\t *migPlusISMtype2 Mutation back in pop ";
					cout << a << ": type " << j << "->" << pathCounter << "; Weight=" <<  Weight << "\n\n" << endl;
				}
#endif
			}

			if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
				cerr << "Numerical errors -> Weight=" << Weight << "Cumul=" << Cumul << endl;
#ifdef DEBUG
				cerr<<"\n\n(!) Problem in weight or cumul computation when looking for an ISM Type2 mutation event" << endl;
				if (mutPathMatrix[j][0] == 0)
					cerr<<"Weight = nZero[a] * (Mig->Ma[a] + ThetaDIG) / Denominator = "<<Weight << endl;
				else
					cerr<<"Weight = ((Effectif[a][pathCounter] + 1.0) * nZero[a] * (Mig->Ma[a] + ThetaDIG)) / Denominator = "<<Weight << endl;
				cerr<<"Cumul = ThetaDIG / (Mig->Ma[a] + ThetaDIG) =  "<<Cumul << endl;
				cerr<<"Mig->PopSize[a] = "<<Mig->PopSize[a] << endl;
				cerr<<"Denominator = "<< Denominator << endl;
				cerr << "\tChoose Pop = " << a << " and type = " << j << endl;
				cerr << "\tPopEffectif[a] = " << PopEffectif[a] << endl;
				cerr << "\tEffectif[a][j] = " << Effectif[a][j]<< endl;
				cout << "\tnZero = " << nZero[a] << endl;
				Print();
				fflush(stdout);
				if (DebugSigmaPi)
					PrintPi(Mig,Mut);
				if (cinGetOnError) cin.get();
#endif
			}
			return Weight;
		}
	}
}	//	end of ISM

/******** Migration ********/
if (mutationModel == "ISM" || mutationModel == "SNP") {
#ifdef DEBUG
    if(mutationModel == "SNP") {
        cerr << "Probleme with ISM/SNP migration event, check what to do for SNPs" << endl;
        if(cinGetOnError) cin.get();
        exit(-1);
    }
#endif

	if (Effectif[a][j] > 0) {
		for (b = 0; b < PSON; b++) {
			if (b != a) {
				PIjb = PIs[j][b];
				Cumul += (PIjb * Mig->Mab[a][b] * nZero[a]) / (PIja * Denominator);//RL->CBR pourquoi il y a nzero[a] et les pi?
				if (Random < Cumul) {
					Weight = (PIja * (Effectif[b][j] + 1.0) * (PopEffectif[a] + 1.0)) / (PIjb * (PopEffectif[b] + 1.0) * (Effectif[a][j] + 1.0));

					//	restoring the gene (now having migrated to pop b) taken out of the sample
					++Effectif[b][j];
					++PopEffectif[b];
					++StateEffectif[j];
					++Size;
#ifdef DEBUG
					if (DebugDGISM) {
						printf("\t *Normal ISM Migration back of type %d : pop %d -> %d;\n\t Cumulate=%g; Random=%g ;",j,a,b,(double) Cumul,(double) Random);
						cout << "\n\t Weight=" << Weight << "\n\n" << endl;
					}
#endif

		            if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
						cerr << "Numerical errors -> Weight=" << Weight << "Cumul=" << Cumul << endl;
#ifdef DEBUG
		               cerr<<"(!) Problem in weight or Cumul computation: normal ISM migration" << endl;
		                cerr<<"Weight = (PIja * (Effectif[b][j] + 1.0) * (PopEffectif[a] + 1.0)) / (PIjb * (PopEffectif[b] + 1.0) * (Effectif[a][j] + 1.0)) = "<<Weight << endl;
		                cerr<<"Cumul += (PIjb * Mig->Mab[a][b] * nZero[a]) / (PIja * Denominator) = "<<Cumul << endl;
		                cerr<<"Mig->Mab[a][b]="<<Mig->Mab[a][b] << endl;
		                cerr<<"Denominator="<< Denominator << endl;
		                cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
			            cerr << "\tEffectif[b][j] = " << Effectif[b][j] << endl;
			            cerr << "\tPopEffectif[a] = " << PopEffectif[a]<< endl;
			            cerr << "\tEffectif[a][j] = " << Effectif[a][j] << endl;
			            cerr << "\tPopEffectif[b] = " << PopEffectif[b] << endl;
			    		cout << "\tnZero = " << nZero[a] << endl;
		    	        cerr << "\tPIjb = " << PIjb << endl;
		    	        cerr << "\tPIja = " << PIja << endl;
		    	        Print();
		                PrintPi(Mig,Mut);
		                fflush(stdout);
						if (DebugSigmaPi)
							PrintPi(Mig,Mut);
		                if(cinGetOnError) cin.get();
#endif
		            }
					return Weight;
				}
			}
		}
	}
	else if ((singleGeneISM == migPlusISMtype1) || (singleGeneISM == migPlusISMtype2)) {
#ifdef DEBUG
        //avirer RL
        if(mutationModel == "SNP") {
            cerr << "Probleme with ISM/SNP migration event, check what to do for SNPs" << endl;
            if(cinGetOnError) cin.get();
            exit(-1);
        }
#endif

		for (b = 0; b < PSON; b++) {
			if (b != a) {
				Cumul += Mig->Mab[a][b] / (Mig->Ma[a] + ThetaDIG);
				if (Random < Cumul) {
					Weight = ((PopEffectif[a] + 1.0) * nZero[a] * (Mig->Ma[a] + ThetaDIG)) / ((PopEffectif[b] + 1.0) * Denominator);

					//	restoring the gene (now having migrated to pop b) taken out of the sample
					++Effectif[b][j];
					++PopEffectif[b];
					++StateEffectif[j];
					++Size;
#ifdef DEBUG
					if (DebugDGISM) {
						printf("\t *migPlusISMtype1/migPlusISMtype2 back of type %d : pop %d -> %d;\n\t Cumulate=%g; Random=%g ;",j,a,b,(double) Cumul,(double) Random);
						cout << "\n\t Weight=" << Weight << "\n\n" << endl;
					}
#endif

		            if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
						cerr << "Numerical errors -> Weight=" << Weight << "Cumul=" << Cumul << endl;
#ifdef DEBUG
		               cerr<<"(!) Problem in weight or Cumul computation: migPlusISMtype1 or migPlusISMtype2" << endl;
		                cerr<<"Weight = ((PopEffectif[a] + 1.0) * nZero[a] * (Mig->Ma[a] + ThetaDIG)) / ((PopEffectif[b] + 1.0) * Denominator) = "<<Weight << endl;
		                cerr<<"Cumul += Mig->Mab[a][b] / (Mig->Ma[a] + ThetaDIG) = "<<Cumul << endl;
		                cerr<<"Mig->Mab[a][b]="<<Mig->Mab[a][b] << endl;
		                cerr<<"Denominator="<< Denominator << endl;
		                cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
			            cerr << "\tEffectif[b][j] = " << Effectif[b][j] << endl;
			            cerr << "\tPopEffectif[a] = " << PopEffectif[a]<< endl;
			            cerr << "\tEffectif[a][j] = " << Effectif[a][j] << endl;
			            cerr << "\tPopEffectif[b] = " << PopEffectif[b] << endl;
			    		cout << "\tnZero = " << nZero[a] << endl;
		    	        cerr << "\tPIjb = " << PIjb << endl;
		    	        cerr << "\tPIja = " << PIja << endl;
		    	        Print();
		                PrintPi(Mig,Mut);
		                fflush(stdout);
						if (DebugSigmaPi)
							PrintPi(Mig,Mut);
		                if(cinGetOnError) cin.get();
#endif
		            }
					return Weight;
				}
			}
		}
	}
	else if (singleGeneISM == migOnly) {
#ifdef DEBUG
        //avirer RL
        if(mutationModel == "SNP") {
            cerr << "Probleme with ISM/SNP migration 1" << endl;
            if(cinGetOnError) cin.get();
            exit(-1);
        }
#endif
		for (b = 0; b < PSON; b++) {
			if (b != a) {
				Cumul += Mig->Mab[a][b] / Mig->Ma[a];
				if (Random < Cumul) {
					Weight = ((PopEffectif[a] + 1.0) * (Effectif[b][j] + 1.0) * nZero[a] * Mig->Ma[a]) / ((PopEffectif[b] + 1.0) * Denominator);

					//	restoring the gene (now having migrated to pop b) taken out of the sample
					++Effectif[b][j];
					++PopEffectif[b];
					++StateEffectif[j];
					++Size;
#ifdef DEBUG
					if (DebugDGISM) {
						printf("\t *migOnly back of type %d : pop %d -> %d;\n\t Cumulate=%g; Random=%g ;",j,a,b,(double) Cumul,(double) Random);
						cout << "\n\t Weight=" << Weight << "\n\n" << endl;
					}
#endif

		            if (Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
						cerr << "Numerical errors -> Weight=" << Weight << "Cumul=" << Cumul << endl;
#ifdef DEBUG
		               cerr<<"(!) Problem in weight or Cumul computation: migOnly" << endl;
		                cerr<<"Weight = ((PopEffectif[a] + 1.0) * (Effectif[b][j] + 1.0) * nZero[a] * Mig->Ma[a]) / ((PopEffectif[b] + 1.0) * Denominator) = "<<Weight << endl;
		                cerr<<"Cumul += Mig->Mab[a][b] / Mig->Ma[a] = "<<Cumul << endl;
		                cerr<<"Mig->Mab[a][b]="<<Mig->Mab[a][b] << endl;
		                cerr<<"Denominator="<< Denominator << endl;
		                cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
			            cerr << "\tEffectif[b][j] = " << Effectif[b][j] << endl;
			            cerr << "\tPopEffectif[a] = " << PopEffectif[a]<< endl;
			            cerr << "\tEffectif[a][j] = " << Effectif[a][j] << endl;
			            cerr << "\tPopEffectif[b] = " << PopEffectif[b] << endl;
			    		cout << "\tnZero = " << nZero[a] << endl;
		    	        cerr << "\tPIjb = " << PIjb << endl;
		    	        cerr << "\tPIja = " << PIja << endl;
		    	        Print();
		                PrintPi(Mig,Mut);
		                fflush(stdout);
						if (DebugSigmaPi)
							PrintPi(Mig,Mut);
		                if(cinGetOnError) cin.get();
#endif
		            }
					return Weight;
				}
			}
		}
	}
}//	end (mutationModel == "ISM")
else {
	for(b=0;b<PSON;b++) {
		if(b!=a) {
			if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) PIjb=altPIs[j][b];
			else PIjb=PIs[j][b];
#ifdef DEBUG
			if(DebugDG) cout << "change/mig-> PIjb[" << j << "][" << b << "]=" << PIjb << endl;
#endif
			//calcul le facteur de la proposal col1 tab2 DIG04ii
			Cumul+=Mig->Mab[a][b]*PIjb/Denominator;
			if(Random<Cumul){
				//calcul le poid correspondant a cette migration col1 tab2 DIG04ii
				Weight=(Effectif[b][j]+1.0)*(PopEffectif[a]+1.0)*PIja;
				Weight/=(Effectif[a][j]+1.0)*(PopEffectif[b]+1.0)*PIjb;
                //ajuste l'echantilllon
				Effectif[b][j]++;
				PopEffectif[b]++;
				StateEffectif[j]++;
				Size++;
	//			disp_coal_events++;
#ifdef DEBUG
				if(DebugDG){
					printf("\t *Migration back of type %d : pop %d -> %d;\n\t Cumulate=%g; Random=%g ;",j,a,b,(double) Cumul,(double) Random);
					//printf("\n\t Weight=%g ",Weight);
					cout << "\n\t Weight=" << Weight << endl;
					//cout << "Migration back, type \t" << j;
					//cout << ":\t "<< a << '\t';
					//cout << " -> \t" << b << '\t';
					if(DebugDG2) cout << " cumulate=" << Cumul;
				    if(DebugDG2) cout << ", Random=" << Random << endl;
				}
#endif
#ifdef ASYMPTO_PAR_DISP_EVENTS
				disp_events++;
#endif
	//cout<<"fin mig"; cout.flush();
	/*if (Weight>1) {cout<<"fin mig "<<Weight; getchar();
		cout<<PIja<<" "<<PIjb;
		ostream_vector(PIs[j],cout);
		getchar();
		}*/
	            if (/*Weight>1 ||*/ Weight <= 0 || std::isnan(Weight) || std::isnan(Cumul) || Cumul>1.001) {
					cerr << "Numerical errors -> Weight=" << Weight << "Cumul=" << Cumul << endl;
#ifdef DEBUG
	               cerr<<"(!) Problem in weightMig or Cumul computation" << endl;
	                cerr<<"Weight=Effectif[b][j]+1.0)*(PopEffectif[a]+1.0)*PIja/(Effectif[a][j]+1.0)*(PopEffectif[b]+1.0)*PIjb = "<<Weight << endl;
	                cerr<<"Cumul+=Mig->Mab[a][b]*PIjb/Denominator = "<<Cumul << endl;
	                cerr<<"Mig->Mab[a][b]="<<Mig->Mab[a][b] << endl;
	                cerr<<"Denominator="<< Denominator << endl;
	                cerr << "\tChoose Pop=" << a << " and type=" << j << endl;
		            cerr << "\tEffectif[b][j] = " << Effectif[b][j] << endl;
		            cerr << "\tPopEffectif[a] = " << PopEffectif[a]<< endl;
		            cerr << "\tEffectif[a][j] = " << Effectif[a][j] << endl;
		            cerr << "\tPopEffectif[b] = " << PopEffectif[b] << endl;
	    	        cerr << "\tPIjb = " << PIjb << endl;
	    	        cerr << "\tPIja = " << PIja << endl;
	    	        Print();
	                //DebugPiSMM=true;
	                //PrintPi(Mig,Mut);
	                fflush(stdout);
					//DebugPiSMM=false;
					if (DebugSigmaPi)
						PrintPi(Mig,Mut);
	                if(cinGetOnError) cin.get();
#endif
	            }
	             //cout << "WeightMigration=" << Weight << endl;
				return Weight;
			}//end if(Random<Cumul)
		}//end if(b!=a)
	}//end for(b=0;b<PSON;b++)
}//	end (mutationModel != "ISM")
/*si on n'est pas sorti par un return ici c'est soit qu'il y a une bug soit que
cumul =1+/- (erreurs num['e]riques) et random> cumul.
Auquel cas while() recommence le calcul avec un nouveau random.*/

#ifdef DEBUG

// petit affichage pour voir la grandeur des impr['e]cisions num['e]riques:
cerr<<"\nRandom="<<Random<<" Cumul="<<Cumul<<" Imprecision=" <<(Random-Cumul)/numeric_limits<cumultype>::epsilon()<<endl;

if(Cumul<0.99 || std::isnan(Cumul)) {
//si on est pas sorti ici c que ca plante : cumul ne somme pas a 1
	cerr << "\nBug in Change(), Cumul is " << Cumul << endl;
	cerr << "random is " << Random << endl;
	cerr << "Choose Pop (a) =" << a << " and Allele (j) ="<<j<<endl;
	cerr << "PopEffectif[a] = " << PopEffectif[a] << endl;
	cerr << "Mig->PopSize[a] = " << Mig->PopSize[a] << endl;
	cerr << "Mig->Ma[a] = " << Mig->Ma[a] << endl;
	if(PopSampledOrNot==2) {
			if(a==0) b=1; else b=0;
			cerr << "Mig->Ma[b] = " << Mig->Ma[b] << endl;
	}
	if(mutationModel=="GSM") cerr << "\tpGSM=" << Mut->pGSM << endl;
	cerr << "ThetaDIG = " << ThetaDIG << endl;
//	if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr << "PopSize=AncThetaDIG/ThetaDIG = " << AncThetaDIG/ThetaDIG << endl;
	cerr << "Denominator is " << Denominator << endl;
//	if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr << "AlternativeDenominator is " << AlternativeDenominator << endl;
	cerr << "PIja is" << PIja << endl;
	cerr << "Denom.remainder (PopEffectif[a]/Mig->PopSize[a]+Mig->Ma[a]+ThetaDIG) is " << (PopEffectif[a]/Mig->PopSize[a]+Mig->Ma[a]+ThetaDIG) << endl;
//	if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool) cerr << "AlternativeDenom.remainder (PopEffectif[a]/(AncThetaDIG/ThetaDIG)+Mig->Ma[a]+ThetaDIG) is " << (PopEffectif[a]/(AncThetaDIG/ThetaDIG)+Mig->Ma[a]+ThetaDIG) << endl;
	if(PSON>1) cout << "Mab[1][0]="<< Mig->Mab[1][0] << endl;
	cerr << "Lasti is " << Lasti << endl;
	cerr << "factor is " << factor << endl;
	cerr << "Mut->Pij["<< Lasti << "][" << j << "] is " << Mut->Pij[Lasti][j] << endl;
	cerr << "Mut->Pij["<< j << "][" << Lasti << "] is " << Mut->Pij[j][Lasti] << endl;
	cerr << "PIia is " << PIia << endl;
	Print();
	if(true) DebugPiSMM=true;
	PrintPi(Mig,Mut);
	if(true) DebugPiSMM=false;

	//cout<<'\a'<<'\a';//<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a'<<'\a';
	cerr << endl;
	//DebugDG=false;
	//DebugDG2=false;
	//DebugChoosePopType=true;
	fflush(stderr);
//	if (Cumul<0.95 or cinGetOnError) {
//		cerr << "Press any key to resume." << endl;
//		fflush(stderr);
//		cin.get();
//	}
	//exit(-1);
}
#endif

            if(Cumul<0.95) break;
} //fin inner while(true) sur évenement APRES choose pop and type : on recommence si on est pas sorti par un return
/** two modes of exit from while(true): either a return for event correctly chosen, or a break if cumul is too low. In the latter case
we go backward by re-putting the gene into the sample : **/
        cerr<<"\n (!!) Cumul<0.95: Random="<<Random<<" Cumul="<<Cumul<<" Imprecision=" <<(Random-Cumul)/numeric_limits<cumultype>::epsilon()<<endl;
        PopEffectif[a]++;
        Effectif[a][j]++;
        StateEffectif[j]++;
        Size++;

    	if(countSeq == 100) {
            cerr << "\nError! After 100 trials Cumul<0.95 in CSample::Change."<<endl;
            cerr << "\n Aborting computation for this point by setting loglik=NaN."<<endl;
            cerr << "\n Contact the author if this message appears more than a few times."<<endl;
            if(cinGetOnError) {
                cerr << "\n...Press any key to resume...\n" << endl;
                cin.get();
            }
            return(-2.0);
//    		exit(-1);
    	}

} //fin de outer while(true) sur choose pop and type et évenements. pour prb GSM, a résoudre!

return(-1.0);
}
/******************* fin Change *****************/

//***********************************************************
// CBR (10.05.2012) procedure for calculating the proposal quantity Q(k|S,n) under ISM (Hobolth et al. (2008) Eq. 10)
//	N.B. CSample::calcHobolthProposal chooses a haplotype DIFFERENTLY from CSample::Change
changetype CSample::calcHobolthChange(CMigMod *Mig,CMutMod *Mut) {

#ifdef DEBUG
	if (DebugDGISM) {
		cout << "-----------------------------------------------------------------------------------\n";
		cout << "Mutation paths before CSample::calcHobolthChange:\n";
		PrintSeq();
		cout << "Pops before CSample::calcHobolthChange:\n";
		Print();
	}
#endif

	events++;

	//	code for returning the weight for the last individual of the MRCA type
	if (Size == 1) {
		Size = 0;
		return 1.0;//RL -> FR, RL -> CBR, ca me parait bizarre, distrib sttationnaire des haplotypes?? sans doute pas très important
	}

	vector<changetype> QHobolth(Sam_AlleleNbr,0.0); // CBR (12.05.2013) stores Q(k|S,n) (Hobolth et al. (2008) Eq. 10)
	vector<changetype> pThetaOfD(*max_element(DmVec.begin(),DmVec.end())+1,0.0); // CBR (03.06.2013) stores p_theta(d) (Hobolth et al. (2008) Eq. 11)
	size_t chosenHapIndex, MRCAIndex;
	bool MRCAIndexSetBool = false;

	changetype Weight,TCompute;
	cumultype rnd, proposalCDF = 0.0;

	vector<HOBEventType> validHapsForEvent(Sam_AlleleNbr, noHOBEvent);
	vector<vector<int>::iterator> validHapMutSiteIndexPtr;

	//	here we use ONLY the time computation part of CSample::ChoosePopType for IM and varsize models
	TCompute = TimeCompute(Mig,Mut,true);
	if (TCompute != 0)
		return TCompute;

	//	assigning pThetaOfD once for each segregating site
	for (size_t i = 0; i < DmVec.size(); i++) {
		if ((DmVec[i] > 0) && (pThetaOfD[DmVec[i]] == 0.0))
			pThetaOfD[DmVec[i]] = calcHobolth_pThetaOfDm(Mut->ThetaDIG, DmVec[i]);
	}

// STEP 1: check for the haplotypes compatible with the possible events (or IF we have reached the top of the tree)
//---------------------------------------------------------------------------------------------------------------
	//	under ideal conditions nZero remains PopEffectif[0] as no gene is taken out of the sample like in CSample::Change
	nZero[0] = PopEffectif[0];

	//	CBR: several checks and event assignments for singleton genes if or not they are capable of mutation
	//	1°	identifies a site and assigns ISMtype1 mutation when the chosen gene contains at least 2 unique mutation sites
	//	2°	identifies a site and assigns ISMtype2 mutation when the chosen gene contains 1 unique mutation site
	//	3°	corrects nZero (i.e. no mutation event) when the chosen singleton gene contains no unique mutation sites
	for (int k = 0; k < Sam_AlleleNbr; k++) {
		//	initializing validHapMutSiteIndexPtr with the "past-the-end" element of mutPathMatrix[k]
		validHapMutSiteIndexPtr.push_back(mutPathMatrix[k].end());

		//	singleton check for EVERY HAPLOTYPE (N.B. here no gene is taken out of the sample like in CSample::Change)
		if (StateEffectif[k] == 1) {
			vector<int>::iterator it1, it2;
			//	looping over the list of segregated sites for each haplotype
			for (it1 = mutPathMatrix[k].begin(); *it1 != 0; it1++) {
				//	IF an unique mutation is found for the haplotype
				if (siteMutCountVec[*it1 - 1] == 1) {
					for (it2 = it1+1; *it2 != 0; it2++) {
						//	check for a second mutation
						if (siteMutCountVec[*it2 - 1] == 1) {
							//	found a second unique mutation site = mutISMtype1
							validHapMutSiteIndexPtr[k] = it1;
							validHapsForEvent[k] = mutISMtype1;
							break;
						}
					}
					if (*it2 == 0) {
						//	found only a single unique mutation site = mutISMtype2
						validHapMutSiteIndexPtr[k] = it1;
						validHapsForEvent[k] = mutISMtype2;
					}
					//	break the inner for loop as the haplotype "k" corresponds to a possible event
					break;
				}
			}
			//	no unique mutation sites found = no possible events for this haplotype
			if (*it1 == 0)
				--nZero[0];
		}
		else if (StateEffectif[k] >= 2)
			validHapsForEvent[k] = coalescence;

		if ((mutPathMatrix[k][0] == 0) && (StateEffectif[k] > 1)) {
			MRCAIndex = k;
			MRCAIndexSetBool = true;
		}
	}

#ifdef DEBUG
	if (DebugDGISM) {
		cout << "\n";
		if (MRCAIndexSetBool)
			cout << "MRCAIndexSetBool = true \n";
		else
			cout << "MRCAIndexSetBool = false \n";

		cout << "pThetaOfD = \n";
		for (size_t j = 0; j < pThetaOfD.size(); j++)
			cout << pThetaOfD[j] << " ";
		cout << "\n";
	}
#endif

	//	IF the pop contains only (but more than 1) MRCA types then the proposal = 1
	if (MRCAIndexSetBool && (StateEffectif[MRCAIndex] == PopEffectif[0])) {
		QHobolth[MRCAIndex] = 1.0;
		chosenHapIndex = MRCAIndex;
	}
	//	IF more than two "event-compatible" haplotypes then continue with the Hobolth proposal stuff
	else {
// STEP 2: calculate the proposal "quantities" for each "event-compatible" haplotype
//---------------------------------------------------------------------------------------------------------------
//		//	IF the pop contains NOT only (but more than 1) MRCA types then don't coalesce the MRCA types as yet
//		if (MRCAIndexSetBool && (StateEffectif[MRCAIndex] < PopEffectif[0]))
//			validHapsForEvent[MRCAIndex] = noHOBEvent;

		for (int k = 0; k < Sam_AlleleNbr; k++) {
			//	IF the haplotype is invalid then skip it for proposal calculation
			if (validHapsForEvent[k] != noHOBEvent) {
				//	loop over the site segregation count (HAPLOTYPE-WISE) vector
				for (size_t i = 0; i < siteMutCountVec.size(); i++) {
					//	IF a size is not segregated then skip the site
					if (siteMutCountVec[i] > 0) {
						vector<int>::iterator it;
						//	loop over with the iterator pointing to the haplotype's site segretion indexes
						for (it = mutPathMatrix[k].begin(); it != mutPathMatrix[k].end(); it++) {
							//	satisfies Hobolth et al. (2008) Eq. 9 (Skm = 1)
							if ( ( (int) (i + 1) ) == *it) {
								QHobolth[k] += (changetype) pThetaOfD[DmVec[i]] * StateEffectif[k] / DmVec[i];
								break;
							}
						}
						//	satisfies Hobolth et al. (2008) Eq. 9 (Skm = 0)
						if (it == mutPathMatrix[k].end())
							QHobolth[k] += (changetype) (1 - pThetaOfD[DmVec[i]]) * StateEffectif[k] / (Size - DmVec[i]);
					}
				}
			}
		}

// STEP 3: transform the proposal "quantities" into proposal probabilities one by one and choose a haplotype
//---------------------------------------------------------------------------------------------------------------
		//	sum the proposal quantities
		changetype sumProp = accumulate(QHobolth.begin(),QHobolth.end(),0.0);

		if (sumProp == 0.0) {
			cerr << "(!) Error in CSample::calcHobolthChange: the sum of the proposal quantities is zero." << endl
				 << "Aborting Migraine...Press any key to exit!" << endl;
			if (cinGetOnError)
				cin.get();
			exit(-1);
		}

		rnd = alea();

#ifdef DEBUG
		if (DebugDGISM) {
			cout << "\nrand = " << rnd << "\n";
			cout << "QHobolth = \n";
			for (size_t j = 0; j < QHobolth.size(); j++)
				cout << QHobolth[j] / sumProp << " ";
			cout << "\n";
		}
#endif
		//	stepwise creation of a discrete proposal CDF and then choose the haplotype
        chosenHapIndex = 0; //RL->CBR initialisation OK?
		for (size_t k = 0; k < QHobolth.size(); k++) {
			if (validHapsForEvent[k] != noHOBEvent) {
				QHobolth[k] /= sumProp;
				proposalCDF += QHobolth[k];
				if (proposalCDF > rnd) {
					chosenHapIndex = k;
					break;
				}
			}
		}
	}

#ifdef DEBUG
		if (DebugDGISM)
			cout << "chosenHapIndex = " << chosenHapIndex << "\n\n";
#endif

// STEP 4: calculate and return the event weight
//---------------------------------------------------------------------------------------------------------------
	cumultype popScaledTotalEventRate = (PopEffectif[0] - 1) / Mig->PopSize[0] + Mut->ThetaDIG;
    if (validHapsForEvent[chosenHapIndex] == coalescence) {
    	Weight = (changetype) 1 / QHobolth[chosenHapIndex] * (StateEffectif[chosenHapIndex] - 1) / Mig->PopSize[0] / popScaledTotalEventRate;

		//	account for the coalescent event
    	--PopEffectif[0];
    	--Effectif[0][chosenHapIndex];
    	--StateEffectif[chosenHapIndex];
    	--Size;

		//	coalescence event => all nucleotide positions containing the mutations of the chosen HAPLOTYPE are reduces by 1
		for (vector<int>::iterator it = mutPathMatrix[chosenHapIndex].begin(); *it != 0; it++)
			--DmVec[*it - 1];

#ifdef DEBUG
		if(DebugDGISM) {
			cout << "\t *coalescence in pop " << "0" ;
			cout << ", of type " << chosenHapIndex << "\n\n" << endl;
		}
#endif

    	return Weight;
    }
    else if (validHapsForEvent[chosenHapIndex] == mutISMtype1) {
    	Weight = (changetype) 1 / QHobolth[chosenHapIndex] * Mut->ThetaDIG / popScaledTotalEventRate / PopEffectif[0];

		//	accounting for a Type 1 mutation
		siteMutCountVec[*validHapMutSiteIndexPtr[chosenHapIndex] - 1] = 0;
		DmVec[*validHapMutSiteIndexPtr[chosenHapIndex] - 1] = 0;
		mutPathMatrix[chosenHapIndex].erase(validHapMutSiteIndexPtr[chosenHapIndex]);

#ifdef DEBUG
		if(DebugDGISM) {
			cout << "\t *mutISMtype1 (mutation into inexistent type) back in pop ";
			cout << "0" << ": old type " << chosenHapIndex << "-> new type " << chosenHapIndex << "\n\n" << endl;
		}
#endif

		return Weight;
    }
    else if (validHapsForEvent[chosenHapIndex] == mutISMtype2) {
		//	accounting for a Type 2 mutation
		siteMutCountVec[*validHapMutSiteIndexPtr[chosenHapIndex] - 1] = 0;
		DmVec[*validHapMutSiteIndexPtr[chosenHapIndex] - 1] = 0;
		mutPathMatrix[chosenHapIndex].erase(validHapMutSiteIndexPtr[chosenHapIndex]);

		//	searching for the existing haplotype into which the chosen type has mutated
		size_t pathCounter;

		//	IF the existing type is mutating into the MRCA then look for an already existing MRCA
		if (mutPathMatrix[chosenHapIndex][0] == 0) {
			for (pathCounter = 0; pathCounter < mutPathMatrix.size(); pathCounter++)
				if ((pathCounter != chosenHapIndex) && (mutPathMatrix[pathCounter] == mutPathMatrix[chosenHapIndex]) && (Effectif[0][pathCounter] > 0))
					break;
		}
		//	ELSE look for a similar non-MRCA type
		else {
			for (pathCounter = 0; pathCounter < mutPathMatrix.size(); pathCounter++)
				if ((pathCounter != chosenHapIndex) && (mutPathMatrix[pathCounter] == mutPathMatrix[chosenHapIndex]))
					break;
		}

		//	CBR: in this case a migPlusISMtype2 mutation gives rise to a previously
		//	inexistent type (the MRCA) in the sample (just like for migPlusISMtype1)
		if (pathCounter == mutPathMatrix.size()) {
	    	Weight = (changetype) 1 / QHobolth[chosenHapIndex] * Mut->ThetaDIG / popScaledTotalEventRate / PopEffectif[0];
#ifdef DEBUG
			if (DebugDGISM) {
				cout << "\t *mutISMtype2 (BUT mutation into inexistent type!!) back in pop ";
				cout << "0" << ": old type " << chosenHapIndex << "-> new type " << chosenHapIndex << "\n\n" << endl;
			}
#endif
		}
		//	CBR: this case corresponds to the classic ISM Type 2 mutation (i.e. from type x -> type y)
		else {
			//	eliminates all traces of the chosen type's site segregation counts
			//	N.B. here the TOTAL SAMPLE site segregation counts (DmVec) does not change as they are also contained within the "target" haplotype y
			for (vector<int>::iterator it = mutPathMatrix[chosenHapIndex].begin(); *it != 0; it++)
				--siteMutCountVec[*it - 1];

			//	assign zero to indicate that this type is non-existent from now onwards
			mutPathMatrix[chosenHapIndex].assign(1,0);

	    	Weight = (changetype) 1 / QHobolth[chosenHapIndex] * Mut->ThetaDIG * (Effectif[0][pathCounter] + 1.0) / popScaledTotalEventRate / PopEffectif[0];

			//	account for the ISM Type 2 mutation (i.e. from type x -> type y) in a single population
	    	--Effectif[0][chosenHapIndex];
	    	--StateEffectif[chosenHapIndex];

	    	++Effectif[0][pathCounter];
	    	++StateEffectif[pathCounter];

#ifdef DEBUG
			if (DebugDGISM) {
				cout << "\t *mutISMtype2 Mutation back in pop ";
				cout << "0" << ": type " << chosenHapIndex << "->" << pathCounter << "\n\n" << endl;
			}
#endif
		}
		return Weight;
    }
	return (-1.0);
}
/*-----------------------------------------------------------------------------*/

//***********************************************************
// CBR (10.05.2013) calculates p_theta(d) for each segregating site given in Hobolth et al. (2008) (Eqs. 11 & 14)
changetype CSample::calcHobolth_pThetaOfDm(const parstype &theta, const int &dm) {

	changetype numerator = 0.0, denominator = 0.0;

	//	p_theta(d) for dm = 1 (Hobolth et al. (2008) Eq. 14 REARRANGED the summation limits!!)
	if (dm == 1) {
		numerator = (changetype) (Size - 1) / (Size - 1 + theta);
		for (int k = 1; k <= Size - 1; k++)
			denominator += (changetype) k / (k + theta);
	}

	//	p_theta(d) for dm > 1 (Hobolth et al. (2008) APPENDIX II, Eq. A.4 AND reverse the summation limits)
	else if (dm > 1) {
		changetype factor = 1.0;
		//	for calculating (dm - 2)!
		for (int ii = 2; ii <= dm - 2; ii++)
			factor *= ii;

		//	special case for k = Size - dm + 1
		numerator += (changetype) (Size - dm) / (Size - dm + theta) * factor;
		denominator += (changetype) (dm - 1) * numerator;

		//	REVERSE SUMMATION!!
		for (int k = Size - dm; k >= 2; k--) {
			factor *= (changetype) (Size - k - 1) / (Size - dm + 1 - k);

			changetype mult = (changetype) (k - 1) / (k - 1 + theta) * factor;

			numerator += mult;
			denominator += (changetype) (Size - k) * mult;
		}
		numerator *= (changetype) (dm - 1);
	}

	if (denominator == 0.0) {
		cerr << "(!) Problem in CSample::calcHobolth_pThetaOfDm : the denominator is zero!" << endl
			 << "Aborting Migraine...Press any key to exit!" << endl;
		if (cinGetOnError)
			cin.get();
		exit(-1);
	}

	return (changetype) (numerator / denominator);
}
/*-----------------------------------------------------------------------------*/

/******************* PACpi *****************/
changetype CSample::PACpi(CMigMod* Mig,CMutMod* Mut){
// draws a gene "randomly" and computes the pi for adding it
//cout<<"debut PACpi";getchar();

int a,j,random,intcumul; //pas locales aux boucles
pascumultype StationAll=1.0/Sam_AlleleNbr;
changetype Weight;

//si il reste un seul genes alors le poids est la proba de l'allele dans la distrib stationnaire
   if(Size==1){
	  for(a=0;a<PopSampledOrNot;a++)
		 if(PopEffectif[a]>0)
			for(vector<int>::iterator i=Effectif[a].begin();i<Effectif[a].end();i++)
				if((*i)==1){
					Weight=StationAll;
					//cout << "\n Final Step, probability of state " << i << "is " << Weight << endl;
					Size=0;
					(*i)=0;
					PopEffectif[a]=0;
//if (Weight>1) {cout<<"stationR! "<<Weight; getchar();}
					return Weight;
				}
	}
//si il reste plus d'un gene
//choose a pop and a type
   random= (int) ( alea.randInt(Size-1) + 1);  // de 1 [`a] Size et non pop ou pop-1...   //CM->RL 23.01.14
   intcumul=0;
   a=-1;
// the boundary case is it stops when intcumul=random=Size
   while(intcumul<random) {a++; intcumul+=PopEffectif[a];}
//cout<<Size<<" "<<random<<" "<<a<<" "<<PopEffectif[a];getchar();
   intcumul-=PopEffectif[a];
   j=-1;
//if (a==-1) {cout<<"a=-1";getchar();}
   while(intcumul<random) {j++; intcumul+=Effectif[a][j];}
//on enleve ce gene de l'echantillon
   PopEffectif[a]--;
   Effectif[a][j]--;
   StateEffectif[j]--;
   Size--;
//on calcul le Pi pour ce gene
//cout<<"("<<j<<" "<<a<<")";

   (this->*PicalcfnPtr)(Mig,Mut,j);

	if(piOnePopEigenbool and PIs[j][a]<epsilonPiGSM){
#ifdef DEBUG
        if (true) cout << "PiOnePop SD used instead of Eigen because PIs[j][a]=" << PIs[j][a] << " < epsilonPiGSM" << endl;
#endif
        

		CalcMatSD(Mut,Mig,Mut->ThetaDIG,false);//pour PiOnePopSD in current population
		picalctype mab,pib=0.0;
		for(int k=0;k<Sam_AlleleNbr;k++) {
			mab=MatriceSD[Size][k][j];
			if(Effectif[0][k]>0.0) pib+=((Effectif[0][k]) * mab )/ (Size);
		}
		PIs[j][a]=pib;

	}


/*
   this->lin_abs_sparsified_1D_penta(Mig,Mut,j);
  if (PIs[j][a]<0) {
    cout<<"argh ";
    cout<<j<<" "<<a<<" "<<PIs[j][a]<<" ";
    // then just calls a more reliable algo which may have debug info added...
    (this->PicalcPIMLDLsym)(Mig,Mut,j);
    cout<<endl<<PIs[j][a]<<endl;
    getchar();
  }
*/
if(std::isnan(PIs[j][a])) {
	cout << "problem in PAC -> Pi=" << PIs[j][a] << endl;
}
  return PIs[j][a];

}
/******************* fin PACpi *****************/

int CSample::PicalcPIMLDLsym(const CMigMod *Mig,const CMutMod *Mut, const int& Type) {
/* solves Mmat.hatPiloc=bvec, using LDL' factorization of symmetric positive definite matrices:
Golub p.138
un peu plus rapide que Cholesky
use symmetry, doesn't requires sqrt()
L_{i,i}=1 mais LL' <> I!! */
const cumultype ThetaDIG=Mut->ThetaDIG;

const int PSON=PopSampledOrNot; // local copy
// a est un iterateur local, par question d'utiliser effaj
    for(int a=0;a<PSON;a++) {
        bvec[a]=Effectif[a][Type]/Mig->PopSize[a]+ThetaDIG/Sam_AlleleNbr;
        Mmat[a][a]=PopEffectif[a]/Mig->PopSize[a]+ThetaDIG+Mig->Ma[a];
    	for(int b=0;b<PSON;b++) if(b!=a) Mmat[a][b]=-(Mig->Mab[a][b]);
    }
/*
// attention cert des lignes ci dessous n'ont de sens que si _sparsified_ a ['e]t['e] appel['e]
    cout<<endl<<"D_A: ";
    for(reg int a=0;a<PopSampledOrNot;a++) cout<<D_A[a]<<" ";
    cout<<endl<<"ligne[a]: ";
    for(reg int a=0;a<PopSampledOrNot;a++) cout<<ligne[a]<<" ";
// ie golub vars (ligne...) ont ['e]t['e] dimensionn['e]es etc...    cout<<endl<<"bvec[a]: ";;
    cout<<ThetaDIG<<" "<<Sam_AlleleNbr;
    cout<<endl;
    for(reg int a=0;a<PopSampledOrNot;a++) cout<<bvec[a]<<" ";
    cout<<endl;
    for(int a=0;a<PopSampledOrNot;a++){
    cout << "\t" ;
	for(int b=0;b<PopSampledOrNot;b++){
		cout << Mmat[a][b] << ',' ;
		}
	cout <<  endl;
	}
getchar();
*/

    for (int j = 0; j < PSON; j++) {
    // at first iteration no elment of L_ is known but the sum is over i<0...
         for (int i = 0; i < j; i++) v_[i]=L_[j][i]*D_[i];
    //if (j>=PSON) cerr<<j<<endl<<v_[j]<<endl<<Mmat[j][j]<<endl;
         v_[j]=Mmat[j][j];
         for (int i = 0; i < j; i++) v_[j]-=L_[j][i]*v_[i];
         D_[j]=v_[j];
         for (int k=j+1;k<PSON;k++) {
          	L_[k][j]=Mmat[k][j];
            for (int i = 0; i < j; i++) L_[k][j]-=L_[k][i]*v_[i];
//            L_ptr=L_[k];
//            for (reg int i = 0; i < j; i++) L_ptr[j]-=L_ptr[i]*v_[i];
            L_[k][j]/=v_[j];
         }
    }
//computes (LDL')^{-1}.b= L^{-1}D^{-1}L'^{-1}b using forward and backward substitution
      // b->L'^{-1}b:
      for (int k = 0; k < PSON; k++)
	  {
         for (int i = 0; i < k; i++)
               bvec[k] -= bvec[i]*L_[k][i];
      }
      // ->D^{-1}.(L'^{-1}b)
      for (int k = 0; k < PSON; k++) bvec[k] /= D_[k];
      // -> L^{-1}.(D^{-1}L'^{-1}b)
      for (int k = PSON-1; k >= 0; k--)
	  {
         for (int i = k+1; i < PSON; i++)
               bvec[k] -= bvec[i]*L_[i][k];
      }
/*    cout<<endl<<"pi[popit]: ";;
    for(reg int a=0;a<PopSampledOrNot;a++) cout<<bvec[a]<<" ";
getchar();
*/

      for (int popit=0;popit<PSON;popit++) {
          PIs[Type][popit]=bvec[popit];
//cout<<" "<<bvec[popit];
      }
//cout<<"\n";getchar();

return 0;
} // end PicalcPIM

//#endif
//#ifdef lin_abs_spTRUE
int CSampleWout::sparsified(const CMigMod *Mig,const CMutMod *Mut, const int& Type) {
cout<<"Arrive dans CSample***Wout***::sparsified";if (cinGetOnError) cin.get();
return 0;
}


int CSampleWout::lin_abs_sparsified_1D_penta(const CMigMod *Mig,const CMutMod *Mut, const int& Type) {
cout<<"Arrive dans CSample***Wout***::lin_abs_sparsified_1D_penta";if (cinGetOnError) cin.get();
return 0;
}

/** FR->FR what could be done : a approximate solution in single precision, then iterative refinement in double precision
see for example http://www.maths.bath.ac.uk/~masrs/ma50177/pdfs/bath-25apr07.pdf, slide 13
Ax=b, Ay=r  are in single precision. However, the gains are not huge according to the slide, and picalctype float shows a moderate effect.
**/

int CSampleWithGolub::sparsified(const CMigMod *Mig,const CMutMod *Mut, const int& Type) {
if (C_c_bool) cerr<<endl<<"entering CSampleWithGolub::sparsified";
//cout<<"Arrive dans CSampleWithGolub::lin_abs_sparsified";getchar();
/* conjugate gradient method for solving Ax=b starting from a solution of Mx=b
Valid in 2D : does not assume constant subdiagonals as in 1D

M here taken as island model approximation of A

(1) solves Mx=b using sherman morrison formula
(2) Then solves M^{-1}.sparse(A)x=M^{-1}b by conjugate gradient iteration on x
(Golub & van Loan 1996, p.534 for conjugate gradient, PressX's sparse()  for the sparse matrix version)
*/

const cumultype ThetaDIG=Mut->ThetaDIG;
const int PSON=PopSampledOrNot; // local copy
static picalctype /*Md,Mdd,*/temp,normr,pAp,zr,newzr,Golubalpha,Golubbeta,
       islandCst,plusIdi,plusIdib,fac;

////// INITIALISATION
    for(int a=0;a<PSON;a++) {
        bvec[a]=Effectif[a][Type]/Mig->PopSize[a]+ThetaDIG/Mut->Mut_AlleleNbr;
        D_A[a]=(PopEffectif[a]/Mig->PopSize[a]+ThetaDIG+Mig->Ma[a]);
        Mmat[a][a]=PopEffectif[a]/Mig->PopSize[a]+ThetaDIG+Mig->Ma[a];
        for(int b=0;b<PSON;b++) if(b!=a) Mmat[a][b]=-(Mig->Mab[a][b]);
    }
    int dist=PSON-1;
    for(int a=3;a<PSON;a++){
        if (Mmat[0][a]>-SPARSIFY_THRESHOLD) {
            dist=a;
	    //	  a=PSON; //quits loop: this assumes that row elements are decreasing!
        }
    }
    for (reg int a=0;a<PSON;a++) {
        bas[a]=max(0,a-dist);
        haut[a]=min(PSON,a+dist+1);
    }

    islandCst= - Mig->DispPars[0]/(PSON-1);
    plusIdi=0.;
    plusIdib=0.;
    /*compact solution of island equation:
    this takes Mmat as a diagonal matrix plus a constant matrix, note that the constant is the outer product of two obvious vectors,
    then applies the Sherman-Morrison formula and a few additional simplifications
     islandCst is the constant term
     plusIdib is the sum of inverse of the diagonal terms
    cf checks in  test_calcul_Pi.nb */
    for(int a=0;a<PSON;a++) {
        D_M[a]=1/(D_A[a]-islandCst);
        plusIdi+=D_M[a];
        plusIdib+=D_M[a]*bvec[a];
    }
    fac=islandCst*plusIdib/(1.+islandCst*plusIdi);
    for(reg int a=0;a<PSON;a++) {
        Golubx[a]=(bvec[a]-fac)*D_M[a];
    } // Golubx solves Misland.x=b

	normr=0;
// calcul Golub:r_0=b-Ax_0;
	for (int a=0;a<PSON;a++) {
		temp=bvec[a];
		for (int k=bas[a];k<a;k++) temp-=Mmat[a][k]*Golubx[k];
		temp-=D_A[a]*Golubx[a];
		for (int k=a+1;k<haut[a];k++) temp-=Mmat[a][k]*Golubx[k];
		Golubr[a]=temp;
		normr+=fabs(temp);
	}

    int ii=0;
  	while(normr>GRADIENT_CONV_THRESHOLD) {
	//Solve Mz_k=r_k using sherman morrison again
        plusIdib=0.;
        for(int a=0;a<PSON;a++) {
            plusIdib+=D_M[a]*Golubr[a];
        }
        fac=islandCst*plusIdib/(1.+islandCst*plusIdi);
        for(int a=0;a<PSON;a++) {
            Golubz[a]=(Golubr[a]-fac)*D_M[a];
        }
    	ii++;
    	if(ii==1) {
    		zr=0.;
    		for (int a=0;a<PSON;a++) {
    			Golubp[a]=Golubz[a];
    			zr+=Golubz[a]*Golubr[a];
    		}
    	} else {
    		newzr=0.;
    		for (reg int a=0;a<PSON;a++) newzr+=Golubz[a]*Golubr[a];
            Golubbeta=newzr/zr;
    		zr=newzr;
    		for (reg int a=0;a<PSON;a++) Golubp[a]=Golubz[a]+Golubbeta*Golubp[a];
    	}
//computes Golub's Ap_k and p_k A p_k
        pAp=0.;
		for (int a=0;a<PSON;a++) {
			temp=0.;
			for (reg int k=bas[a];k<a;k++) temp+=Mmat[a][k]*Golubp[k];
			temp+=D_A[a]*Golubp[a];
			for (reg int k=a+1;k<haut[a];k++) temp+=Mmat[a][k]*Golubp[k];
			GolubAp[a]=temp;
			pAp+=Golubp[a]*temp;
		}
    	if (pAp==0.0) { //FER normal mais rare (cf PressX)
           for (int j=0;j<PSON;j++) PIs[Type][j]=Golubx[j];
//         cout<<"fin sparsified\n";getchar();
           return 0;
        } else
    	Golubalpha=zr/pAp;
    	normr=0.;
   		for (int a=0;a<PSON;a++) {
   			Golubx[a]+=Golubalpha*Golubp[a];
   			Golubr[a]-=Golubalpha*GolubAp[a];
			normr+=fabs(Golubr[a]);
   			}
	}  //fin while

	for (int j=0;j<PSON;j++) {
        PIs[Type][j]=Golubx[j];
    }
    if (C_c_bool) {cerr<<endl<<"leaving CSampleWithGolub::sparsified";}
	return 0;
} // end sparsified

int CSampleWithGolub::lin_abs_sparsified_1D_penta(const CMigMod *Mig,const CMutMod *Mut, const int& Type) {
if (C_c_bool) cerr<<endl<<"normr values in CSampleWithGolub::lin_abs_sparsified_1D_penta: ";
//cout<<"Arrive dans CSampleWithGolub::lin_abs_sparsified_1D_penta";getchar();
/* conjugate gradient method for solving Ax=b starting from a solution of Mx=b
M here taken as pentadiagonal, constant diagonals approximation of A
(1) solves Mx=b by LDL' factorization of pentadiagonal symmetric positive definite matrices
(2) Then solves M^{-1}.sparse(A)x=M^{-1}b by conjugate gradient iteration on x
(Golub & van Loan 1996, p.534)
*** Uses constant subdiagonals (not simply for initial matrix) which a 1D feature... not valid in 2D !!!
if the matrix is incorrectly forced into one with constant subdiagonal by applying this algo,
the row sum may not be positive, the matrix not positive definite... the pi values negative...
*/

///FR->FR implementer initialisation par island model pour g large // PSON ?
// static pointers are faster !!! new's without delete's !!!
const cumultype ThetaDIG=Mut->ThetaDIG;
const int PSON=PopSampledOrNot; // local copy
static picalctype Md,Mdd,temp,normr,pAp,zr,newzr,Golubalpha,Golubbeta;


////// INITIALISATION
for(reg int a=0;a<PSON;a++) {
    bvec[a]=Effectif[a][Type]/Mig->PopSize[a]+ThetaDIG/Mut->Mut_AlleleNbr;
	D_A[a]=(PopEffectif[a]/Mig->PopSize[a]+ThetaDIG+Mig->Ma[a]);
    D_M[a]=D_A[a];
	ligne[a]=-(Mig->Mab[0][a]);
}

int dist=PSON-1;
for(reg int a=3;a<PSON;a++){
	if (ligne[a]>-SPARSIFY_THRESHOLD) {
	    dist=a;
	    /*then quits loop: this assumes that row elements are decreasing !*/
	    a=PSON;
    }
}
//cout<<dist;getchar();
for (reg int a=0;a<PSON;a++) {
	bas[a]=max(0,a-dist);
	haut[a]=min(PSON,a+dist+1);
}
Mdd=ligne[2];
Md=ligne[1];

int ii=0;
//    Factor Golub:M=LDL' (reused later)  //D is modified
    gamma[0]=Md/D_M[0];
    delta[0]=Mdd/D_M[0];

    D_M[1]-=Md*gamma[0];
    gamma[1]=(Md-Mdd*gamma[0])/D_M[1];
    delta[1]=Mdd/D_M[1];

    for (reg int k=2;k<PSON-2;k++){
        D_M[k]+=-Mdd*delta[k-2]-D_M[k-1]*gamma[k-1]*gamma[k-1];
        gamma[k]=(Md-Mdd*gamma[k-1])/D_M[k];
        delta[k]=Mdd/D_M[k];
    }

    D_M[PSON-2]+=-Mdd*delta[PSON-4]-D_M[PSON-3]*gamma[PSON-3]*gamma[PSON-3];
    gamma[PSON-2]=(Md-Mdd*gamma[PSON-3])/D_M[PSON-2];
    D_M[PSON-1]+=-Mdd*delta[PSON-3]-D_M[PSON-2]*gamma[PSON-2]*gamma[PSON-2];


////// calcul Golub:x_0= /*Solution of pentadiagonal system*/

//      computes (LDL')^{-1}.b= L^{-1}D^{-1}L'^{-1}b

      // b->L'^{-1}b: le fait qu'on puisse le calculer comme ca n'est pas ['e]vident
      // Il faut utiliser la structure de L'

    Golubx[0]=bvec[0];
    Golubx[1]=bvec[1]-gamma[0]*Golubx[0]; //Golubx a ['e]t['e] initialis['e]...
    for (reg int k=2;k<PSON;k++) Golubx[k]=bvec[k]-gamma[k-1]*Golubx[k-1]-delta[k-2]*Golubx[k-2];
      // ->D^{-1}.(L'^{-1}b)
    for (reg int k=0;k<PSON;k++) Golubx[k]/=D_M[k];
      // -> L'^{-1}.(D^{-1}L'^{-1}b)
    Golubx[PSON-2]-=gamma[PSON-2]*Golubx[PSON-1];
    for (reg int k=PSON-3;k>=0;k--) Golubx[k]-=gamma[k]*Golubx[k+1]+delta[k]*Golubx[k+2];

///// a ce point on a une solution du systeme pentadiagonal

/*for(reg int a=0;a<PSON;a++) cout<<Golubx[a]<<" ";
cout<<endl;*/


	normr=0;
// calcul Golub:r_0=b-Ax_0;
	for (reg int a=0;a<PSON;a++) {
		temp=bvec[a];
		for (reg int k=bas[a];k<a;k++) temp-=ligne[a-k]*Golubx[k];
		temp-=D_A[a]*Golubx[a];
		for (reg int k=a+1;k<haut[a];k++) temp-=ligne[k-a]*Golubx[k];
		Golubr[a]=temp;
		normr+=fabs(temp);
	}
/*for(reg int a=0;a<PSON;a++) cout<<Golubr[a]<<" ";
cout<<endl;*/

  	while(normr>GRADIENT_CONV_THRESHOLD) {
       if (C_c_bool) cerr<<normr<<" ";
	//Solve Mz_k=r_k
//      computes Golubz=(LDL')^{-1}.Golubr= L^{-1}D^{-1}L'^{-1}Golubr

      // b->L'^{-1}Golubr: le fait qu'on puisse le calculer comme ca n'est pas ['e]vident
      // Il faut utiliser la structure de L'
		Golubz[0]=Golubr[0];
    	Golubz[1]=Golubr[1]-gamma[0]*Golubz[0];
    	for (reg int k=2;k<PSON;k++) Golubz[k]=Golubr[k]-gamma[k-1]*Golubz[k-1]-delta[k-2]*Golubz[k-2];
      // ->D^{-1}.(L'^{-1}Golubr)
    	for (reg int k=0;k<PSON;k++) Golubz[k]/=D_M[k];
      // -> L'^{-1}.(D^{-1}L'^{-1}Golubr)
    	Golubz[PSON-2]-=gamma[PSON-2]*Golubz[PSON-1];
    	for (reg int k=PSON-3;k>=0;k--) Golubz[k]-=gamma[k]*Golubz[k+1]+delta[k]*Golubz[k+2];


/*cout<<"Golubz ";
for(reg int a=0;a<PSON;a++) cout<<Golubz[a]<<" ";
cout<<endl;*/

    	ii++;

    	if(ii==1) {
    		zr=0.;
    		for (reg int a=0;a<PSON;a++) {
    			Golubp[a]=Golubz[a];
    			zr+=Golubz[a]*Golubr[a];
    		}
    	} else {
    		newzr=0.;
    		for (reg int a=0;a<PSON;a++) newzr+=Golubz[a]*Golubr[a];
            Golubbeta=newzr/zr;
    		zr=newzr;
    		for (reg int a=0;a<PSON;a++) Golubp[a]=Golubz[a]+Golubbeta*Golubp[a];
    	}
//computes Golub's Ap_k and p_k A p_k
//cout<<zr<<endl;
        pAp=0.;
		for (reg int a=0;a<PSON;a++) {
			temp=0.;
			for (reg int k=bas[a];k<a;k++) temp+=ligne[a-k]*Golubp[k];
			temp+=D_A[a]*Golubp[a];
			for (reg int k=a+1;k<haut[a];k++) temp+=ligne[k-a]*Golubp[k];
			GolubAp[a]=temp;
			pAp+=Golubp[a]*temp;
		}
    	if (pAp==0.0) { //FER normal mais rare (cf PressX)
           for (reg int j=0;j<PSON;j++) PIs[Type][j]=Golubx[j];
//         cout<<"fin lin_abs_sparsified_1D_penta\n";getchar();
           return 0;
        } else
    	Golubalpha=zr/pAp;
/*cout<<"GolubAp ";
for(reg int a=0;a<PSON;a++) cout<<GolubAp[a]<<" ";
cout<<endl;

cout<<zr<<" "<<pAp<<" "<<Golubalpha<<endl;
getchar();*/

    	normr=0.;
   		for (reg int a=0;a<PSON;a++) {
   			Golubx[a]+=Golubalpha*Golubp[a];
   			Golubr[a]-=Golubalpha*GolubAp[a];
			normr+=fabs(Golubr[a]);
   			}
	}  //fin while

	for (reg int j=0;j<PSON;j++) {
        PIs[Type][j]=Golubx[j];
    }
    if (C_c_bool) {cerr<<endl<<"leaving CSampleWithGolub::lin_abs_sparsified_1D_penta";}
		return 0;
} // end lin_abs_sparsified_1D_penta

/*-----------Quelques Procedures Picalc de DG2002 r['e]-adapt['e]e pour migraine---------------*/


//***********************************************************
// CBR (30.04.2012) procedure for calculating Pi_hats under ISM (NB. only for coalescence & migration events)
int CSample::PiCalcISM(const CMigMod *Mig,const CMutMod *Mut, const int &Type){
	reg int a, b;
	const cumultype ThetaDIG = Mut->ThetaDIG;
	const int PSON = PopSampledOrNot; // local copy
	pirestype Denominator;

	for (a = 0; a < PSON; a++) {
		if (a != chosenPop)
			setNZeroForPiCalcISM(Mig, a, Type);
		Denominator = (PopEffectif[a] + 1) * (PopEffectif[a]/Mig->PopSize[a] + Mig->Ma[a] + ThetaDIG);
		bvec[a] = Effectif[a][Type] / Mig->PopSize[a];
		Mmat[a][a] = Denominator / nZero[a];
		for (b = 0; b < PSON; b++)
			if (b != a)
				Mmat[a][b] = -(Mig->Mab[a][b]);
	}

	inversion(Mmat,PSON);

#ifdef DEBUG
	reg int i, j;
	if (gaussdebug) {
		printf("\n\n\tfinal step \n");
		for (i = 0; i < PSON; i++) {
			printf("\n\t");
			for (j = 0; j < PSON; j++)
				printf("Mmat[%d][%d]=%3.4e ", i, j, (double) Mmat[i][j]);
			printf("bvec[%d]=%.4e", i, (double) bvec[i]);
		}
	}
	if (DebugPi) {
		for (a = 0; a < PSON; a++) {
			printf("\n\n\t\tPi_Pim_general(%d,%d)", Type, a);
			//printf("\n\t\t pi=b_gauss[%d]=%3.4e ;\n",a,bvec[a]);
			cout << "\n\t\t Pi=b_gauss[" << a << "]=" << bvec[a] << endl;
		}
	}
#endif

	for (a = 0; a < PSON; a++) {
		PIs[Type][a] = 0.0;
		for (b = 0; b < PSON; b++)
			PIs[Type][a] += (Mmat[a][b] * bvec[a]);
	}
	return 0;
}
/*-----------------------------------------------------------------------------*/

//***********************************************************
// CBR (03.05.2012) procedure for setting nZero for all PopSampledOrNot (CSample::PiCalcISM)
//	(N.B. this is a tailored version of what  already exists in CSample::Change)
void CSample::setNZeroForPiCalcISM(const CMigMod *Mig, const int &Pop, const int &Type){

	//	under ideal conditions nZero remains PopEffectif[Pop]
	nZero[Pop] = PopEffectif[Pop];

	for (reg int k = 0; k < Sam_AlleleNbr; k++) {
		//	"GLOBALLY singleton" check for EVERY HAPLOTYPE accounting for the gene taken out of the sample
		if (((k != Type) && (Effectif[Pop][k] == 1) && (StateEffectif[k] == 1))
				|| ((k == Type) && (Effectif[Pop][k] == 0) && (StateEffectif[k] == 0))) {

			vector<int>::iterator it1, it2;
			//	looping over the list of segregated sites for each haplotype
			for (it1 = mutPathMatrix[k].begin(); *it1 != 0; it1++)
				//	IF a unique mutation is found for the haplotype
				if (siteMutCountVec[*it1 - 1] == 1)
					//	break loop as the current haplotype corresponds to a possible event
					break;

			//	no unique mutation sites found => no possible events for this haplotype => modify nZero
			if (*it1 == 0)
				--nZero[Pop];
		}

		//	CBR: correct nZero and count singleton genes capable of ONLY immigration
		bool singletonMigration = false;
		//	"LOCALLY singleton" check for EVERY HAPLOTYPE accounting for the gene taken out of the sample
		if (((k != Type) && (Effectif[Pop][k] == 1) && (StateEffectif[k] > 1))
				|| ((k == Type) && (Effectif[Pop][k] == 0) && (StateEffectif[k] > 0))) {
			//	loop over second pop
			for (reg int b = 0; b < PopSampledOrNot; b++) {
				//	IF second pop is different from chosen + backward migration possible + chosen type exists in second pop
				if ((b != Pop) && (Mig->Mab[Pop][b] != 0) && (Effectif[b][k] > 0)) {
					//	THEN the current haplotype is capable of IMMIGRATING FROM at least one other pop
					singletonMigration = true;
					break;
				}
			}
			//	IF the current haplotype couldn't have immigrated into the chosen pop THEN modify nZero
			if ((!singletonMigration) && (k != Type))
				--nZero[Pop];
		}
	}
}
/*-----------------------------------------------------------------------------*/

//***********************************************************
//(DG2002.cpp)->marche  pour 2pop 2 alleles ... 5 pop 5 alleles
//Migraine->comparaison avec DG2002 OK pour 2 pops et PIM 5 alleles a v['e]rifier plus largement (RL 28/07/2006)
// Verified IBD with Mathematica for 2-pop and 2-alleles under the IM and stable 2-pop scenarios (CBR 17.12.2013)

int CSample::PiCalcPimDG2002(const CMigMod *Mig,const CMutMod *Mut, const int& Type){
// pop structuree en PIM
reg int a,b;


const cumultype ThetaDIG=Mut->ThetaDIG;

const int PSON=PopSampledOrNot; // local copy

for(a=0;a<PSON;a++){
	bvec[a]=Effectif[a][Type]/Mig->PopSize[a]+ThetaDIG/Sam_AlleleNbr;
	Mmat[a][a]=PopEffectif[a]/Mig->PopSize[a]+Mig->Ma[a]+ThetaDIG;
	for(b=0;b<PSON;b++)	if(b!=a) Mmat[a][b]=-(Mig->Mab[a][b]);
}

Gaussian(Mmat,bvec,PSON);
#ifdef DEBUG
reg int i,j;
if(gaussdebug){
   	printf("\n\n\tfinal step \n");
		for(i=0;i<PSON;i++) {
			printf("\n\t");
			for(j=0;j<PSON;j++) printf("Mmat[%d][%d]=%3.4e ",i,j,(double) Mmat[i][j]);
			printf("bvec[%d]=%.4e",i,(double) bvec[i]);
		}
	}
if(DebugPi) {
	for(a=0;a<PSON;a++) {
		printf("\n\n\t\tPi_Pim_general(%d,%d)",Type,a);
		//printf("\n\t\t pi=b_gauss[%d]=%3.4e ;\n",a,bvec[a]);
		cout << "\n\t\t Pi=b_gauss[" << a << "]=" << bvec[a] << endl;
	}
}
#endif
for (int popit=0;popit<PSON;popit++) {
    //cout << "Type=" << Type << " popit=" << popit << endl;
    //cout << "bvec=" << bvec[popit]  << endl;
PIs[Type][popit]=bvec[popit];
    //cout<<"PIs=" << PIs[Type][popit]<<endl;
    //cin.get();
}
//ostream_vector(hpv,cout);cin.get();
return 0;
}
/*-----------------------------------------------------------------------------*/

//***********************************************************
//Proc['e]dure PiTwo(Pim) directement import['e]e de DG2002
//DG2002->normalement pi two prb : mais visiblement marche pas si non pim
//DG2002->verifiee pour Pim avec 2 ou 3 alleles
//DG2002->pour autre que Pim ->
int CSample::PiCalcTwoPopDG2002(const CMigMod *Mig,const CMutMod *Mut,const int& Type){
reg int a,b;
reg picalctype det;
reg vector<picalctype> bb;
//reg vector< vector<picalctype> > A;

const cumultype ThetaDIG=Mut->ThetaDIG;

const int PSON=PopSampledOrNot; // local copy

if(PSON!=2) {
	cout<<"(!) Problem in Pi calculation : Migraine try to use the procedure PiCalcTwoPopDG2002() for a model with "<<PSON<<" pop(s);"<<endl;
	cout<<"(!) I exit."<<endl;
	if (cinGetOnError) cin.get();
	exit(-1);
}

for(a=0;a<PSON;a++) {
    b=1-a;
    Mmat[0][0]=PopEffectif[a]/Mig->PopSize[a]+Mig->Ma[a]+ThetaDIG;
    Mmat[1][1]=PopEffectif[b]/Mig->PopSize[b]+Mig->Ma[b]+ThetaDIG;
    Mmat[0][1]=-(Mig->Ma[a]);
    Mmat[1][0]=-(Mig->Ma[b]);
    det=Mmat[0][0]*Mmat[1][1]-Mmat[0][1]*Mmat[1][0];
#ifdef DEBUG
    if(DebugPi) {
        printf("\n\n\t\tPi_TwoPop(%d,%d)",Type,a);
        printf("\n\t\tdeterminant=%g",(double) det);
    }
#endif
    if(fabs(det)<numeric_limits<picalctype>::epsilon()){
        cout << "Bug in PiTwoPop(); det is" ;
	//cout << det << endl;
        printf("%g \n",(double) det);
        if (cinGetOnError) cin.get();
        exit(-1);
    }

    bvec[a]=Mmat[1][1]*(Effectif[a][Type]/Mig->PopSize[a]+ThetaDIG*Mut->Pij[0][Type]);
    bvec[a] -=Mmat[0][1]*(Effectif[b][Type]/Mig->PopSize[b]+ThetaDIG*Mut->Pij[0][Type]);
    bvec[a]/=det;
}
for (int popit=0;popit<PSON;popit++) {
	PIs[Type][popit]=bvec[popit];
	//cout<<PIs[Type][popit];cin.get();
}
//ostream_vector(hpv,cout);cin.get();
return 0;
}
/*------------------------------------------------------------------------------*/

//***********************************************************
//Porcedure PiTwoSMM2 directement imoprt['e]e de DG2002
//DG2002->nouvelle procedure 03032003 les pi somment a 1
//DG2002->verification 042003 ok pour deux genes
//DG2002->donne la bonne prob avec un run pour 2 genes
int CSample::PiCalcTwoPopSmmDG2002(const CMigMod *Mig,const CMutMod *Mut,const int &Type){
vector <picalctype> LLambda,x,r;
picalctype Ak0,Ak1,PiLocal,temp,temp1;
int k,a,b,stepNbr;


const parstype ThetaDIG=Mut->ThetaDIG;

if(PopSampledOrNot!=2) {
	cout<<"(!) Problem in Pi calculation : Migraine try to use the procedure PiCalcTwoPopSMMDG2002() for a model with "<<PopSampledOrNot<<" pops;"<<endl;
	cout<<"(!) I exit."<<endl;
	if (cinGetOnError) cin.get();
	exit(-1);
}

if(mutationModel!="SMM") {
	cout<<"(!) Problem in Pi calculation : Migraine try to use the procedure PiCalcTwoPopSMMDG2002() for a  "<<mutationModel<<" mutational model;"<<endl;
	cout<<"(!) I exit."<<endl;
	if (cinGetOnError) cin.get();
	exit(-1);
}

x.resize(2);
LLambda.resize(2);
r.resize(2);

x[0]=PopEffectif[0]/Mig->PopSize[0]+Mig->Ma[0];
x[1]=PopEffectif[1]/Mig->PopSize[1]+Mig->Ma[1];
//	CBR,RL (02.12.2013) avoids precision problems when Mig->Ma << 1
if (x[0] == x[1]) {
	LLambda[0]=-x[0] + sqrtl(Mig->Ma[0]*Mig->Ma[1]);
	LLambda[1]=-x[0] - sqrtl(Mig->Ma[0]*Mig->Ma[1]);
}
else {
	LLambda[0]=(-(x[0]+x[1]) + sqrtl(sq((x[0]+x[1]))-4.0*(x[0]*x[1]-Mig->Ma[0]*Mig->Ma[1]) ) )/2.0;
	LLambda[1]=(-(x[0]+x[1]) - sqrtl(sq((x[0]+x[1]))-4.0*(x[0]*x[1]-Mig->Ma[0]*Mig->Ma[1]) ) )/2.0;
}
r[0]=ThetaDIG/(ThetaDIG-LLambda[0]);
r[1]=ThetaDIG/(ThetaDIG-LLambda[1]);

for(a=0;a<PopSampledOrNot;a++) {
	if(a==0) b=1;
	 else b=0;

	PiLocal=0.0;
	for(k=AllelicRange[0];k<=AllelicRange[1];k++){
		//RL 28062011 : passage en circulaire par stepNbr pour K>100,
		if(Sam_AlleleNbr > 100 && (abs(Type-k) > Sam_AlleleNbr/2)) stepNbr=Sam_AlleleNbr-abs(Type-k);//car SMM circulaire pour k>100
		 else stepNbr=abs(Type-k);
		Ak0=Ak1=0.0;
		if(Effectif[a][k]>0) {
			temp=Effectif[a][k]/Mig->PopSize[a];
			temp1=PopEffectif[b]/Mig->PopSize[b]+Mig->Ma[b]+ThetaDIG;
			Ak0=temp*(temp1*Ccc(stepNbr,r[0])-(ThetaDIG/2)*(Ccc(stepNbr+1,r[0])+Ccc(stepNbr-1,r[0])));
			Ak1=temp*(temp1*Ccc(stepNbr,r[1])-(ThetaDIG/2)*(Ccc(stepNbr+1,r[1])+Ccc(stepNbr-1,r[1])));
		}
		if(Effectif[b][k]>0) {
			temp=Effectif[b][k]/Mig->PopSize[b]*Mig->Ma[a];
			Ak0+=temp*Ccc(stepNbr,r[0]);
			Ak1+=temp*Ccc(stepNbr,r[1]);

		}
		PiLocal+=(Ak0/(ThetaDIG-LLambda[0])-Ak1/(ThetaDIG-LLambda[1]));
	}
	PiLocal/=(LLambda[0]-LLambda[1]);
	bvec[a]=PiLocal;
}//fin boucle sur pops (a)

for (int popit=0;popit<PopSampledOrNot;popit++) PIs[Type][popit]=bvec[popit];

#ifdef DEBUG
	if(DebugPi) {
		for(a=0;a<PopSampledOrNot;a++) {
			printf("\n\n\t\tPiCalcTwoPopSmmDG2002(%d,%d)",Type,a);
			//printf("\n\t\t pi=b_gauss[%d]=%3.4e ;\n",a,bvec[a]);
			cout << "\n\t\t Pi[" << a << "]=" << bvec[a] << endl;
		}
	}
#endif
return 0;
}



inline picalctype Ccc(int pas,picalctype r){
picalctype temp,roots;

roots=(sqrtl(1-sq(r)));
temp=1/roots*powl(r/(1+roots),abs(pas));

return temp;
}
//***********************************************************
//Procedure de calcul des Pi pouyr une pop :
// algo DG2004B PIM, SD2000 JRSL (option speciale, tout modèles) ou DGLR2005 pour SMM et extension FR-RL-PP 2011 pour GSM

//26/11/2010 : modified to consider AlternativeIS_AncPopSize and FRExpMixAncActPopSize

int CSample::PiCalcOnePop(const CMigMod *Mig,const CMutMod *Mut,const int& ChosenType){
int k,StepNbr;
picalctype pib,mab;
//picalctype altpib,pib1,expFactor;

const picalctype ThetaDIG=Mut->ThetaDIG;
const int SampleSize=Size;
static parstype prevPopSize=0.0;
static parstype prevTheta=0;
static int prevSampleSize=0;

// on gagne 20% de temps de calculs en mettant tout en float, mais quelques imprécisions
//passage en static ne fait gagner que 10% de temps de calculs

//for SMM and GSM
static parstype RRho;
static picalctype SqrtsqRRho;
static picalctype InsideBrackets;
static picalctype denom;

//for GSM only
static parstype g;
static parstype term;
static picalctype RRR;
static picalctype SqrtRRR;
static picalctype RRRinvSqrtRRR;
static picalctype invDenom;

//for PiOnePopEigen et PIM only
static picalctype EventRate=SampleSize/Mig->PopSize[0]+ThetaDIG;

//for PiOnePopEigen et PIM only
static picalctype Term1=EventRate*Mig->PopSize[0];
static picalctype Term2=ThetaDIG/EventRate;

//for PIM only
static picalctype 	Term3=ThetaDIG/(EventRate*Sam_AlleleNbr);
static picalctype 	Term4=1.0/(Mig->PopSize[0]*EventRate);


//const picalctype AncThetaDIG=Mut->AncThetaDIG;
//const picalctype PopSizeRatio=AncThetaDIG/ThetaDIG;
//const picalctype AltRho=ThetaDIG/(SampleSize/PopSizeRatio+ThetaDIG);//pour alternativeIS_AncPopSize et


#ifdef DEBUG
	if(DebugPiSMM) {
        cout << "\n\nDebuging Pi_OnePop : " << endl;
		Print();
		if(DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush") cout << "scaled_generation=" << scaled_generation << " and MinChangeTime" << Mig->MinChangeTime << endl;
        cout << "Sam_AlleleNbr=" << Sam_AlleleNbr << endl;
		if(mutationModel=="GSM") cout << "pGSM=" << Mut->pGSM << endl;
/*		if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool)
			cout << "Mig->PopSize[0]=AncThetaDIG/ThetaDIG" << PopSizeRatio << endl;
		else*/ cout << "Mig->PopSize[0]=" << Mig->PopSize[0] << endl;
        cout << "ThetaDIG=" << ThetaDIG << endl;
        cout << "SampleSize=" << SampleSize << endl;
		cout << "ChosenTypeType="<< ChosenType << endl;
//		if((AlternativeIS_AncPopSizebool ||  AlternativeIS_FRExpMixAncActPopSizebool) && !Mig->InAncStablePopbool) {
//			cout << "AltRho="<< AltRho << endl;
//		}
	}
#endif



//if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) altpib=0.0;
if(mutationModel=="PIM" ) {
/*******************************/
	//PIM DG2004

//  if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool)
//	   pib=(AltRho/Sam_AlleleNbr)+Effectif[0][ChosenType]/PopSizeRatio/(SampleSize/PopSizeRatio+ThetaDIG);
//	else {
//		if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool)
//			altpib=(AltRho/Sam_AlleleNbr)+Effectif[0][ChosenType]/PopSizeRatio/(SampleSize/PopSizeRatio+ThetaDIG);

	if( prevSampleSize!=SampleSize
		or fabs(Mig->PopSize[0]-prevPopSize)/Mig->PopSize[0]>epsilonOPVSChange
			or fabs(ThetaDIG-prevTheta)/ThetaDIG>epsilonOPVSChange) {
		EventRate=SampleSize/Mig->PopSize[0]+ThetaDIG;
		Term3=ThetaDIG/(EventRate*Sam_AlleleNbr);
		Term4=1.0/(Mig->PopSize[0]*EventRate);
		prevSampleSize=SampleSize;
		prevTheta=ThetaDIG;
		prevPopSize=Mig->PopSize[0];
	}
	pib=Term3+Effectif[0][ChosenType]*Term4;
//	}
} else if(piOnePopSDbool) {
/*******************************/
//GSM et SMM ou autre en SD2000
#ifdef DEBUG
	if(DebugPiSMM) cout << "Precedure SD : " << endl;
#endif

	if((DemographicModel=="OnePopVarSize" || DemographicModel=="OnePopFounderFlush")
			and ( prevSampleSize!=SampleSize
				or fabs(Mig->PopSize[0]-prevPopSize)/Mig->PopSize[0]>epsilonOPVSChange
					or fabs(ThetaDIG-prevTheta)/ThetaDIG>epsilonOPVSChange )
                ) {
		   CalcMatSD(Mut,Mig,Mut->ThetaDIG,false);
		   prevPopSize=Mig->PopSize[0];
		   prevTheta=ThetaDIG;
		   prevSampleSize=SampleSize;
#ifdef DEBUG
            if(DebugPiSMM) cout << "SD Matrix recomputed" << Term1 << endl;
#endif
	   }



	pib=0.0;
	for(k=0;k<Sam_AlleleNbr;k++) {
        mab=MatriceSD[SampleSize][k][ChosenType];
		if(Effectif[0][k]>0.0) pib+=(Effectif[0][k]) * mab;
#ifdef DEBUG
		if(DebugPiSMM)
			cout << "k=" << k << " -> mab=MatriceSD[SampleSize][k][ChosenType]=" << mab << endl;
#endif
    }

} else if(piOnePopEigenbool) {


	if(prevSampleSize!=SampleSize
			or fabs(Mig->PopSize[0]-prevPopSize)/Mig->PopSize[0]>epsilonOPVSChange
				or fabs(ThetaDIG-prevTheta)/ThetaDIG>epsilonOPVSChange) {
		EventRate=SampleSize/Mig->PopSize[0]+ThetaDIG;
		Term1=EventRate*Mig->PopSize[0];
		Term2=ThetaDIG/EventRate;
		prevSampleSize=SampleSize;
		prevTheta=ThetaDIG;
		prevPopSize=Mig->PopSize[0];
	}

	pib=0.0;
	picalctype ScalProdnR;

	for(k=0;k<Sam_AlleleNbr;k++) {
		ScalProdnR=0.0;
		//for(int j=0;j<Sam_AlleleNbr;j++)
		for(int j=AllelicRange[0];j<=AllelicRange[1];j++)
			if(Effectif[0][j]>0) ScalProdnR+=Effectif[0][j]*Mut->REigenVectors[j][k];
		pib+=ScalProdnR*Mut->LEigenVectors[k][ChosenType]/(Term1*(1.0 - Term2*Mut->EigenValues[k]));
		/*cout << "k=" << k  << endl;
		cout << "REigenVectors[j][k]=";
		for(int j=0;j<Sam_AlleleNbr;j++) cout << " " << Mut->REigenVectors[j][k];
		cout << "\nScalProdnR=" << ScalProdnR << endl;
		cout << "LEigenVectors[k][ChosenType]=" << Mut->LEigenVectors[k][ChosenType] << endl;*/
    }

#ifdef DEBUG
	if(DebugPiSMM) {
        cout << "Pi One Pop Eigen (exFRRL) computation : " << endl;
        cout << "EventRate=SampleSize/Mig->PopSize[0]+ThetaDIG=" << EventRate << endl;
		cout << "Term1=EventRate*Mig->PopSize[0]=" << Term1 << endl;
        cout << "Term2=ThetaDIG/EventRate=" << Term2 << endl;
	}
#endif


} else {
/*************************************/
// SMM/GSM en DGLR2005 - FR-RL-PP 2011
if(mutationModel=="SMM") {
	pib=0.0;

	if(prevSampleSize!=SampleSize
			or fabs(Mig->PopSize[0]-prevPopSize)/Mig->PopSize[0]>epsilonOPVSChange
				or fabs(ThetaDIG-prevTheta)/ThetaDIG>epsilonOPVSChange) {
		RRho=ThetaDIG/(SampleSize/Mig->PopSize[0]+ThetaDIG);
		SqrtsqRRho=sqrt(1-sq(RRho));
		InsideBrackets=RRho/(1+SqrtsqRRho);
		denom=SqrtsqRRho*SampleSize/(1.0-RRho);
		prevSampleSize=SampleSize;
		prevTheta=ThetaDIG;
		prevPopSize=Mig->PopSize[0];
	}
	for(k=AllelicRange[0];k<=AllelicRange[1];k++){
        if(Effectif[0][k]>0.0) {
			//if(abs(ChosenType-k) > Sam_AlleleNbr/2) StepNbr=Sam_AlleleNbr-abs(ChosenType-k);//car SMM circulaire pour DG
			 //else StepNbr=abs(ChosenType-k);
            //RL devrait être plus rapide
            StepNbr=min(abs(ChosenType-k),Sam_AlleleNbr-abs(ChosenType-k));
//			if(AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool)
//				pib+=((Effectif[0][k]) * (1-AltRho)/sqrt(1-sq(AltRho)) * pow( (double) AltRho/(1+sqrt(1-sq(AltRho))),StepNbr) )/ (SampleSize);
//			else {
//				if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool)
//					altpib+=((Effectif[0][k]) * (1-AltRho)/sqrt(1-sq(AltRho)) * pow( (double) AltRho/(1+sqrt(1-sq(AltRho))),StepNbr) )/ (SampleSize);
				pib+=Effectif[0][k] * pow( (double) InsideBrackets,StepNbr) / denom;
//			}
#ifdef DEBUG
            if(DebugSD || DebugPiSMM) {
//				if((AlternativeIS_AncPopSizebool ||  AlternativeIS_FRExpMixAncActPopSizebool) && !Mig->InAncStablePopbool) cout << "k=" << k <<" -> (alt)Pib+=" <<
//					((Effectif[0][k]) * (1-AltRho)/sqrt(1-sq(AltRho)) * pow( (double) AltRho/(1+sqrt(1-sq(AltRho))),StepNbr) )/ (SampleSize) << endl;
				if(! (AlternativeIS_AncPopSizebool && !Mig->InAncStablePopbool)) cout << "k=" << k <<" -> Pib+=" <<
					Effectif[0][k] * pow( (double) InsideBrackets,StepNbr)/ denom << endl;
			}
#endif
        }// end if(Effectif[0][k]>0.0)
    }//end for(k=AllelicRange[0];k<=AllelicRange[1];k++)

} else if(mutationModel=="GSM" /*and Mut->pGSM<=0.99*/) {//end if SMM
	pib=0.0;
	//picalctype corrFiniteAlleleNbrPart1,CorrFiniteAlleleNbr;

	if( prevSampleSize!=SampleSize
			or fabs(Mig->PopSize[0]-prevPopSize)/Mig->PopSize[0]>epsilonOPVSChange
				or fabs(ThetaDIG-prevTheta)/ThetaDIG>epsilonOPVSChange ) {
		RRho=ThetaDIG/(SampleSize/Mig->PopSize[0]+ThetaDIG);
		g=Mut->pGSM;
		term=1+sq(g);
		RRR=(2.0*g+RRho*(1.0-g))/(term+RRho*g*(1.0-g));
		SqrtRRR=sqrt(1.0-sq(RRR));
		RRRinvSqrtRRR=RRR/(1.0+SqrtRRR);
		invDenom=(1-RRho)/(term+RRho*g*(1.0-g))/(SampleSize*SqrtRRR);
		prevSampleSize=SampleSize;
		prevTheta=ThetaDIG;
		prevPopSize=Mig->PopSize[0];
	}

#ifdef DEBUG
	if(DebugPiSMM) {
		cout << "pGSM="<< g << endl;
		cout << "RRho="<< RRho << endl;
		cout << "RRR="<< RRR << endl;
	}
#endif

	for(k=AllelicRange[0];k<=AllelicRange[1];k++){
        if(Effectif[0][k]>0.0) {

            //if(abs(ChosenType-k) > Sam_AlleleNbr/2) StepNbr=Sam_AlleleNbr-abs(ChosenType-k); else StepNbr=abs(ChosenType-k);//car GSM circulaire en DG
            //RL devrait être plus rapide
            StepNbr=min(abs(ChosenType-k),Sam_AlleleNbr-abs(ChosenType-k));
            if(StepNbr!=0) pib+=Effectif[0][k]*pow(RRRinvSqrtRRR,StepNbr)*( term - g*( RRRinvSqrtRRR + 1.0/RRRinvSqrtRRR ) )*invDenom;
                else pib+=Effectif[0][k]*1*( term - g*( RRRinvSqrtRRR + RRRinvSqrtRRR ) )*invDenom;
#ifdef DEBUG
            if(DebugPiSMM) {
				cout<< "k="<<k;
				cout << ", ccc(l=" << StepNbr<<",RRho)="<<pow( RRRinvSqrtRRR,StepNbr)/SqrtRRR<< endl;
				cout << "\tccc(l=" << StepNbr<<"+1,RRho)="<<pow( RRRinvSqrtRRR,StepNbr+1)/SqrtRRR<<endl;
				cout << "\tccc(l=" << StepNbr<<"-1,RRho)="<<pow( RRRinvSqrtRRR,abs(StepNbr-1))/SqrtRRR<<endl;
				cout << "\tI(l=" << StepNbr<<",RRho)="<<(term*pow( RRRinvSqrtRRR,StepNbr) - g*( pow( RRRinvSqrtRRR,StepNbr+1) + pow( RRRinvSqrtRRR,abs(StepNbr-1)) ))*invDenom/(1-RRho)*SampleSize<< endl;
				cout << "\t\t -> Pib+="<<Effectif[0][k]*( term*pow( RRRinvSqrtRRR,StepNbr) - g*( pow( RRRinvSqrtRRR,StepNbr+1) + pow( RRRinvSqrtRRR,abs(StepNbr-1)) ) )*invDenom<<endl;
			}
#endif
        }// end if(Effectif>0)
    }//end for(k=AllelicRange...
} else {
	cerr <<"Error in PiOnePop Computations, MutationModel should be PIM, SMM or GSM"<< endl;
	if(cinGetOnError) cin.get();
	exit(-1);
}
}//end SMM/GSM en DGLR2005 - FR-RL-PP 2011


/*f(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) {
	expFactor=exp(-0.075*pow( (long double) (scaled_generation-Mig->PrevChangeTime), (long double) 0.5)
				  -0.1*pow((long double)(scaled_generation-Mig->PrevChangeTime), (long double) 4));

	//(0.036257 - 0.798301*(scaled_generation-Mig->PrevChangeTime)
	//			+ 7.33485*pow((scaled_generation-Mig->PrevChangeTime),2)
	//			- 29.9804*pow((scaled_generation-Mig->PrevChangeTime),3)
	//			+ 45.1554*pow((scaled_generation-Mig->PrevChangeTime),4)
	//			- 1/(1+ThetaDIG*pow(PopSizeRatio,(scaled_generation-Mig->PrevChangeTime)/(Mig->MinChangeTime-Mig->PrevChangeTime))))
	//			/(1/(1+ThetaDIG*PopSizeRatio)
	//			- 1/(1+ThetaDIG*pow(PopSizeRatio,(scaled_generation-Mig->PrevChangeTime)/(Mig->MinChangeTime-Mig->PrevChangeTime))));
//
//	cout << "term1=" << 0.06791996453432293 + 0.007683218773999549*(scaled_generation-Mig->PrevChangeTime)
//	+ 0.00035091585796903857*pow((scaled_generation-Mig->PrevChangeTime),2)
//	- 7.139381303126858e-6*pow((scaled_generation-Mig->PrevChangeTime),3)
//	+ 5.3584885387418616e-8*pow((scaled_generation-Mig->PrevChangeTime),4)
//	<< endl;
//	cout << "term2=" << 1/(1+ThetaDIG*pow(PopSizeRatio,(scaled_generation-Mig->PrevChangeTime)/(Mig->MinChangeTime-Mig->PrevChangeTime))) << endl;
//	cout << "term3=" << 1/(1+ThetaDIG) << endl;


	//exp(0.5*((scaled_generation-Mig->MinChangeTime)*ThetaDIG
	//			  +4*(Mig->MinChangeTime-Mig->PrevChangeTime)*(1./PopSizeRatio-
	//			  pow(PopSizeRatio,-(scaled_generation-Mig->PrevChangeTime)/(Mig->MinChangeTime-Mig->PrevChangeTime)))/log(PopSizeRatio)));

	altPIs[ChosenType][0]=expFactor*altpib+(1-expFactor)*pib;
	if(PIjabool) proposalCorrectionFactor=(expFactor*((SampleSize/PopSizeRatio+ThetaDIG)*altpib
										   +Effectif[0][ChosenType]/Mig->PopSize[0]-Effectif[0][ChosenType]/PopSizeRatio)
										   +(1-expFactor)*(SampleSize/Mig->PopSize[0]+ThetaDIG)*pib)
										   /((SampleSize/Mig->PopSize[0]+ThetaDIG)*altPIs[ChosenType][0]);//correction so that the sum of all proposal probabilities in cumul is 1
	//cout << "altpib=" << altpib << " and pib=" << pib << endl;
	//cout << "expfactor=" << expFactor << endl;
	//cout << " -> altPIs==expFactor*altpib+(1-expFactor)*pib=" << altPIs[ChosenType][0] << endl;
}*/
PIs[ChosenType][0]=pib;


//cout << proposalCorrectionFactor << endl;
//cin.get();

#ifdef DEBUG
if(DebugPiSMM) {
	printf("Pi_OnePop(%d,%d)->",ChosenType,0);
/*	if(AlternativeIS_FRExpMixAncActPopSizebool && !Mig->InAncStablePopbool) {
		cout << "\tPopSizeRatio=" << PopSizeRatio << endl;
		cout << "\texpFactor=" << expFactor << endl;
		cout << "\taltpib=" << altpib << " and pib=" << pib << endl;
		cout << "\texpfactor=" << expFactor << endl;
		cout << "\tpib=" << pib << endl;
		cout << " \t-> altPIs=expFactor*altpib+(1-expFactor)*pib=" << altPIs[ChosenType][0] << endl;
	} else */ cout << "PIs["<<ChosenType<<"][0]=" << PIs[ChosenType][0] << endl<< endl;
	fflush(stdout);
}
#endif

return 0;
}
/*-----------------------------------------------------------------------------*/


/****************************************************************************/
int inversion( picalctype** A,int _Size){
int i,j,im,jm,id,l,k,inte;
picalctype piv,ap,temp;
vector<int> ir,ic;

ir.resize(_Size);
ic.resize(_Size);

if(_Size==1) {
	if(A[0][0]==0) {printf("pb inversion d'une matrice nulle a une dimension");exit(1);}
	 else A[0][0]=1/A[0][0];
} else {//debut boucle1
	for(id=0;id<_Size;id++) {
		piv=A[id][id];im=id;jm=id;
		for(i=id;i<_Size;i++)
		 for(j=id;j<_Size;j++)
		  if(fabs(piv)<fabs(A[i][j])) {piv=A[i][j];im=i;jm=j;}
		if(fabs(piv)<pow((double)10,-30)) {
			printf("pb 1 a l'inversion de la matrice");
			exit(1);
		} else {//debut boucle2
			if(im!=id) for(j=0;j<_Size;j++) {
				temp=A[id][j];
				A[id][j]=A[im][j];
				A[im][j]=temp;
			}
			ir[id]=im;
			if(jm!=id) for(i=0;i<_Size;i++) {
				temp=A[i][id];
				A[i][id]=A[i][jm];
				A[i][jm]=temp;
			}
			ic[id]=jm;
			for(i=0;i<_Size;i++) {
				ap=A[i][id]/piv;
				for(j=0;j<_Size;j++)
				 if((i!=id)&&(j!=id)) A[i][j]=A[i][j]-ap*A[id][j];
			}
			for(i=0;i<_Size;i++) A[i][id]=A[i][id]/piv;
			for(j=0;j<_Size;j++) A[id][j]=A[id][j]/(-piv);
			A[id][id]=1.0/piv;
		}//fin boucle2
	}//fin boucle1

	for(l=0;l<_Size;l++) {
		k=_Size-1-l;
		if(ir[k]!=k) {
			inte=ir[k];
			for(i=0;i<_Size;i++) {
				temp=A[i][k];
				A[i][k]=A[i][inte];
				A[i][inte]=temp;
			}
		}
	}

	for(l=0;l<_Size;l++) {
		k=_Size-1-l;
		if(ic[k]!=k) {
			inte=ic[k];
			for(j=0;j<_Size;j++) {
				temp=A[k][j];
				A[k][j]=A[inte][j];
				A[inte][j]=temp;
			}
		}
	}

}//fin boucle0
return 0;
}

